﻿//  ******************************************
//  Frequency Manager Browse Window
//  Copyright © Jeff Knapp 2012-2013 
//
//  This window displays the frequency database records.
//  It allows a user to select a frequency and send it
//  to the radio, to edit a chosen frequency record, and to 
//  filter the grid by frequency, description, or group.
//  ******************************************
//  Date        Changes
//  01/31/13    made PostitionGridFreq select multiple matching rows.
//  02/02/13    support for editing multiple rows.
//  02/03/13    refactored positiongridbyfreq for better performance.
//  02/04/13    refactored positiongridbyfreqanddesc for better reliabilty and performance.
//              added double-buffer extension to grid for better performance.
//  02/15/13    fixed issue with live track no selecting rows
//  02/17/13    fixed issue with window size cause by closing while minimized
//  ******************************************

using System;
using System.Collections;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Data.SQLite;
using System.Diagnostics;
using System.Drawing;
using System.Globalization;
using System.IO;
using System.Linq;
using System.Reflection;
using System.Runtime.InteropServices;
using System.Runtime.Serialization.Formatters.Binary;
using System.Windows.Forms;
using SDRSharp.Radio;

namespace SDRSharp.FreqMgr
{
    internal sealed partial class Browse : Form
    {
        #region local vars

        internal DataGridViewCellStyle FlaggedFreqFormat = new DataGridViewCellStyle();
        internal DataGridViewCellStyle LockedFreqFormat = new DataGridViewCellStyle();
        internal DataGridViewCellStyle DefaultCellFormat = new DataGridViewCellStyle();
        internal event EventHandler<StationEditedEventArgs> StationEdited;
        internal event EventHandler<StationSelectedEventArgs> StationSelected;
        private Boolean _isResizing;
        private Boolean _isScanning;
        private Boolean _loadingGroups = true;
        private const DataGridViewContentAlignment MiddleRight = DataGridViewContentAlignment.MiddleRight;
        private const String Copyright = "Copyright © Jeff Knapp 2012-2013";
        private const String LiveTrackConfigKey = "freqMgrBrowseLiveTrack";
        private const String WindowConfigKey = "freqMgrBrowseWindow";
        private DataSet _ds;
        private readonly BackgroundWorker _positionGridByFreqAndDesc = new BackgroundWorker();
        private readonly BackgroundWorker _positionGridFreq = new BackgroundWorker();
        private readonly FreqMgrCollapsiblePanel _freqmgr;
        private readonly FreqRecord _fr = new FreqRecord();
        private readonly long _currFreq;


        #endregion

        /// <summary>
        /// Constructor
        /// </summary>
        /// <param name="currFrq">The frequency at which to position the gridview</param>
        /// <param name="fm">The form that opened this window</param>
        public Browse(long currFrq, FreqMgrCollapsiblePanel fm)
        {
            try
            {
                InitializeComponent();
                _currFreq = currFrq;
                _freqmgr = fm;

                //initialize the datagridview dataset
                _ds = new DataSet();
                _ds.Tables.Add(new DataTable());
                //initialize the flag filter dropdown
                filterFlagsList.SelectedIndex = 0;
                //load the group list
                LoadGroupList();
                //build the grid columns
                PrepareDGV();
                //load the datagridview with the initial data
                LoadDGV(_currFreq);
                
                //init window
                Size sz;
                Point lo;
                Common.GetWindowSettings(WindowConfigKey, out sz, out lo);
                //get window size
                if (!sz.IsEmpty)
                {
                    //we have to check to see if the browse window was minimized
                    //when it was saved; if not we can use the stored size
                    //otherwise we let Windows determine the size
                    if ((!lo.IsEmpty) &&
                        (lo.X > -32000) &&
                        (lo.Y > -32000))
                    {
                        Size = sz;
                    }
                }
                //get window position
                //we have to check to see if the browse window was minimized
                //when it was saved; if not we can use the stored location
                //otherwise we let Windows determine the location
                if ((!lo.IsEmpty) &&
                    (lo.X > -32000) &&
                    (lo.Y > -32000))
                {
                    Location = lo;
                }
                //get state of LiveTrack button
                liveTrack.Checked = Utils.GetBooleanSetting(LiveTrackConfigKey);
                //get whether or not the main form is scanning
                _isScanning = fm.IsScanning;
                //get colors for locked & flagged freqs
                LockedFreqFormat.BackColor = Utils.GetColorSetting("freqMgrLockedFreqBackColor", SystemColors.ControlDark);
                LockedFreqFormat.ForeColor = Utils.GetColorSetting("freqMgrLockedFreqForeColor", SystemColors.ButtonHighlight);
                FlaggedFreqFormat.BackColor = Utils.GetColorSetting("freqMgrFlaggedFreqBackColor", SystemColors.Info);
                FlaggedFreqFormat.ForeColor = Utils.GetColorSetting("freqMgrFlaggedFreqForeColor", SystemColors.ControlText);
                //get colors for default cell styles
                DefaultCellFormat = dgv.DefaultCellStyle;
                //attach events
                _freqmgr.ScannerStateChanged += FreqMgrScannerStateChanged;
                _freqmgr.GroupsUpdated += GroupsUpdated;
                LocationChanged += BrowseLocationChanged;
                Closing += BrowseClosing;
                _positionGridFreq.DoWork += PositionGridByFreq;
                _positionGridByFreqAndDesc.DoWork += PositionGridByFreqAndDesc;

#if (TESTING)
                Location = new Point(1030, -900);
                Size = new Size(650, 860);
                liveTrack.Checked = true;

#endif
            }
            catch (Exception ex)
            {
                MessageBox.Show(String.Format("Error in Browse ctor: {0}\n\r{1}", ex.Message, ex.StackTrace), "Error",
                                MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
        }

        #region control events

        /// <summary>
        /// toggles the flagged/notflagged status of the selected records
        /// </summary>
        private void FlaggedMenuClick(object sender, EventArgs e)
        {
            if (dgv.SelectedRows.Count == 0) return;
            dgv.SuspendLayout();

            try
            {
                Int32 gridFirstSelectedRowIndex;
                Int32 gridFirstRow = dgv.FirstDisplayedScrollingRowIndex;
                Int32 topRowFreq = Convert.ToInt32(dgv.SelectedRows[0].Cells["Frequency"].Value);
                using (SQLiteConnection conn = Common.GetSqlConnection())
                using (SQLiteTransaction sqlTransaction = conn.BeginTransaction())
                {
                    gridFirstSelectedRowIndex = dgv.SelectedRows[0].Index;
                    foreach (DataGridViewRow i in dgv.SelectedRows)
                    {
                        if (Convert.ToInt64(i.Cells["Locked"].Value) == 1) continue;
                        //set or clear the flag
                        const String upd = "update data set flagged = @flagged where Id = @freqId";
                        var cmd = new SQLiteCommand(upd, conn);
                        long status = Convert.ToInt64(i.Cells["Flagged"].Value) == 0 ? 1 : 0;
                        cmd.Parameters.AddWithValue("@flagged", status);
                        cmd.Parameters.AddWithValue("@freqId", Convert.ToInt32(i.Cells["Id"].Value));
                        cmd.ExecuteNonQuery();
                    }
                    sqlTransaction.Commit();
                }
                //reload and reposition the grid
                LoadDGV(topRowFreq);
                if (dgv.Rows.Count > 0)
                {
                    dgv.FirstDisplayedScrollingRowIndex = gridFirstRow;
                    dgv.ClearSelection();
                    try
                    {
                        dgv.Rows[gridFirstSelectedRowIndex].Selected = true;
                    }
                    catch (Exception)
                    {
                        dgv.Rows[gridFirstRow].Selected = true;
                    }
                }
            }
            catch (Exception ex)
            {
                MessageBox.Show(String.Format("Error in FlaggedMenuClick: {0}\n\r{1}", ex.Message, ex.StackTrace),
                                "Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
            dgv.ResumeLayout(true);
            //reload the freqmgrpanel station lookup table
            StationEdited(this, new StationEditedEventArgs(new FreqRecord()));
        }

        /// <summary>
        /// toggles the locked/notlocked status of the selected records
        /// </summary>
        private void LockedMenuClick(object sender, EventArgs e)
        {
            if (dgv.SelectedRows.Count == 0) return;
            dgv.SuspendLayout();
            try
            {
                Int32 gridFirstSelectedRowIndex;
                Int32 gridFirstRow = dgv.FirstDisplayedScrollingRowIndex;
                Int32 topRowFreq = Convert.ToInt32(dgv.SelectedRows[0].Cells["Frequency"].Value);
                using (SQLiteConnection conn = Common.GetSqlConnection())
                using (SQLiteTransaction sqlTransaction = conn.BeginTransaction())
                {
                    gridFirstSelectedRowIndex = dgv.SelectedRows[0].Index;
                    foreach (DataGridViewRow i in dgv.SelectedRows)
                    {
                        if (Convert.ToInt64(i.Cells["Flagged"].Value) == 1) continue;
                        //set or clear the lock
                        const String upd = "update data set locked = @locked where Id = @freqId";
                        var cmd = new SQLiteCommand(upd, conn);
                        long status = Convert.ToInt64(i.Cells["Locked"].Value) == 0 ? 1 : 0;
                        cmd.Parameters.AddWithValue("@locked", status);
                        cmd.Parameters.AddWithValue("@freqId", Convert.ToInt32(i.Cells["Id"].Value));
                        cmd.ExecuteNonQuery();
                    }
                    sqlTransaction.Commit();
                }
                LoadDGV(topRowFreq);
                if (dgv.Rows.Count > 0)
                {
                    dgv.FirstDisplayedScrollingRowIndex = gridFirstRow;
                    dgv.ClearSelection();
                    try
                    {
                        dgv.Rows[gridFirstSelectedRowIndex].Selected = true;
                    }
                    catch (Exception)
                    {
                        dgv.Rows[gridFirstRow].Selected = true;
                    }
                }
            }
            catch (Exception ex)
            {
                MessageBox.Show(String.Format("Error in LockedMenuClick: {0}\n\r{1}", ex.Message, ex.StackTrace),
                                "Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
            dgv.ResumeLayout(true);
            //reload the freqmgrpanel station lookup table
            StationEdited(this, new StationEditedEventArgs(new FreqRecord()));
        }

        /// <summary>
        /// filter the grid by flags
        /// </summary>
        private void FilterFlagsSelectedValueChanged(object sender, EventArgs e)
        {
            dgv.ClearSelection();
            dgv.MultiSelect = false;
            if ((filterFlagsList.Text != "Neither") && (String.IsNullOrEmpty(filterFlagsList.Text) == false))
            {
                ClearFilter();
                filterText.Enabled = false;
                clearFilterButton.Enabled = false;
                groupList.Enabled = false;
                lock (_ds)
                {
                    _ds.Tables[0].DefaultView.RowFilter = filterFlagsList.Text == "Flagged" ? "Flagged = 1" : "Locked = 1";
                }
                lock (dgv)
                {
                    if (dgv.Rows.Count > 0)
                    {
                        dgv.Rows[0].Selected = true;
                        dgv.FirstDisplayedScrollingRowIndex = 0;
                    }
                }
            }
            else
            {
                ClearFilter();
                filterText.Enabled = true;
                clearFilterButton.Enabled = true;
                groupList.Enabled = true;
            }
            dgv.MultiSelect = true;
            SetWindowText();
        }

        /// <summary>
        /// saves grid column settings to config
        /// </summary>
        private void BrowseClosing(object sender, CancelEventArgs e)
        {
            List<ColInfo> ci =
                (from DataGridViewColumn c in dgv.Columns
                 select new ColInfo { Name = c.Name, Index = c.DisplayIndex, Width = c.Width }).ToList();
            var ser = new BinaryFormatter();
            byte[] bytes;
            using (var ms = new MemoryStream())
            {
                ser.Serialize(ms, ci);
                bytes = ms.ToArray();
            }
            String s = Convert.ToBase64String(bytes);
            Utils.SaveSetting("freqMgrBrowseColumnsConfig", s);
            Common.SaveWindowSettings(WindowConfigKey, Size, Location);
        }

        /// <summary>
        /// gets grid column settings from config
        /// and configures the columns
        /// </summary>
        private void GetColumnSettings()
        {
            try
            {
                String s = Utils.GetStringSetting("freqMgrBrowseColumnsConfig", String.Empty);
                if (s.Length <= 0) return;
                List<ColInfo> ci;
                byte[] b = Convert.FromBase64String(s);
                using (var stream = new MemoryStream(b))
                {
                    var formatter = new BinaryFormatter();
                    stream.Seek(0, SeekOrigin.Begin);
                    ci = (List<ColInfo>)formatter.Deserialize(stream);
                }
                foreach (ColInfo c in ci)
                {
                    dgv.Columns[c.Name].DisplayIndex = c.Index;
                    dgv.Columns[c.Name].Width = c.Width;
                }
            }
            catch (Exception ex)
            {
                MessageBox.Show(String.Format("Error in GetColumnSettings: {0}\n\r{1}", ex.Message, ex.StackTrace),
                                "Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
        }

        /// <summary>
        /// saves the window's location to the config file
        /// </summary>
        private void BrowseLocationChanged(object sender, EventArgs e)
        {
            Common.SaveWindowSettings(WindowConfigKey, Size, Location);
        }

        /// <summary>
        /// Handles a double-click on the datagridview and
        /// sends the row data to the radio
        /// </summary>
        private void DGVDoubleClick(object sender, EventArgs e)
        {
            if ((dgv.CurrentRow == null) || (dgv.CurrentRow.Index <= -1) || (dgv.SelectedRows.Count > 1)) return;
            SaveRowValues();
            StationSelected(this, new StationSelectedEventArgs(_fr));
        }

        /// <summary>
        /// Handles formatting cells during datagridview loading
        /// </summary>
        private void DGVCellFormatting(object sender, DataGridViewCellFormattingEventArgs e)
        {
            try
            {
                String colname = dgv.Columns[e.ColumnIndex].Name;
                //we want to display our data zeroes as blanks and ones as 'Yes'
                if ((colname == "Squelch Enabled") || (colname == "Shift Enabled"))
                {
                    String testValue = e.Value.ToString();
                    if ((testValue == "0") || (String.IsNullOrEmpty(testValue)))
                    {
                        e.Value = String.Empty;
                    }
                    else
                    {
                        e.Value = "Yes";
                    }
                }
                //format date
                if (colname == "Date")
                {
                    if (!String.IsNullOrEmpty(e.Value.ToString()))
                    {
                        e.Value = Convert.ToDateTime(e.Value).ToString(Common.DateOnlyFormat);
                    }
                }
                try
                {
                    //format locked freq
                    if (Convert.ToBoolean(dgv.Rows[e.RowIndex].Cells["Locked"].Value))
                    {
                        dgv.Rows[e.RowIndex].DefaultCellStyle = LockedFreqFormat;
                    }
                    //format flagged freq
                    if (Convert.ToBoolean(dgv.Rows[e.RowIndex].Cells["Flagged"].Value))
                    {
                        dgv.Rows[e.RowIndex].DefaultCellStyle = FlaggedFreqFormat;
                    }
                }
                catch (Exception)
                {
                    //swallow it
                }
            }
            catch (Exception ex)
            {
                MessageBox.Show(String.Format("Error in DGVCellFormatting: {0}\n\r{1}", ex.Message, ex.StackTrace),
                                "Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
        }

        /// <summary>
        /// sends the selected station to the radio
        /// </summary>
        private void SendToRadioButtonClick(object sender, EventArgs e)
        {
            if ((dgv.SelectedRows.Count <= 0) || (dgv.SelectedRows.Count > 1))
                return; //have to have only 1 selected            
            DGVDoubleClick(new object(), new EventArgs());
        }

        /// <summary>
        /// handles the Close button click
        /// </summary>
        private void CloseButtonClick(object sender, EventArgs e)
        {
            Close();
        }

        /// <summary>
        /// handles the deletion of a selected frequency
        /// </summary>
        private void DeleteButtonClick(object sender, EventArgs e)
        {
#if (LOG)
            Common.Log("enter");
#endif
            try
            {
                if (dgv.SelectedRows.Count <= 0) return; //must have 1 or more selected
                //save the first record's data for use when reloading grid later
                var targetFreq = (long)dgv.SelectedRows[0].Cells["Frequency"].Value;
                var currRecord = (Int64)dgv.SelectedRows[0].Cells["Id"].Value;
                //save the index of the row at the top of the grid for use when reloading grid later
                Int32 topRow = dgv.FirstDisplayedScrollingRowIndex;
                //save the row index of the first selected row for use when reloading grid later
                Int32 selectedRow = dgv.SelectedRows[0].Index;
                if (currRecord > 0)
                {
                    DialogResult result;
                    //get permission to proceed
                    if (dgv.SelectedRows.Count == 1)
                    {
                        String msg = String.Format(
                            "{0}Hz\n\r{1}\n\r\n\rAre you sure you want to delete this Frequency?",
                            dgv.SelectedRows[0].Cells["Frequency"].EditedFormattedValue,
                            dgv.SelectedRows[0].Cells["Description"].Value);
                        result = MessageBox.Show(msg, "Delete Record", MessageBoxButtons.OKCancel,
                                                 MessageBoxIcon.Warning);
                    }
                    else
                    {
                        String msg =
                            String.Format(
                                "DELETE  {0}  FREQUENCIES\n\r\n\rAre you sure you want to delete these Frequencies?",
                                dgv.SelectedRows.Count);
                        result = MessageBox.Show(msg, "Delete Records", MessageBoxButtons.OKCancel,
                                                 MessageBoxIcon.Warning);
                    }

                    if (result == DialogResult.OK)
                    {
                        const String delFreq = "delete from Data where Id = @id";
                        const String delGroup = "delete from GroupFreq where freqId = @id";
                        using (SQLiteConnection conn = Common.GetSqlConnection())
                        using (SQLiteTransaction sqlTransaction = conn.BeginTransaction())
                        {
                            //delete the records
                            DataGridViewSelectedRowCollection records = dgv.SelectedRows;
                            foreach (DataGridViewRow item in records)
                            {
                                //delete from Data
                                var cmd = new SQLiteCommand(delFreq, conn);
                                var recId = (Int64)item.Cells["Id"].Value;
                                cmd.Parameters.AddWithValue("@id", recId);
                                cmd.ExecuteNonQuery();
                                //delete from GroupFreq
                                cmd = new SQLiteCommand(delGroup, conn);
                                cmd.Parameters.AddWithValue("@id", recId);
                                cmd.ExecuteNonQuery();
                            }
                            sqlTransaction.Commit();
                        }
                        //reload grid
                        LoadDGV(targetFreq);
                        //reset top row if necessary
                        if ((dgv.Rows.Count > 0) && (topRow > 0))
                        {
                            try
                            {
                                dgv.FirstDisplayedScrollingRowIndex = topRow;
                                dgv.Rows[selectedRow].Selected = true;
                            }
                            catch (Exception)
                            {
                                //swallow it
                            }
                        }
                    }
                }
                //reload the freqmgrpanel station lookup table
                StationEdited(this, new StationEditedEventArgs(new FreqRecord()));
            }
            catch (Exception ex)
            {
                MessageBox.Show(
                    String.Format("Error in Browse DeleteButtonClick: {0}\n\r{1}", ex.Message, ex.StackTrace), "Error",
                    MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
#if (LOG)
            Common.Log("exit");
#endif
        }

        /// <summary>
        /// delete the contents of the filter textbox
        /// thereby removing the filter
        /// </summary>
        private void ClearFilterButtonClick(object sender, EventArgs e)
        {
            if (String.IsNullOrEmpty(filterText.Text)) return;
            ClearFilter();
        }

        /// <summary>
        /// removes the filter from the dgv
        /// </summary>
        private void ClearFilter()
        {
            try
            {
                DataGridViewRow row = null;
                long freq = 0;
                String desc = String.Empty;
                if (dgv.SelectedRows.Count > 0)
                {
                    row = dgv.SelectedRows[0];
                    freq = (long)row.Cells["Frequency"].Value;
                    desc = row.Cells["Description"].Value.ToString();
                }
                filterText.Text = String.Empty;
                if (_ds != null)
                {
                    lock (_ds)
                    {
                        _ds.Tables[0].DefaultView.RowFilter = "";
                    }
                }
                if (freq > 0)
                {
                    PositionGrid(freq, desc);
                }
                if (row == null) return;
                row.Dispose();
            }
            catch (Exception ex)
            {
                MessageBox.Show(String.Format("Error in Browse ClearFilter: {0}\n\r{1}", ex.Message, ex.StackTrace),
                                "Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
        }

        /// <summary>
        /// deselects any selected row; used when there is no frequency
        /// match between what's selected in the spectrum window
        /// and the database.
        /// </summary>
        public void DeselectCurrentRow()
        {
            try
            {
                if ((!liveTrack.Checked) || (dgv.SelectedRows.Count <= 0)) return;
                dgv.ClearSelection();
            }
            catch (Exception)
            {
                //swallow it
            }
        }

        /// <summary>
        /// display the edit form for the selected record
        /// </summary>
        private void EditButtonClick(object sender, EventArgs e)
        {
            if (dgv.SelectedRows.Count <= 0) return;
            try
            {
                if (dgv.SelectedRows.Count == 1)
                {
                    #region edit single frequency

                    //build the frequency data record for passage to the edit form
                    DataGridViewRow row = dgv.SelectedRows[0];
                    var rec = new FreqRecord
                    {
                        Id = Convert.ToInt32(row.Cells["Id"].Value),
                        Frequency = (long)row.Cells["Frequency"].Value,
                        Center = (long)row.Cells["Center"].Value,
                        Description = row.Cells["Description"].Value.ToString(),
                        Mode = row.Cells["Mode"].Value.ToString(),
                        Protocol = row.Cells["Protocol"].Value.ToString(),
                        Callsign = row.Cells["Call Sign"].Value.ToString(),
                        Service = row.Cells["Service"].Value.ToString(),
                        ShiftEnabled = Convert.ToInt32(row.Cells["Shift Enabled"].Value),
                        Shift = (long)row.Cells["Shift"].Value,
                        FilterType = row.Cells["Filter Type"].Value.ToString(),
                        FilterBandwidth = Convert.ToInt32(row.Cells["Filter Bandwidth"].Value),
                        FilterOrder = Convert.ToInt32(row.Cells["Filter Order"].Value),
                        SquelchEnabled = Convert.ToInt32(row.Cells["Squelch Enabled"].Value),
                        Squelch = Convert.ToInt32(row.Cells["Squelch"].Value),
                        CWShift = Convert.ToInt32(row.Cells["CW Shift"].Value),
                        Notes = row.Cells["Notes"].Value.ToString(),
                        Date = row.Cells["Date"].Value.ToString(),
                        Flagged = Convert.ToBoolean(row.Cells["Flagged"].Value),
                        Locked = Convert.ToBoolean(row.Cells["Locked"].Value)
                    };

                    //set the remaining parms                    
                    Int32 filterTypeSelectedIndex = Array.IndexOf(Common.FilterTypeArray, rec.FilterType);

                    //display the edit form
                    using (var form = new Edit(rec, Common.FilterTypeArray, filterTypeSelectedIndex, rec.Mode))
                    {
                        //set the form's icon
                        form.Icon = Icon;
                        //save the current row
                        Int32 currRow = dgv.SelectedRows[0].Index;
                        //save the current top row's position
                        Int32 currTopRow = dgv.FirstDisplayedScrollingRowIndex;
                        //display the edit form modally and wait for return
                        if (form.ShowDialog() != DialogResult.OK) return;

                        //data was added or changed; update sdr# and FreqMgr
                        rec = form.RecordData;
                        LoadDGV(rec.Frequency);
                        _loadingGroups = true;
                        GroupListSelectedValueChanged(new object(), new EventArgs());
                        //determine whether add or update
                        if (form.FreqUpdated)
                        {
                            try
                            {
                                dgv.ClearSelection();
                                dgv.FirstDisplayedScrollingRowIndex = currTopRow;
                                dgv.Rows[currRow].Selected = true;
                            }
                            catch (Exception)
                            {
                                //swallow it
                            }
                        }
                        else //new record
                        {
                            PositionGrid(rec.Frequency);
                        }
                        //reload the freqmgrpanel station lookup table
                        StationEdited(this, new StationEditedEventArgs(rec));
                        _loadingGroups = false;
                    }

                    #endregion
                }
                else
                {
                    #region edit multiple frequencies
                    //store info about the selected rows
                    long[] rows = new long[dgv.SelectedRows.Count]; //the database record Id
                    Int32[] selectedRows = new Int32[dgv.SelectedRows.Count]; //the grid row id
                    for (int i = dgv.SelectedRows.Count - 1; i >= 0; --i)
                    {
                        rows[i] = (long)dgv.SelectedRows[i].Cells["Id"].Value;
                        selectedRows[i] = dgv.SelectedRows[i].Index;
                    }

                    //display the edit form
                    using (var form = new Edit(rows, Common.FilterTypeArray))
                    {
                        //set the form's icon
                        form.Icon = Icon;
                        //save the top selected row's frequency for later reference
                        long currFreq = (long)dgv.SelectedRows[selectedRows.Length - 1].Cells["Frequency"].Value;
                        //display the edit form modally and wait for return
                        if (form.ShowDialog() != DialogResult.OK) return;

                        //position the grid to the first selected freq 
                        //and then re-select the others
                        LoadDGV(currFreq);
                        try
                        {
                            foreach (int t in selectedRows)
                            {
                                dgv.Rows[t].Selected = true;
                            }
                        }
                        catch (Exception)
                        {
                            //swallow it
                        }
                        //reload the freqmgrpanel station lookup table
                        StationEdited(this, new StationEditedEventArgs(null));
                    }
                    #endregion
                }
                SetButtonStates();
            }
            catch (Exception ex)
            {
                String msg = String.Format("An error occurred.  The error was: \n\r{0}\n\r{1}", ex.Message, ex.StackTrace.Trim());
                MessageBox.Show(msg, "Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
        }

        /// <summary>
        /// set/clear the dgv filter based on the groups dropdown
        /// </summary>
        private void GroupListSelectedValueChanged(object sender, EventArgs e)
        {
            if (_loadingGroups) return;
            dgv.ClearSelection();
            dgv.MultiSelect = false;
            if ((groupList.Text != Common.NoneText) && (String.IsNullOrEmpty(groupList.Text) == false))
            {
                ClearFilter();
                filterText.Enabled = false;
                clearFilterButton.Enabled = false;
                filterFlagsList.Enabled = false;
                lock (dgv)
                {
                    _ds.Tables[0].DefaultView.RowFilter = String.Format("Groups LIKE '*{0}*'", groupList.Text);
                    if (dgv.Rows.Count > 0)
                    {
                        dgv.Rows[0].Selected = true;
                        dgv.FirstDisplayedScrollingRowIndex = 0;
                    }
                }
            }
            else
            {
                ClearFilter();
                filterText.Enabled = true;
                clearFilterButton.Enabled = true;
                filterFlagsList.Enabled = true;
            }
            dgv.MultiSelect = true;
            SetWindowText();
        }

        /// <summary>
        /// processes a right-click on the datagridview
        /// </summary>
        private void DGVCellMouseClick(object sender, DataGridViewCellMouseEventArgs e)
        {
            if (e.RowIndex < 0 || e.ColumnIndex < 0 || e.Button != MouseButtons.Right) return;
            //select the row and determine the menu position
            dgv.Rows[e.RowIndex].Selected = true;
            Rectangle rect = dgv.GetCellDisplayRectangle(e.ColumnIndex, e.RowIndex, true);

            //disable the sendToRadioButton if more than one row is selected
            sendToRadioButton.Enabled = dgv.SelectedRows.Count == 1;

            //disable the other menus and buttons as is appropriate
            if ((_isScanning) && (liveTrack.Checked))
            {
                deleteButton.Enabled = false;
                menuEdit.Enabled = false;
                menuDelete.Enabled = false;
                menuFlagged.Enabled = false;
                menuLocked.Enabled = false;
                menuGroupAdd.Enabled = false;
                menuGroupRemove.Enabled = false;
            }
            else
            {
                menuDelete.Enabled = true;
                deleteButton.Enabled = true;
                menuFlagged.Enabled = true;
                menuLocked.Enabled = true;
                menuGroupAdd.Enabled = true;
                menuGroupRemove.Enabled = true;
                //populate the Add menu
                menuGroupAdd.MenuItems.Clear();
                MenuItem[] groups = GetGroupMenuItems(AddGroupItemClick);
                menuGroupAdd.MenuItems.AddRange(groups);

                //populate the Remove menu
                menuGroupRemove.MenuItems.Clear();
                groups = GetGroupMenuItems(RemoveGroupItemClick);
                menuGroupRemove.MenuItems.AddRange(groups);
            }
            //show the menu
            dgvEditMenu.Show((Control)sender, new Point(rect.Left + e.X, rect.Top + e.Y));
        }

        /// <summary>
        /// enables/disables buttons based on number of rows that are selected
        /// </summary>
        private void DGVSelectionChanged(object sender, EventArgs e)
        {
           SetButtonStates();
        }

        /// <summary>
        /// handles the edit menu click; simulates clicking the Edit button
        /// </summary>
        private void EditMenuClick(object sender, EventArgs e)
        {
            editButton.PerformClick();
        }

        /// <summary>
        /// handles the delete menu click; simulates clicking the Delete button
        /// </summary>
        private void DeleteMenuClick(object sender, EventArgs e)
        {
            DeleteButtonClick(new object(), new EventArgs());
        }

        /// <summary>
        /// processes the click of a group from the Add Group submenu
        /// </summary>
        private void AddGroupItemClick(object sender, EventArgs e)
        {
#if (LOG)
            Common.Log("enter");
#endif
            try
            {
                dgv.SuspendLayout();
                Int32 gridFirstRow = dgv.FirstDisplayedScrollingRowIndex;
                Int32 topRow = Convert.ToInt32(dgv.SelectedRows[0].Cells["Frequency"].Value);
                var mi = (MenuItem)sender;
                using (SQLiteConnection conn = Common.GetSqlConnection())
                using (SQLiteTransaction sqlTransaction = conn.BeginTransaction())
                {
                    foreach (DataGridViewRow i in dgv.SelectedRows)
                    {
                        //first see if this freq is already assigned to this group
                        const string sel =
                            "select count(*) from GroupFreq where GroupId = @groupId and FreqId = @freqId";
                        var cmd = new SQLiteCommand(sel, conn);
                        cmd.Parameters.AddWithValue("@groupId", Convert.ToInt32(mi.Tag));
                        cmd.Parameters.AddWithValue("@freqId", Convert.ToInt32(i.Cells["Id"].Value));
                        Int32 ct = Convert.ToInt32(cmd.ExecuteScalar());
                        //skip it if the freq is already assigned to the group
                        if (ct > 0) continue;
                        //add the group assignment
                        const string ins = "insert into GroupFreq (GroupId,FreqId) values (@groupId,@freqId)";
                        cmd = new SQLiteCommand(ins, conn);
                        cmd.Parameters.AddWithValue("@groupId", Convert.ToInt32(mi.Tag));
                        cmd.Parameters.AddWithValue("@freqId", Convert.ToInt32(i.Cells["Id"].Value));
                        cmd.ExecuteNonQuery();
                    }
                    sqlTransaction.Commit();
                }
                LoadDGV(topRow);
                try
                {
                    dgv.FirstDisplayedScrollingRowIndex = gridFirstRow;
                }
                catch (Exception)
                {
                    //swallow it
                }
                dgv.ResumeLayout(true);
            }
            catch (Exception ex)
            {
                MessageBox.Show(
                    String.Format("Error in Browse DGVSelectionChanged: {0}\n\r{1}", ex.Message, ex.StackTrace), "Error",
                    MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
#if (LOG)
            Common.Log("exit");
#endif
        }

        /// <summary>
        /// process the click of a group from the Remove Group submenu
        /// </summary>
        private void RemoveGroupItemClick(object sender, EventArgs e)
        {
            dgv.SuspendLayout();
            Int32 gridFirstRow = dgv.FirstDisplayedScrollingRowIndex;
            Int32 topRow = Convert.ToInt32(dgv.SelectedRows[0].Cells["Frequency"].Value);
            var mi = (MenuItem)sender;
            using (SQLiteConnection conn = Common.GetSqlConnection())
            using (SQLiteTransaction sqlTransaction = conn.BeginTransaction())
            {
                foreach (string del in from DataGridViewRow i in dgv.SelectedRows
                                       select
                                           String.Format("delete from GroupFreq where GroupId = {0} and FreqId = {1}",
                                                         Convert.ToInt32(mi.Tag), Convert.ToInt32(i.Cells["Id"].Value)))
                {
                    var cmd = new SQLiteCommand(del, conn);
                    cmd.ExecuteNonQuery();
                }
                sqlTransaction.Commit();
            }
            LoadDGV(topRow);
            try
            {
                dgv.FirstDisplayedScrollingRowIndex = gridFirstRow;
            }
            catch (Exception)
            {
                //swallow it. this exception can occur when you delete all the freqs from
                //a group and you are filtering by that group - gridFirstRow
                //would end up being null
            }
            dgv.ResumeLayout(true);
        }

        /// <summary>
        /// sets button states based on checked property
        /// </summary>
        private void LiveTrackCheckChanged(object sender, EventArgs e)
        {
            Utils.SaveSetting(LiveTrackConfigKey, liveTrack.Checked);
            SetButtonStates();
        }

        /// <summary>
        /// handles ScannerStateChanged event from FreqMgr
        /// </summary>
        private void FreqMgrScannerStateChanged(object sender, FreqMgrCollapsiblePanel.ScannerStateChangedEventArgs e)
        {
            _isScanning = e.IsScanning;
            SetButtonStates();
        }

        /// <summary>
        /// Event sink for groupsUpdated
        /// </summary>
        private void GroupsUpdated(object sender, FreqMgrCollapsiblePanel.GroupsUpdatedEventArgs e)
        {
            //preserve the existing group selection
            String currGroup = Common.NoneText;
            var x = (DataRowView)groupList.SelectedItem;
            if (x != null)
            {
                currGroup = x.Row["Name"].ToString();
            }
            //load the group list
            LoadGroupList();
            //restore selection
            groupList.SelectedIndex = groupList.FindStringExact(currGroup, 0);
            if (groupList.SelectedIndex == -1)
            {
                groupList.SelectedIndex = groupList.FindStringExact(Common.NoneText, 0);
                groupList.Text = Common.NoneText;
            }
            groupList.Text = currGroup;
        }

        #endregion

        #region utility members

        /// <summary>
        /// Expose internal data to other classes
        /// </summary>
        public FreqRecord RecordData
        {
            get { return _fr; }
        }

        /// <summary>
        /// builds an array of menu items for the edit context menu
        /// </summary>
        /// <param name="handler">Name of the event handler for the menu items</param>
        /// <returns>MenuItem array</returns>
        private static MenuItem[] GetGroupMenuItems(EventHandler handler)
        {
            MenuItem[] mi = null;
            try
            {
                DataView dv = Common.GetGroupList(false);
                mi = new MenuItem[dv.Table.Rows.Count];
                for (int i = 0; i < dv.Table.Rows.Count; i++)
                {
                    DataRow dr = dv.Table.Rows[i];
                    var m = new MenuItem { Tag = dr["Id"], Text = dr["Name"].ToString() };
                    m.Click += handler;
                    mi[i] = m;
                }
            }
            catch (Exception ex)
            {
                MessageBox.Show(
                    String.Format("Error in Browse GetGroupMenuItems: {0}\n\r{1}", ex.Message, ex.StackTrace), "Error",
                    MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
            return mi;
        }


        /// <summary>
        /// loads the datagridview from the database
        /// </summary>
        /// <param name="targetFreq">frequency at which to position the datagridview</param>
        public void LoadDGV(long targetFreq)
        {
#if (LOG)
            Common.Log("enter");
#endif
            try
            {
                lock (dgv)
                {
                    dgv.SuspendLayout();
                    String sortCol = "Frequency";
                    //save existing sort order if any
                    var sortDir = SortOrder.Ascending;
                    if ((dgv.Columns.Count > 0) && (dgv.SortedColumn != null))
                    {
                        sortCol = dgv.SortedColumn.HeaderText;
                        sortDir = dgv.SortOrder;
                    }
                    //get data
                    const string sel =
                        "select d.id, d.Locked, d.Flagged, d.Frequency, d.Center, d.Description, " +
                        "d.Mode, d.Protocol, d.[Call Sign], d.Service, group_concat(g.Name) as Groups, " +
                        "d.[Shift Enabled], d.Shift, d.[Filter Type], d.[Filter Bandwidth], d.[Filter Order], " +
                        "d.[Squelch Enabled], d.Squelch, d.[CW Shift],d.Notes, Date(d.Date) as Date, " +
                        "cast(d.Frequency as varchar(12)) as FreqString from data as d " +
                        "left join groupfreq as gf on gf.freqId = d.id left join Groups as g " +
                        "on g.id = gf.groupid group by d.id order by d.Frequency, d.Description";
                    using (SQLiteConnection conn = Common.GetSqlConnection())
                    using (var cmd = new SQLiteCommand(sel, conn))
                    using (var da = new SQLiteDataAdapter(cmd))
                    {
                        lock (_ds)
                        {
                            _ds = new DataSet();
                            da.Fill(_ds);
                        }
                    }

                    //bind the new data
                    dgv.DataSource = _ds.Tables[0];

                    //restore the current sort column and order
                    if (sortCol != "Groups")
                    {
                        dgv.Columns[sortCol].SortMode = DataGridViewColumnSortMode.Automatic;
                        dgv.Sort(dgv.Columns[sortCol],
                                 sortDir == SortOrder.Ascending
                                     ? ListSortDirection.Ascending
                                     : ListSortDirection.Descending);
                        dgv.Columns[sortCol].HeaderCell.SortGlyphDirection = sortDir;
                    }
                    //position the grid to the target frequency
                    PositionGrid(targetFreq);
                    //reapply any existing filter
                    if (filterText.Text.Trim().Length > 0)
                    {
                        FilterTextTextChanged(new Object(), new EventArgs());
                    }
                    if (groupList.SelectedIndex != groupList.FindStringExact(Common.NoneText, 0))
                    {
                        GroupListSelectedValueChanged(new Object(), new EventArgs());
                    }
                    if (filterFlagsList.SelectedIndex != 0)
                    {
                        FilterFlagsSelectedValueChanged(new Object(), new EventArgs());
                    }
                    dgv.Refresh();
                    SetWindowText();
                }
            }
            catch (Exception ex)
            {
                MessageBox.Show(String.Format("Error in LoadDGV: {0}\n\r{1}", ex.Message, ex.StackTrace), "Error",
                                MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
            finally
            {
                dgv.ResumeLayout(true);
            }
#if (LOG)
            Common.Log("exit");
#endif
        }

        /// <summary>
        /// initializes the dgv columns
        /// </summary>
        private void PrepareDGV(){
            try
            {
                dgv.DoubleBuffered(true);
                //disable auto columns
                dgv.AutoGenerateColumns = false;

                //connect events
                dgv.DoubleClick += DGVDoubleClick;
                dgv.CellFormatting += DGVCellFormatting;

                var newCell = new DataGridViewTextBoxCell();

                var newColumn = new DataGridViewColumn
                                    {
                                        Name = "Id",
                                        CellTemplate = newCell,
                                        FillWeight = 1,
                                        SortMode = DataGridViewColumnSortMode.Automatic
                                    };
                dgv.Columns.Add(newColumn);
                dgv.Columns["Id"].Visible = false;
                dgv.Columns["Id"].DisplayIndex = 0;
                dgv.Columns["Id"].DataPropertyName = "Id";

                newColumn = new DataGridViewColumn
                                {
                                    Name = "Locked",
                                    CellTemplate = newCell,
                                    FillWeight = 1,
                                    SortMode = DataGridViewColumnSortMode.Automatic
                                };
                dgv.Columns.Add(newColumn);
                dgv.Columns["Locked"].Visible = false;
                dgv.Columns["Locked"].DataPropertyName = "Locked";

                newColumn = new DataGridViewColumn
                                {
                                    Name = "Flagged",
                                    CellTemplate = newCell,
                                    FillWeight = 1,
                                    SortMode = DataGridViewColumnSortMode.Automatic
                                };
                dgv.Columns.Add(newColumn);
                dgv.Columns["Flagged"].Visible = false;
                dgv.Columns["Flagged"].DataPropertyName = "Flagged";

                newColumn = new DataGridViewColumn
                                {
                                    Name = "Frequency",
                                    CellTemplate = newCell,
                                    FillWeight = 1,
                                    SortMode = DataGridViewColumnSortMode.Automatic
                                };
                dgv.Columns.Add(newColumn);
                dgv.Columns["Frequency"].HeaderCell.Style.Alignment = MiddleRight;
                dgv.Columns["Frequency"].DefaultCellStyle.Format = Common.NumericFormat;
                dgv.Columns["Frequency"].DefaultCellStyle.Alignment = MiddleRight;
                dgv.Columns["Frequency"].DataPropertyName = "Frequency";

                newColumn = new DataGridViewColumn
                                {
                                    Name = "Center",
                                    CellTemplate = newCell,
                                    FillWeight = 1,
                                    SortMode = DataGridViewColumnSortMode.Automatic
                                };
                dgv.Columns.Add(newColumn);
                dgv.Columns["Center"].HeaderCell.Style.Alignment = MiddleRight;
                dgv.Columns["Center"].DefaultCellStyle.Format = Common.NumericFormat;
                dgv.Columns["Center"].DefaultCellStyle.Alignment = MiddleRight;
                dgv.Columns["Center"].SortMode = DataGridViewColumnSortMode.Automatic;
                dgv.Columns["Center"].DataPropertyName = "Center";

                newColumn = new DataGridViewColumn
                                {
                                    Name = "Description",
                                    CellTemplate = newCell,
                                    FillWeight = 1,
                                    SortMode = DataGridViewColumnSortMode.Automatic
                                };
                dgv.Columns.Add(newColumn);
                dgv.Columns["Description"].Width = 300;
                dgv.Columns["Description"].DataPropertyName = "Description";

                newColumn = new DataGridViewColumn
                                {
                                    Name = "Mode",
                                    CellTemplate = newCell,
                                    FillWeight = 1,
                                    SortMode = DataGridViewColumnSortMode.Automatic
                                };
                dgv.Columns.Add(newColumn);
                dgv.Columns["Mode"].DataPropertyName = "Mode";

                newColumn = new DataGridViewColumn
                                {
                                    Name = "Protocol",
                                    CellTemplate = newCell,
                                    FillWeight = 1,
                                    SortMode = DataGridViewColumnSortMode.Automatic
                                };
                dgv.Columns.Add(newColumn);
                dgv.Columns["Protocol"].DataPropertyName = "Protocol";

                newColumn = new DataGridViewColumn
                                {
                                    Name = "Call Sign",
                                    CellTemplate = newCell,
                                    FillWeight = 1,
                                    SortMode = DataGridViewColumnSortMode.Automatic
                                };
                dgv.Columns.Add(newColumn);
                dgv.Columns["Call Sign"].DataPropertyName = "Call Sign";

                newColumn = new DataGridViewColumn
                                {
                                    Name = "Service",
                                    CellTemplate = newCell,
                                    FillWeight = 1,
                                    SortMode = DataGridViewColumnSortMode.Automatic
                                };
                dgv.Columns.Add(newColumn);
                dgv.Columns["Service"].DataPropertyName = "Service";

                newColumn = new DataGridViewColumn
                                {
                                    Name = "Groups",
                                    CellTemplate = newCell,
                                    FillWeight = 1,
                                    SortMode = DataGridViewColumnSortMode.Automatic
                                };
                dgv.Columns.Add(newColumn);
                dgv.Columns["Groups"].DataPropertyName = "Groups";

                newColumn = new DataGridViewColumn
                                {
                                    Name = "Shift Enabled",
                                    CellTemplate = newCell,
                                    FillWeight = 1,
                                    SortMode = DataGridViewColumnSortMode.Automatic
                                };
                dgv.Columns.Add(newColumn);
                dgv.Columns["Shift Enabled"].DataPropertyName = "Shift Enabled";

                newColumn = new DataGridViewColumn
                                {
                                    Name = "Shift",
                                    CellTemplate = newCell,
                                    FillWeight = 1,
                                    SortMode = DataGridViewColumnSortMode.Automatic
                                };
                dgv.Columns.Add(newColumn);
                dgv.Columns["Shift"].HeaderCell.Style.Alignment = MiddleRight;
                dgv.Columns["Shift"].DefaultCellStyle.Alignment = MiddleRight;
                dgv.Columns["Shift"].DefaultCellStyle.Format = Common.NumericFormat;
                dgv.Columns["Shift"].DataPropertyName = "Shift";

                newColumn = new DataGridViewColumn
                                {
                                    Name = "Filter Type",
                                    CellTemplate = newCell,
                                    FillWeight = 1,
                                    SortMode = DataGridViewColumnSortMode.Automatic
                                };
                dgv.Columns.Add(newColumn);
                dgv.Columns["Filter Type"].DataPropertyName = "Filter Type";

                newColumn = new DataGridViewColumn
                                {
                                    Name = "Filter Bandwidth",
                                    CellTemplate = newCell,
                                    FillWeight = 1,
                                    SortMode = DataGridViewColumnSortMode.Automatic
                                };
                dgv.Columns.Add(newColumn);
                dgv.Columns["Filter Bandwidth"].HeaderCell.Style.Alignment = MiddleRight;
                dgv.Columns["Filter Bandwidth"].DefaultCellStyle.Format = Common.NumericFormat;
                dgv.Columns["Filter Bandwidth"].DefaultCellStyle.Alignment = MiddleRight;
                dgv.Columns["Filter Bandwidth"].DataPropertyName = "Filter Bandwidth";

                newColumn = new DataGridViewColumn
                                {
                                    Name = "Filter Order",
                                    CellTemplate = newCell,
                                    FillWeight = 1,
                                    SortMode = DataGridViewColumnSortMode.Automatic
                                };
                dgv.Columns.Add(newColumn);
                dgv.Columns["Filter Order"].HeaderCell.Style.Alignment = MiddleRight;
                dgv.Columns["Filter Order"].DefaultCellStyle.Format = Common.NumericFormat;
                dgv.Columns["Filter Order"].DefaultCellStyle.Alignment = MiddleRight;
                dgv.Columns["Filter Order"].DataPropertyName = "Filter Order";

                newColumn = new DataGridViewColumn
                                {
                                    Name = "Squelch Enabled",
                                    CellTemplate = newCell,
                                    FillWeight = 1,
                                    SortMode = DataGridViewColumnSortMode.Automatic
                                };
                dgv.Columns.Add(newColumn);
                dgv.Columns["Squelch Enabled"].DataPropertyName = "Squelch Enabled";

                newColumn = new DataGridViewColumn
                                {
                                    Name = "Squelch",
                                    CellTemplate = newCell,
                                    FillWeight = 1,
                                    SortMode = DataGridViewColumnSortMode.Automatic
                                };
                dgv.Columns.Add(newColumn);
                dgv.Columns["Squelch"].HeaderCell.Style.Alignment = MiddleRight;
                dgv.Columns["Squelch"].DefaultCellStyle.Format = Common.NumericFormat;
                dgv.Columns["Squelch"].DefaultCellStyle.Alignment = MiddleRight;
                dgv.Columns["Squelch"].DataPropertyName = "Squelch";

                newColumn = new DataGridViewColumn
                                {
                                    Name = "CW Shift",
                                    CellTemplate = newCell,
                                    FillWeight = 1,
                                    SortMode = DataGridViewColumnSortMode.Automatic
                                };
                dgv.Columns.Add(newColumn);
                dgv.Columns["CW Shift"].HeaderCell.Style.Alignment = MiddleRight;
                dgv.Columns["CW Shift"].DefaultCellStyle.Format = Common.NumericFormat;
                dgv.Columns["CW Shift"].DefaultCellStyle.Alignment = MiddleRight;
                dgv.Columns["CW Shift"].DataPropertyName = "CW Shift";

                newColumn = new DataGridViewColumn
                                {
                                    Name = "Notes",
                                    CellTemplate = newCell,
                                    FillWeight = 1,
                                    SortMode = DataGridViewColumnSortMode.Automatic
                                };
                dgv.Columns.Add(newColumn);
                dgv.Columns["Notes"].DataPropertyName = "Notes";

                newColumn = new DataGridViewColumn
                                {
                                    Name = "Date",
                                    CellTemplate = newCell,
                                    FillWeight = 1,
                                    SortMode = DataGridViewColumnSortMode.Automatic
                                };
                dgv.Columns.Add(newColumn);
                dgv.Columns["Date"].HeaderCell.Style.Alignment = MiddleRight;
                dgv.Columns["Date"].DefaultCellStyle.Alignment = MiddleRight;
                dgv.Columns["Date"].DataPropertyName = "Date";

                newColumn = new DataGridViewColumn
                                {
                                    CellTemplate = newCell,
                                    FillWeight = 1,
                                    SortMode = DataGridViewColumnSortMode.Automatic
                                };
                dgv.Columns.Add(newColumn);
                dgv.Columns[dgv.Columns.Count - 1].Width = 40;

                newColumn = new DataGridViewColumn
                                {
                                    Name = "FreqString",
                                    CellTemplate = newCell,
                                    FillWeight = 1,
                                    SortMode = DataGridViewColumnSortMode.Automatic
                                };
                dgv.Columns.Add(newColumn);
                dgv.Columns["FreqString"].Visible = false;
                dgv.Columns["FreqString"].DataPropertyName = "FreqString";

                newCell.Dispose();
                newColumn.Dispose();

                //get the sequence and width from the config file if they exist
                GetColumnSettings();
            }
            catch (Exception ex)
            {
                MessageBox.Show(String.Format("Error in PrepareDGV: {0}\n\r{1}", ex.Message, ex.StackTrace), "Error",
                                MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
        }

        /// <summary>
        /// loads the group list from the database
        /// </summary>
        private void LoadGroupList()
        {
            try
            {
                _loadingGroups = true;
                groupList.DataSource = null;
                DataView dv = Common.GetGroupList(true);
                groupList.DataSource = dv;
                groupList.DisplayMember = "Name";
                groupList.ValueMember = "Id";
                groupList.SelectedIndex = groupList.FindStringExact(Common.NoneText, 0);
                _loadingGroups = false;
            }
            catch (Exception ex)
            {
                MessageBox.Show(String.Format("Error in LoadGroupList: {0}\n\r{1}", ex.Message, ex.StackTrace), "Error",
                                MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
            _loadingGroups = false;
        }

        /// <summary>
        /// positions the datagridview to the row 
        /// containing the specified frequency
        /// via a backgroundworker
        /// </summary>
        public void PositionGrid(long targetFreq)
        {
            if (
                (!liveTrack.Checked) ||
                (targetFreq <= 0) ||
                (dgv.Rows.Count == 0) ||
                (_isResizing)
                )
            {
                return;
            }
            if (!_positionGridFreq.IsBusy) _positionGridFreq.RunWorkerAsync(targetFreq);
        }

        /// <summary>
        /// positions the datagridview to the row 
        /// containing the specified frequency
        /// </summary>
        /// <remarks>Crude but effective</remarks>
        private void PositionGridByFreq(object sender, DoWorkEventArgs e)
        {
            long targetFreq = (long)e.Argument;
#if (LOG)
            Common.Log("1 enter");
#endif
            try
            {
                lock (dgv)
                {
                    dgv.SuspendLayout();
                    System.Threading.Thread.Sleep(5);
                    dgv.ClearSelection();
                    //save the index of the first row displayed for later use
                    int firstIndex = dgv.FirstDisplayedScrollingRowIndex;
                    try
                    {
                        //I tried this LINQ but it kept hanging the program when there wasn't 
                        //a row that qualified. It didn't throw exceptions, it just hung.
                        
                        //row = dgv.Rows.Cast<DataGridViewRow>().First(r => Convert.ToInt64(r.Cells["Frequency"].Value).Equals(targetFreq));

                        //this is crude, and I hate it, but it is reliable.
                        foreach (DataGridViewRow r in dgv.Rows)
                        {
                            if ((Convert.ToInt64(r.Cells["Frequency"].Value).Equals(targetFreq)) &&
                                (!Convert.ToBoolean(r.Cells["Locked"].Value)))
                            {
                                r.Selected = true;
                            }
                        }
                    }
                    catch (Exception)
                    {
                        //swallow it
#if (LOG)
                        Common.Log("1 find exception: " + ex.Message);
#endif
                    }
                    if (dgv.SelectedRows.Count == 0)
                    {
                        dgv.ResumeLayout();
#if (LOG)
                        Common.Log("1 not found, exit");
#endif
                        return;
                    }
                    //reposition to the first row of the grid if necessary
                    Int32 lastRowVisible = firstIndex + dgv.DisplayedRowCount(true);
                    if ((dgv.SelectedRows[0].Index > firstIndex) && (dgv.SelectedRows[0].Index < lastRowVisible))
                    {
                        dgv.ResumeLayout();
#if (LOG)
                        Common.Log("1 exit");
#endif
                        return;
                    }
                    dgv.FirstDisplayedScrollingRowIndex = dgv.SelectedRows[0].Index;
                    dgv.ResumeLayout();
                }
            }
            catch (ArgumentOutOfRangeException)
            {
                //swallow it
            }
            catch (InvalidOperationException)
            {
                //swallow it. this exception usually happens when the app 
                //starts and the current sdr# frequency is not in the grid
            }
            catch (Exception)
            {
                //swallow it
            }
#if (LOG)
            Common.Log("1 exit");
#endif
        }

        /// <summary>
        /// positions the datagridview to the row 
        /// containing the specified frequency and description
        /// via a backgroundworker
        /// </summary>
        public void PositionGrid(long targetFreq, String desc)
        {
            if (!liveTrack.Checked) return;
            if (targetFreq <= 0) return;
            if (dgv.Rows.Count == 0) return;
            Object[] args = {targetFreq,desc};            
            if (!_positionGridByFreqAndDesc.IsBusy) _positionGridByFreqAndDesc.RunWorkerAsync(args);
        }

        /// <summary>
        /// positions the datagridview to the row 
        /// containing the specified frequency and description
        /// </summary>
        /// <remarks>Crude but effective</remarks>
        private void PositionGridByFreqAndDesc(object sender, DoWorkEventArgs e)
        {
            Object[] args = (Object[]) e.Argument;
            long targetFreq = (long)args[0];
            String desc = (String) args[1];
#if (LOG)
            Common.Log("2 enter");
#endif
            try
            {
                lock (dgv)
                {
                    dgv.SuspendLayout();
                    dgv.ClearSelection();
                    //save the index of the first row displayed for later use
                    int firstIndex = dgv.FirstDisplayedScrollingRowIndex;
                    try
                    {
                        //this is crude, and I hate it, but it is reliable.
                        foreach (DataGridViewRow r in dgv.Rows)
                        {
                            if ((Convert.ToInt64(r.Cells["Frequency"].Value).Equals(targetFreq)) &&
                                (r.Cells["Description"].Value.Equals(desc)) && 
                                (!Convert.ToBoolean(r.Cells["Locked"].Value)))
                            {
                                r.Selected = true;
                            }
                        }
                    }
                    catch (Exception)
                    {
                        //swallow it
#if (LOG)
                        Common.Log("2 find exception: " + ex.Message);
#endif
                    }
                    if (dgv.SelectedRows.Count == 0)
                    {
                        dgv.ResumeLayout();
#if (LOG)
                        Common.Log("2 not found, exit");
#endif
                        return;
                    }
                    //reposition the grid if necessary
                    Int32 lastRowVisible = firstIndex + dgv.DisplayedRowCount(true);
                    if ((dgv.SelectedRows[0].Index > firstIndex) && (dgv.SelectedRows[0].Index < lastRowVisible))
                    {
                        dgv.ResumeLayout();
#if (LOG)
                        Common.Log("2 exit");
#endif
                        return;
                    }
                    dgv.FirstDisplayedScrollingRowIndex = dgv.SelectedRows[0].Index;
                    dgv.ResumeLayout();
                }
            }
            catch (ArgumentOutOfRangeException)
            {
                //swallow it
            }
            catch (InvalidOperationException)
            {
                //swallow it
            }
            catch (Exception)
            {
                //swallow it
            }
#if (LOG)
            Common.Log("2 exit");
#endif
        }

        /// <summary>
        /// Puts row values into a FreqRecord for later
        /// retrieval by FreqMgr form
        /// </summary>
        private void SaveRowValues()
        {
            try
            {
                _fr.Callsign = dgv.CurrentRow.Cells["Call Sign"].Value.ToString();
                _fr.Center = (long)dgv.CurrentRow.Cells["Center"].Value;
                _fr.CWShift = Convert.ToInt32(dgv.CurrentRow.Cells["CW Shift"].Value);
                _fr.Date = dgv.CurrentRow.Cells["Date"].Value.ToString();
                _fr.Description = dgv.CurrentRow.Cells["Description"].Value.ToString();
                _fr.FilterBandwidth = Convert.ToInt32(dgv.CurrentRow.Cells["Filter Bandwidth"].Value);
                _fr.FilterOrder = Convert.ToInt32(dgv.CurrentRow.Cells["Filter Order"].Value);
                _fr.FilterType = dgv.CurrentRow.Cells["Filter Type"].Value.ToString();
                _fr.Frequency = (long)dgv.CurrentRow.Cells["Frequency"].Value;
                _fr.Id = Convert.ToInt32(dgv.CurrentRow.Cells["Id"].Value);
                _fr.Mode = dgv.CurrentRow.Cells["Mode"].Value.ToString();
                _fr.Notes = dgv.CurrentRow.Cells["Notes"].Value.ToString();
                _fr.Protocol = dgv.CurrentRow.Cells["Protocol"].Value.ToString();
                _fr.Service = dgv.CurrentRow.Cells["Service"].Value.ToString();
                _fr.ShiftEnabled = Convert.ToInt32(dgv.CurrentRow.Cells["Shift Enabled"].Value);
                _fr.Shift = (long)dgv.CurrentRow.Cells["Shift"].Value;
                _fr.SquelchEnabled = Convert.ToInt32(dgv.CurrentRow.Cells["Squelch Enabled"].Value);
                _fr.Squelch = Convert.ToInt32(dgv.CurrentRow.Cells["Squelch"].Value);
                _fr.Flagged = Convert.ToBoolean(dgv.CurrentRow.Cells["Flagged"].Value);
                _fr.Locked = Convert.ToBoolean(dgv.CurrentRow.Cells["Locked"].Value);
            }
            catch (Exception ex)
            {
                MessageBox.Show(String.Format("Error in SaveRowValues: {0}\n\r{1}", ex.Message, ex.StackTrace), "Error",
                                MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
        }

        /// <summary>
        /// processes the datagridview filter settings on-the-fly
        /// </summary>
        private void FilterTextTextChanged(object sender, EventArgs e)
        {
            //set control states
            if (filterText.Text.Length > 0)
            {
                groupList.Enabled = false;
                filterFlagsList.Enabled = false;
            }
            else
            {
                groupList.Enabled = true;
                filterFlagsList.Enabled = true;
            }
            //try to convert the filter text into a number
            Decimal frq;
            Boolean result = Decimal.TryParse(filterText.Text, Common.NsThou, CultureInfo.CurrentCulture, out frq);
            try
            {
                //if it's a frequency being entered, filter on frequencies
                if (result)
                {
                    _ds.Tables[0].DefaultView.RowFilter = String.Format("FreqString LIKE '{0}*'", frq);
                }
                else
                {
                    //filter based on description or notes
                    _ds.Tables[0].DefaultView.RowFilter = String.Format("Description LIKE '*{0}*' OR Notes LIKE '*{0}*'", filterText.Text);
                }
                //select the first row shown and set display it
                dgv.Rows[0].Selected = true;
                dgv.FirstDisplayedScrollingRowIndex = 0;
                SetWindowText();
            }
            catch (Exception)
            {
                //swallow it
            }
        }

        /// <summary>
        /// utility to set button states based on current conditions
        /// </summary>
        private void SetButtonStates()
        {
#if (LOG)
            Common.Log("enter");
#endif
            if (dgv.SelectedRows.Count > 1)
            {
                sendToRadioButton.Enabled = false;
                //editButton.Enabled = false;
#if (LOG)
                Common.Log("exit");
#endif
                return;
            }
            if ((_isScanning) && (liveTrack.Checked))
            {
                sendToRadioButton.Enabled = false;
                editButton.Enabled = false;
                deleteButton.Enabled = false;
            }
            else
            {
                sendToRadioButton.Enabled = true;
                editButton.Enabled = true;
                deleteButton.Enabled = true;
            }
#if (LOG)
            Common.Log("exit");
#endif
        }

        /// <summary>
        /// reloads the grid on command
        /// </summary>
        private void RefreshButtonClick(object sender, EventArgs e)
        {
            LoadDGV(0);
        }

        /// <summary>
        /// sets the grid locked or flagged cells for a frequency 
        /// and invalidates the grid so it repaints.  By doing this 
        /// we don't need to reload the grid when changing states.
        /// </summary>
        public void SetFreqFlags(long freq, long locked, long flagged)
        {
#if (LOG)
            Common.Log("enter");
#endif
            try
            {
                DataGridViewRow row = dgv.Rows
                    .Cast<DataGridViewRow>().First(r => Convert.ToInt64(r.Cells["Frequency"].Value).Equals(freq));
                Int32 index = row.Index;
                dgv.Rows[index].Cells[1].Value = locked;
                dgv.Rows[index].Cells[2].Value = flagged;
                dgv.Invalidate();
            }
            catch (Exception)
            {
                //swallow it
            }
#if (LOG)
            Common.Log("exit");
#endif
        }

        /// <summary>
        /// Adds a new record to the grid.  This assumes
        /// it has been added to the database elsewhere.
        /// Doing this prevents needing to call a reload
        /// of the whole grid.s
        /// </summary>
        public void GridRowAdd(FreqRecord fr, Boolean updateGroups)
        {
            if (!liveTrack.Checked) return;
#if (LOG)
            Common.Log("enter");
#endif
            try
            {
                //create a row of data
                DataRow dr = _ds.Tables[0].NewRow();
                dr["Id"] = fr.Id;
                dr["Frequency"] = fr.Frequency;
                dr["Center"] = fr.Center;
                dr["Description"] = fr.Description;
                dr["Mode"] = fr.Mode;
                dr["Protocol"] = fr.Protocol;
                dr["Call Sign"] = fr.Callsign;
                dr["Service"] = fr.Service;
                dr["Shift Enabled"] = fr.ShiftEnabled;
                dr["Shift"] = fr.Shift;
                dr["Filter Type"] = fr.FilterType;
                dr["Filter Bandwidth"] = fr.FilterBandwidth;
                dr["Filter Order"] = fr.FilterOrder;
                dr["Squelch Enabled"] = fr.SquelchEnabled;
                dr["Squelch"] = fr.Squelch;
                dr["CW Shift"] = fr.CWShift;
                dr["Notes"] = fr.Notes;
                dr["Date"] = fr.Date;
                dr["Locked"] = fr.Locked;
                dr["Flagged"] = fr.Flagged;
                dr["FreqString"] = fr.Frequency.ToString(CultureInfo.InvariantCulture);
                //if groups need to be updated, get them from the db
                if (updateGroups)
                {
                    using (SQLiteConnection conn = Common.GetSqlConnection())
                    {
                        const String sel =
                            "select group_concat(g.[Name]) from Groups g join GroupFreq f on g.Id = f.GroupId " +
                            "where f.FreqId = @freqId";
                        var cmd = new SQLiteCommand(sel, conn);
                        cmd.Parameters.AddWithValue("@freqId", fr.Id);
                        dr["Groups"] = cmd.ExecuteScalar();
                    }
                }
                //add the datarow to the dataset, which causes it to display in the grid
                if (_ds != null)
                {
                    lock (_ds)
                    {
                        _ds.Tables[0].Rows.Add(dr);
                        _ds.Tables[0].AcceptChanges();
                    }
                }
                SetWindowText();
                //position the grid to the new row
                PositionGrid(fr.Frequency);
            }
            catch (Exception ex)
            {
                MessageBox.Show(String.Format("An error occurred: {0}\n\r{1}", ex.Message, ex.StackTrace), "Error",
                                MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
#if (LOG)
            Common.Log("exit");
#endif
        }

        /// <summary>
        /// sets the window caption
        /// </summary>
        private void SetWindowText()
        {
            Text = String.Format("Browse {0} Frequencies", dgv.Rows.Count); //window caption
        }

        /// <summary>
        /// Updates a record on the grid.  This assumes
        /// it has been updated in the database elsewhere.
        /// </summary>
        public void GridRowUpdate(FreqRecord fr, Boolean updateGroups)
        {
            if (!liveTrack.Checked) return;
#if (LOG)
            Common.Log("enter");
#endif
            try
            {
                //find the row containing the frequency
                dgv.ClearSelection();
                DataGridViewRow row = dgv.Rows
                    .Cast<DataGridViewRow>().First(r => Convert.ToInt64(r.Cells["Id"].Value).Equals(fr.Id));
                Int32 rowId = row.Index;
                //update the cells in this row
                dgv.Rows[rowId].Selected = true;
                dgv.Rows[rowId].Cells["Id"].Value = fr.Id;
                dgv.Rows[rowId].Cells["Frequency"].Value = fr.Frequency;
                dgv.Rows[rowId].Cells["Center"].Value = fr.Center;
                dgv.Rows[rowId].Cells["Description"].Value = fr.Description;
                dgv.Rows[rowId].Cells["Mode"].Value = fr.Mode;
                dgv.Rows[rowId].Cells["Protocol"].Value = fr.Protocol;
                dgv.Rows[rowId].Cells["Call Sign"].Value = fr.Callsign;
                dgv.Rows[rowId].Cells["Service"].Value = fr.Service;
                dgv.Rows[rowId].Cells["Shift Enabled"].Value = fr.ShiftEnabled;
                dgv.Rows[rowId].Cells["Shift"].Value = fr.Shift;
                dgv.Rows[rowId].Cells["Filter Type"].Value = fr.FilterType;
                dgv.Rows[rowId].Cells["Filter Bandwidth"].Value = fr.FilterBandwidth;
                dgv.Rows[rowId].Cells["Filter Order"].Value = fr.FilterOrder;
                dgv.Rows[rowId].Cells["Squelch Enabled"].Value = fr.SquelchEnabled;
                dgv.Rows[rowId].Cells["Squelch"].Value = fr.Squelch;
                dgv.Rows[rowId].Cells["CW Shift"].Value = fr.CWShift;
                dgv.Rows[rowId].Cells["Notes"].Value = fr.Notes;
                dgv.Rows[rowId].Cells["Date"].Value = fr.Date;
                dgv.Rows[rowId].Cells["Locked"].Value = fr.Locked;
                dgv.Rows[rowId].Cells["Flagged"].Value = fr.Flagged;
                dgv.Rows[rowId].Cells["FreqString"].Value = fr.Frequency.ToString(CultureInfo.InvariantCulture);
                //if groups need to be updated, get them from the db
                if (updateGroups)
                {
                    using (SQLiteConnection conn = Common.GetSqlConnection())
                    {
                        const String sel =
                            "select group_concat(g.[Name]) from Groups g join GroupFreq f on g.Id = f.GroupId " +
                            "where f.FreqId = @freqId";
                        var cmd = new SQLiteCommand(sel, conn);
                        cmd.Parameters.AddWithValue("@freqId", fr.Id);
                        dgv.Rows[rowId].Cells["Groups"].Value = cmd.ExecuteScalar();
                    }
                }
                //force the grid to repaint
                dgv.Invalidate();
            }
            catch (Exception ex)
            {
#if (LOG)
            Common.Log(ex.Message + "\n\r" + ex.StackTrace);
#endif
            }
#if (LOG)
            Common.Log("exit");
#endif
        }

        /// <summary>
        /// traps and ignores the grid DataError event
        /// </summary>
        private void DGVDataError(object sender, DataGridViewDataErrorEventArgs e)
        {
            //Debug.WriteLine(e.Exception.Message);
        }

        /// <summary>
        /// resizes the grid when the window size changes.
        /// </summary>
        private void BrowseSizeChanged(object sender, EventArgs e)
        {
            _isResizing = true;  //tells PositionGrid not to do its work
            dgv.Height = ClientRectangle.Height - panel1.Height;
            _isResizing = false;
        }

        #endregion

    }

    #region ColInfo class

    /// <summary>
    /// used for serializing grid column information in and out of config
    /// </summary>
    [Serializable]
    internal class ColInfo
    {
        private const String Copyright = "Copyright © Jeff Knapp 2012-2013";
        internal String Name { get; set; }
        internal Int32 Index { get; set; }
        internal Int32 Width { get; set; }
    }

    #endregion

    #region double-buffer extension class
    
    /// <summary>
    /// installs double-buffering on the datagridview to improve performance
    /// </summary>
    public static class ExtensionMethods
    {
        public static void DoubleBuffered(this DataGridView dgv, bool setting)
        {
            Type dgvType = dgv.GetType();
            PropertyInfo pi = dgvType.GetProperty("DoubleBuffered",
                                                  BindingFlags.Instance | BindingFlags.NonPublic);
            pi.SetValue(dgv, setting, null);
        }
    }

    #endregion


}