﻿//  ******************************************
//  Frequency Manager + Scanner Common Classes
//  Copyright © Jeff Knapp 2012-2013 
//
//  Contains classes and vars used across several freqmgr objects.
//  ******************************************
//  Date        Changes
//  12/12/12    globally refactor creation of connection object
//  12/13/12    improvements to GetSqlConnection.  added DateSaveFormat.
//  01/07/13    gave config keys consistent naming.
//  01/14/13    fixed incorrect name of alternate db key.
//  01/19/13    added ASC to sort in GetGroupList.
//  02/02/13    added addNoneToList parm to GetGroupList.
//  ******************************************

using System;
using System.Data;
using System.Data.SQLite;
using System.Diagnostics;
using System.Drawing;
using System.Globalization;
using System.IO;
using System.Runtime.CompilerServices;
using System.Windows.Forms;
using SDRSharp;
using SDRSharp.Radio;

namespace SDRSharp.FreqMgr
{
    internal static class Common
    {
        #region local vars

        private const String Copyright = "Copyright © Jeff Knapp 2012-2013";
        internal static String DbPath;
        internal static String DbConnString;
        internal static String DbFileName;
        internal static String NumericFormat = "N0"; //numeric format 
        internal static String DateOnlyFormat = "d"; //date format that includes only the date in local format
        internal static String DateSaveFormat = "s"; //date format that is used to save culture invariant date to the db
        internal static readonly String NoneText = Utils.GetStringSetting("freqMgrNoneText", "None");
        internal static readonly String UnidText = Utils.GetStringSetting("freqMgrUnidText", "Unidentified");
        internal static NumberStyles NsThou = NumberStyles.AllowThousands; //number format that includes thousands
        internal static NumberStyles NsThouSign = NumberStyles.AllowThousands | NumberStyles.AllowLeadingSign;//number format that includes thousands and leading sign
        //private copy of the array for use in Edit because we don't have legal access to SDR#'s   
        internal static String[] FilterTypeArray = new[]
                                                       {
                                                           "Hamming", "Blackman", "Blackman-Harris", "Hann-Poisson",
                                                           "Youssef"
                                                       };                                                               
        private const String TimeFormat = "H.mm.ss.fffffff";
        #endregion
        
        /// <summary>
        /// constructor
        /// </summary>
        static Common()
        {
            DbPath = Environment.GetFolderPath(Environment.SpecialFolder.LocalApplicationData) + "\\SDRSFM";
            DbFileName = "FreqMgr.db";
            //test for alternate path, which must include the filename
            String altPath = Utils.GetStringSetting("freqMgrAlternateDb", null);
            if (!String.IsNullOrEmpty(altPath))
            {
                DbPath = Path.GetDirectoryName(altPath);
                DbFileName = Path.GetFileName(altPath);
            }
            //compose the connectionstring
            DbConnString = String.Format("Data Source={0}\\{1};Version=3;New=True;Pooling=True;Max Pool Size=100;",
                                         DbPath, DbFileName);
        }

        /// <summary>
        /// Generate a SQLite connection object if none exists
        /// </summary>
        internal static SQLiteConnection GetSqlConnection()
        {
            SQLiteConnection conn = null;
            try
            {
                conn = new SQLiteConnection(DbConnString);
                conn.Open();
            }
            catch (Exception ex)
            {
                String msg =
                    String.Format(
                        "Unable to connect to Frequency Manager database.  Cannot continue.  The error was: \n\r{0}",
                        ex.Message + "\n\r" + ex.StackTrace.Trim());
                MessageBox.Show(msg, "Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
            return conn;
        }

        /// <summary>
        /// retrieves window settings based on the windowName
        /// </summary>
        /// <param name="windowNameKey">String matching the key of the config setting</param>
        /// <param name="sz">Outputs a System.Drawing.Size object</param>
        /// <param name="lo">Outputs a System.Drawing.Point object</param>
        internal static void GetWindowSettings(String windowNameKey, out Size sz, out Point lo)
        {
            try
            {
                String windowData = Utils.GetStringSetting(windowNameKey, "0,0|100,100");
                String[] settings = windowData.Split('|');
                //extract size
                String sizeValue = settings[0];
                var cu = new ConfigUtils();
                Int32[] data = cu.ExtractValues(sizeValue);
                sz = new Size(data[0], data[1]);
                //extract location
                String locationValue = settings[1];
                data = cu.ExtractValues(locationValue);
                lo = new Point(data[0], data[1]);
            }
            catch (Exception)
            {
                //return empty objects
                sz = new Size();
                lo = new Point();
            }
        }

        /// <summary>
        /// saves window size and location to the app config file
        /// </summary>
        /// <param name="windowNameKey">window name/key name</param>
        /// <param name="sz">System.Drawing.Size object</param>
        /// <param name="lo">System.Drawing.Point object</param>
        internal static void SaveWindowSettings(String windowNameKey, Size sz, Point lo)
        {
            var data = new String[2];
            data[0] = String.Format("{0},{1}", sz.Width, sz.Height);
            data[1] = String.Format("{0},{1}", lo.X, lo.Y);
            String setting = String.Join("|", data);
            Utils.SaveSetting(windowNameKey, setting);
        }

        /// <summary>
        /// Generates a DataView containing a sorted table of Group names
        /// </summary>
        /// <returns>DataView containing sorted table</returns>
        internal static DataView GetGroupList(Boolean addNoneToList)
        {
            DataSet dsLocal = null;
            try
            {
                const string sel = "select Id, Name from Groups order by Name";
                using (SQLiteConnection conn = GetSqlConnection())
                using (var cmd = new SQLiteCommand(sel, conn))
                using (var da = new SQLiteDataAdapter(cmd))
                {
                    dsLocal = new DataSet();
                    da.Fill(dsLocal);
                    if (addNoneToList)
                    {
                        DataRow row = dsLocal.Tables[0].NewRow();
                        row["Id"] = -1;
                        row["Name"] = NoneText;
                        dsLocal.Tables[0].Rows.InsertAt(row, 0);
                        dsLocal.Tables[0].AcceptChanges();
                    }                    
                    dsLocal.Tables[0].DefaultView.Sort = "Name ASC";
                }
            }
            catch (Exception)
            {
            }
            return dsLocal.Tables[0].DefaultView;
        }

        /// <summary>
        /// round a frequency downward to the number of decimal places. 
        /// </summary>
        /// <remarks>From http://stackoverflow.com/questions/13482159/how-to-round-up-or-down-in-c </remarks>
        internal static Double RoundDownFreq(Double number, int decimalPlaces)
        {
            return Math.Floor(number*Math.Pow(10, decimalPlaces))/Math.Pow(10, decimalPlaces);
        }

        /// <summary>
        /// round a frequency upward to the number of decimal places. 
        /// </summary>
        /// <remarks>From http://stackoverflow.com/questions/13482159/how-to-round-up-or-down-in-c </remarks>
        internal static Double RoundUpFreq(Double number, int decimalPlaces)
        {
            return Math.Ceiling(number*Math.Pow(10, decimalPlaces))/Math.Pow(10, decimalPlaces);
        }

        /// <summary>
        /// Logs the name of the calling routine to the debug window
        /// </summary>
        /// <remarks>Based on http://stackoverflow.com/questions/2652460/c-sharp-how-to-get-the-name-of-the-current-method-from-code </remarks>
        [MethodImpl(MethodImplOptions.NoInlining)]
        internal static void Log()
        {
#if  (LOG)
            // get call stack
            var stackTrace = new StackTrace();
            // get calling method name
            String name = stackTrace.GetFrame(1).GetMethod().Name;
            Debug.WriteLine(DateTime.Now.ToString(TimeFormat) + "\t" + name);
#endif
        }

        /// <summary>
        /// Logs the name of the calling routine and some text to the debug window
        /// </summary>
        /// <remarks>Based on http://stackoverflow.com/questions/2652460/c-sharp-how-to-get-the-name-of-the-current-method-from-code </remarks>
        [MethodImpl(MethodImplOptions.NoInlining)]
        internal static void Log(String text)
        {
#if  (LOG)
            // get call stack
            var stackTrace = new StackTrace();
            // get calling method name
            String name = stackTrace.GetFrame(1).GetMethod().Name;
            Debug.WriteLine(DateTime.Now.ToString(TimeFormat) + "\t" + name + ": " + text);
#endif
        }
    }

    #region configuration utils

    /// <summary>
    /// utility class for config file functions
    /// </summary>
    internal sealed class ConfigUtils
    {
        /// <summary>
        /// utility function to split a comma-separated string into an array of integers.
        /// we have it here because it can't be in a static class.
        /// </summary>
        public Int32[] ExtractValues(String dataPair)
        {
            try
            {
                String[] values = dataPair.Split(',');
                var returnData = new Int32[2];
                returnData[0] = Int32.Parse(values[0]);
                returnData[1] = Int32.Parse(values[1]);
                return returnData;
            }
            catch (Exception)
            {
                return null;
            }
        }
    }

    #endregion

    #region event argument classes

    /// <summary>
    /// event argument for station selected event
    /// </summary>
    internal sealed class StationSelectedEventArgs : EventArgs
    {
        private readonly FreqRecord _fr;

        internal StationSelectedEventArgs(FreqRecord rec)
        {
            _fr = rec;
        }

        internal FreqRecord Record
        {
            get { return _fr; }
        }
    }

    /// <summary>
    /// event argument for station edited event
    /// </summary>
    internal sealed class StationEditedEventArgs : EventArgs
    {
        private readonly FreqRecord _fr;

        internal StationEditedEventArgs(FreqRecord rec)
        {
            _fr = rec;
        }

        internal FreqRecord Record
        {
            get { return _fr; }
        }
    }

    /// <summary>
    /// event argument for GroupUpdated event
    /// </summary>
    internal sealed class GroupsUpdatedEventArgs : EventArgs
    {
        //this space intentionally left blank
    }

    #endregion
}