﻿//  ******************************************
//  Frequency Manager + Scanner Edit Dialog
//  Copyright © Jeff Knapp 2012-2013 
//
//  This displays the edit dialog.
//  It allows the user to add and edit frequency records.
//  Unlike the other freqmgr windows, this is shown
//  as a modal dialog.
//  ******************************************
//  Date        Changes
//  12/23/12    minor UI cleanup
//  01/07/13    gave config keys consistent naming
//  02/01/13    added support for bulk-editing records when called from Browse
//  ******************************************

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Data.SQLite;
using System.Drawing;
using System.Globalization;
using System.Linq;
using System.Windows.Forms;

namespace SDRSharp.FreqMgr
{
    internal sealed partial class Edit : Form
    {
        #region local vars

        private const String Copyright = "Copyright © Jeff Knapp 2012-2013";
        private const String WindowConfigKey = "freqMgrEditWindow";
        private readonly FreqRecord _fr = new FreqRecord();
        private Boolean _allowValidation;
        private readonly Int32 _numberRecordsEditing;
        private readonly long[] _recs;
        private readonly Boolean _editingMultipleRecs;
        private readonly Boolean _okToLoad = true;
        
        #endregion

        /// <summary>
        /// constructor for editing a single record
        /// </summary>
        /// <param name="rec">FreqRecord object containing the data to edit</param>
        /// <param name="filterTypes">array of possible filter types</param>
        /// <param name="filterTypeSelectedIndex">the current filtertype</param>
        /// <param name="currMode">the current mode</param>
        public Edit(FreqRecord rec, IEnumerable<string> filterTypes, int filterTypeSelectedIndex, String currMode)
        {
            try
            {
                InitializeComponent();
                Cursor = Cursors.WaitCursor;
                Text = "Edit a Frequency";
                //init dropdown contents
                InitDropdowns(rec.Id, filterTypes, filterTypeSelectedIndex);
                //set mode
                mode.SelectedItem = currMode;
                //save incoming data to private record
                RecordData = rec;
                _numberRecordsEditing = 1;

                #region set UI control values

                frequency.Text = RecordData.Frequency.ToString(Common.NumericFormat);
                center.Text = RecordData.Center.ToString(Common.NumericFormat);
                description.Text = RecordData.Description;
                mode.Text = RecordData.Mode;
                protocol.Text = RecordData.Protocol.Length > 0 ? RecordData.Protocol : Common.UnidText;
                callsign.Text = RecordData.Callsign;
                service.Text = RecordData.Service.Length > 0 ? RecordData.Service : Common.UnidText;
                shiftEnabled.Checked = Convert.ToBoolean(RecordData.ShiftEnabled);
                shift.Text = RecordData.Shift.ToString(Common.NumericFormat);
                shift.Enabled = shiftEnabled.Checked;
                filterType.Text = RecordData.FilterType;
                filterBandwidth.Text = RecordData.FilterBandwidth.ToString(Common.NumericFormat);
                filterOrder.Text = RecordData.FilterOrder.ToString(Common.NumericFormat);
                squelchEnabled.Checked = Convert.ToBoolean(RecordData.SquelchEnabled);
                squelch.Text = RecordData.Squelch.ToString(Common.NumericFormat);
                squelch.Enabled = squelchEnabled.Checked;
                cwShift.Text = RecordData.CWShift.ToString(Common.NumericFormat);
                notes.Text = RecordData.Notes;
                lockedCheckBox.Checked = RecordData.Locked;
                flaggedCheckBox.Checked = RecordData.Flagged;
                updateButton.Enabled = (RecordData.Id > 0);

                #endregion

                #region connect event handlers

                FormClosing += EditFormClosing;
                Shown += EditShown;
                frequency.Validated += FrequencyValidated;
                frequency.Validating += FrequencyValidating;
                center.Validated += CenterValidated;
                center.Validating += CenterValidating;
                shift.Validated += ShiftValidated;
                shift.Validating += ShiftValidating;
                filterBandwidth.Validated += FilterBandwidthValidated;
                filterBandwidth.Validating += FilterBandwidthValidating;
                filterOrder.Validated += FilterOrderValidated;
                filterOrder.Validating += FilterOrderValidating;
                cwShift.Validated += CWShiftValidated;
                cwShift.Validating += CWShiftValidating;
                squelch.Validating += SquelchValidating;
                squelch.Validated += SquelchValidated;
                squelchEnabled.Click += SquelchEnabledClick;
                shiftEnabled.Click += ShiftEnabledClick;

                #endregion

                #region format window

                Size sz;
                Point lo;
                Common.GetWindowSettings(WindowConfigKey, out sz, out lo);
                //get window size
                if (!sz.IsEmpty)
                {
                    Size = sz;
                }
                //get window position
                if (!lo.IsEmpty)
                {
                    Location = lo;
                }
                #endregion
                Cursor = Cursors.Default;
            }
            catch (Exception ex)
            {
                Cursor = Cursors.Default;
                _okToLoad = false;
                Location = new Point(-10000, -10000); //hide the window because it failed to load correctly
                String msg = String.Format("An error occurred and the Edit dialog cannot be loaded.  The error was: \n\r{0}",
                           ex.Message + "\n\r" + ex.StackTrace.Trim());
                MessageBox.Show(msg, "Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
        }

        /// <summary>
        /// constructor for editing multiple records
        /// </summary>
        /// <param name="rec">Array of long database record Ids</param>
        /// <param name="filterTypes">array of possible filter types</param>
        public Edit(long[] rec, IEnumerable<string> filterTypes)
        {
            try
            {
                InitializeComponent();
                Cursor = Cursors.WaitCursor;
                Load += EditLoad;
                Text = "Edit Multiple Frequencies";
                _recs = rec;
                _editingMultipleRecs = true;
                _numberRecordsEditing = rec.Length;

                //init dropdown contents
                InitDropdowns(-1, filterTypes, 0);

                #region set UI control values
                //init controls
                updateButton.Enabled = true;
                addButton.Enabled = false;

                //rig the checkboxes to support 3 states
                squelchEnabled.ThreeState = true;
                squelchEnabled.CheckState = CheckState.Indeterminate;
                shiftEnabled.ThreeState = true;
                shiftEnabled.CheckState = CheckState.Indeterminate;
                lockedCheckBox.ThreeState = true;
                lockedCheckBox.CheckState = CheckState.Indeterminate;
                flaggedCheckBox.ThreeState = true;
                flaggedCheckBox.CheckState = CheckState.Indeterminate;
                //at this point all controls are at default states for a mix of records
                #endregion

                #region survey the records to determine how to set some of the controls
                int nbrLocked = 0;
                int nbrFlagged = 0;
                int nbrSquelchEnabled = 0;
                int nbrShiftEnabled = 0;
                List<long> frequencies = new List<long>();
                List<long> centers = new List<long>();
                List<String> descriptions = new List<String>();
                List<String> modes = new List<String>();
                List<String> protocols = new List<String>();
                List<String> services = new List<String>();
                List<String> filters = new List<String>();
                List<String> filterBws = new List<String>();
                List<String> cwShifts = new List<String>();
                List<String> shifts = new List<String>();
                List<String> filterOrders = new List<String>();
                List<String> squelches = new List<String>();
                const string sel = "select Frequency, Center, Description, [Squelch Enabled], [Shift Enabled], Locked, Flagged, Mode, " +
                                   "Protocol, Service, [Filter Type], [Filter Bandwidth], [CW Shift], " +
                                   "Shift, [Filter Order], Squelch from data where id = @id";
                using (SQLiteConnection conn = Common.GetSqlConnection())
                {
                    var cmd = new SQLiteCommand(sel, conn);
                    for (int i = 0; i < _numberRecordsEditing; i++)
                    {
                        cmd.Parameters.Clear();
                        cmd.Parameters.AddWithValue("@id", _recs[i]);
                        DataSet ds = new DataSet();
                        using (var da = new SQLiteDataAdapter(cmd))
                        {
                            da.Fill(ds);
                        }
                        DataRow dr = ds.Tables[0].Rows[0];
                        if (Convert.ToBoolean(dr["Squelch Enabled"])) nbrSquelchEnabled += 1;
                        if (Convert.ToBoolean(dr["Shift Enabled"])) nbrShiftEnabled += 1;
                        if (Convert.ToBoolean(dr["Locked"])) nbrLocked += 1;
                        if (Convert.ToBoolean(dr["Flagged"])) nbrFlagged += 1;
                        if (!frequencies.Contains((long)dr["Frequency"])) frequencies.Add((long)dr["Frequency"]);
                        if (!centers.Contains((long)dr["Center"])) centers.Add((long)dr["Center"]);
                        if (!descriptions.Contains(dr["Description"].ToString())) descriptions.Add(dr["Description"].ToString());
                        if (!modes.Contains(dr["Mode"].ToString())) modes.Add(dr["Mode"].ToString());
                        if (!protocols.Contains(dr["Protocol"].ToString())) protocols.Add(dr["Protocol"].ToString());
                        if (!services.Contains(dr["Service"].ToString())) services.Add(dr["Service"].ToString());
                        if (!filters.Contains(dr["Filter Type"].ToString())) filters.Add(dr["Filter Type"].ToString());
                        if (!filterBws.Contains(dr["Filter Bandwidth"].ToString())) filterBws.Add(dr["Filter Bandwidth"].ToString());
                        if (!cwShifts.Contains(dr["CW Shift"].ToString())) cwShifts.Add(dr["CW Shift"].ToString());
                        if (!shifts.Contains(dr["Shift"].ToString())) shifts.Add(dr["Shift"].ToString());
                        if (!filterOrders.Contains(dr["Filter Order"].ToString())) filterOrders.Add(dr["Filter Order"].ToString());
                        if (!squelches.Contains(dr["Squelch"].ToString())) squelches.Add(dr["Squelch"].ToString());
                    }
                }
                #endregion

                #region set the checkbox states according to the survey results

                if (nbrFlagged == 0)
                {
                    flaggedCheckBox.CheckState = CheckState.Unchecked;
                }
                if (nbrFlagged == _numberRecordsEditing)
                {
                    flaggedCheckBox.CheckState = CheckState.Checked;
                }

                if (nbrLocked == 0)
                {
                    lockedCheckBox.CheckState = CheckState.Unchecked;
                }
                if (nbrLocked == _numberRecordsEditing)
                {
                    lockedCheckBox.CheckState = CheckState.Checked;
                }

                if (nbrSquelchEnabled == 0)
                {
                    squelchEnabled.CheckState = CheckState.Unchecked;
                }
                if (nbrSquelchEnabled == _numberRecordsEditing)
                {
                    squelchEnabled.CheckState = CheckState.Checked;
                }

                if (nbrShiftEnabled == 0)
                {
                    shiftEnabled.CheckState = CheckState.Unchecked;
                }
                if (nbrShiftEnabled == _numberRecordsEditing)
                {
                    shiftEnabled.CheckState = CheckState.Checked;
                }
                squelch.Enabled = squelchEnabled.Checked;
                shift.Enabled = shiftEnabled.Checked;
                #endregion

                #region set the dropdowns according to the survey results

                if (modes.Count == 1) mode.SelectedIndex = mode.FindStringExact(modes[0]);
                if (protocols.Count == 1) protocol.SelectedIndex = protocol.FindStringExact(protocols[0]);
                if (services.Count == 1) service.SelectedIndex = service.FindStringExact(services[0]);
                if (filters.Count == 1) filterType.SelectedIndex = filterType.FindStringExact(filters[0]);

                #endregion

                #region set textboxes according to the survey results
                if (frequencies.Count == 1) frequency.Text = frequencies[0].ToString(Common.NumericFormat);
                if (centers.Count == 1) center.Text = centers[0].ToString(Common.NumericFormat);
                if (descriptions.Count == 1) description.Text = descriptions[0];
                if (filterBws.Count == 1) filterBandwidth.Text = filterBws[0];
                if (cwShifts.Count == 1) cwShift.Text = cwShifts[0];
                if (shifts.Count == 1) shift.Text = shifts[0];
                if (filterOrders.Count == 1) filterOrder.Text = filterOrders[0];
                if (squelches.Count == 1) squelch.Text = squelches[0];

                #endregion

                #region connect event handlers

                FormClosing += EditFormClosing;
                Shown += EditShown;
                shift.Validated += ShiftValidated;
                shift.Validating += ShiftValidating;
                filterBandwidth.Validated += FilterBandwidthValidated;
                filterBandwidth.Validating += FilterBandwidthValidating;
                filterOrder.Validated += FilterOrderValidated;
                filterOrder.Validating += FilterOrderValidating;
                cwShift.Validated += CWShiftValidated;
                cwShift.Validating += CWShiftValidating;
                squelch.Validating += SquelchValidating;
                squelch.Validated += SquelchValidated;
                squelchEnabled.Click += SquelchEnabledClick;
                shiftEnabled.Click += ShiftEnabledClick;

                #endregion

                #region format window

                Size sz;
                Point lo;
                Common.GetWindowSettings(WindowConfigKey, out sz, out lo);
                //get window size
                if (!sz.IsEmpty)
                {
                    Size = sz;
                }
                //get window position
                if (!lo.IsEmpty)
                {
                    Location = lo;
                }

                #endregion

                _editingMultipleRecs = false;
                Cursor = Cursors.Default;
            }
            catch (Exception ex)
            {
                Cursor = Cursors.Default;
                _okToLoad = false;
                Location = new Point(-10000, -10000); //hide the window because it failed to load correctly
                String msg = String.Format("An error occurred and the Edit dialog cannot be loaded.  The error was: \n\r{0}",
                           ex.Message + "\n\r" + ex.StackTrace.Trim());
                MessageBox.Show(msg, "Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
        }

        /// <summary>
        /// handles the load event
        /// </summary>
        /// <remarks>closes the window if it is not ok to finish loading</remarks>
        private void EditLoad(object sender, EventArgs e)
        {
            if (!_okToLoad)
            {
                Close();
            }
        }

        #region utility members

        /// <summary>
        /// initializes the dropdowns and their contents
        /// </summary>
        /// <param name="recId">Record Id for single record, -1 for multiple records</param>
        /// <param name="filterTypes">List of filter types</param>
        /// <param name="filterTypeSelectedIndex">Selected filter index for single record, any integer for multiple records</param>
        private void InitDropdowns(int recId, IEnumerable<string> filterTypes, int filterTypeSelectedIndex)
        {
            InitMode();
            InitProtocol();
            InitService();
            InitGroups(recId);
            //we do this stupid loop to get around the covariance issue
            //when converting a string array to an object array as
            //would happen if we used filterType.Items.AddRange(filterTypes)
            foreach (string s in filterTypes)
            {
                filterType.Items.Add(s);
            }
            if (recId >= 0) filterType.SelectedIndex = filterTypeSelectedIndex;

        }

        /// <summary>
        /// Set and Get private data
        /// </summary>
        public FreqRecord RecordData
        {
            get { return _fr; }
            set
            {
                //store incoming values to private variable
                _fr.Callsign = value.Callsign;
                _fr.Center = value.Center;
                _fr.CWShift = value.CWShift;
                _fr.Date = String.IsNullOrEmpty(_fr.Date)
                               ? DateTime.Now.ToString(Common.DateSaveFormat, CultureInfo.InvariantCulture)
                               : value.Date;
                _fr.Description = value.Description;
                _fr.FilterBandwidth = value.FilterBandwidth;
                _fr.FilterOrder = value.FilterOrder;
                _fr.FilterType = value.FilterType;
                _fr.Flagged = value.Flagged;
                _fr.Frequency = value.Frequency;
                _fr.Id = value.Id;
                _fr.Locked = value.Locked;
                _fr.Mode = value.Mode;
                _fr.Notes = value.Notes;
                _fr.Protocol = value.Protocol.Length > 0 ? value.Protocol : Common.UnidText;
                _fr.Service = value.Service.Length > 0 ? value.Service : Common.UnidText;
                _fr.ShiftEnabled = value.ShiftEnabled;
                _fr.Shift = value.Shift;
                _fr.SquelchEnabled = value.SquelchEnabled;
                _fr.Squelch = value.Squelch;
                _fr.Flagged = value.Flagged;
                _fr.Locked = value.Locked;
            }
        }

        /// <summary>
        /// public property indicating whether the edited freq was updated
        /// </summary>
        public bool FreqUpdated { get; private set; }

        /// <summary>
        /// initializes the group list.  this reads the groups from
        /// the db and then checkmarks the ones that the frequency
        /// is assigned to.
        /// </summary>
        private void InitGroups(Int32 freqId)
        {
            try
            {
                //load the groups list
                using (SQLiteConnection conn = Common.GetSqlConnection())
                {
                    String sel = "select Id, Name from Groups order by Name";
                    var cmd = new SQLiteCommand(sel, conn);
                    var da = new SQLiteDataAdapter(cmd);
                    var ds = new DataSet();
                    da.Fill(ds);
                    foreach (ListViewItem li in from DataRow item in ds.Tables[0].Rows select new ListViewItem { Text = item["Name"].ToString(), Tag = item["Id"].ToString() })
                    {
                        groupList.Items.Add(li);
                    }

                    //set checkmarks for the items to which the freq is assigned
                    if (freqId <= 0) return;
                    sel = "select groups.name from groupfreq join groups on groups.id = groupfreq.groupid where groupfreq.freqid = @freqId";
                    cmd = new SQLiteCommand(sel, conn);
                    cmd.Parameters.AddWithValue("@freqId", freqId);
                    da = new SQLiteDataAdapter(cmd);
                    ds = new DataSet();
                    da.Fill(ds);
                    foreach (ListViewItem li in from DataRow item in ds.Tables[0].Rows select groupList.FindItemWithText(item["Name"].ToString()))
                    {
                        li.Checked = true;
                    }
                }
            }
            catch (Exception ex)
            {
                String msg = String.Format("An error occurred.  The error was: \n\r{0}",
                                           ex.Message + "\n\r" + ex.StackTrace.Trim());
                MessageBox.Show(msg, "Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
        }

        /// <summary>
        /// Initialize the values in the Mode dropdown because
        /// we don't have legal access to the ones in SDR#
        /// </summary>
        private void InitMode()
        {
            mode.Items.Add("AM");
            mode.Items.Add("CW-L");
            mode.Items.Add("CW-U");
            mode.Items.Add("DSB");
            mode.Items.Add("LSB");
            mode.Items.Add("NFM");
            mode.Items.Add("USB");
            mode.Items.Add("WFM");
        }

        /// <summary>
        /// Initialize the values in the Protocol combobox
        /// </summary>
        private void InitProtocol()
        {
            try
            {
                using (SQLiteConnection conn = Common.GetSqlConnection())
                {
                    const string sel = "select Name from Protocol order by Name";
                    var cmd = new SQLiteCommand(sel, conn);
                    var da = new SQLiteDataAdapter(cmd);
                    var ds = new DataSet();
                    da.Fill(ds);
                    protocol.DataSource = ds.Tables[0];
                    protocol.DisplayMember = "Name";
                    if (_editingMultipleRecs)
                    {
                        protocol.SelectedIndex = -1;
                    }
                }
            }
            catch (Exception ex)
            {
                String msg = String.Format("An error occurred.  The error was: \n\r{0}",
                                           ex.Message + "\n\r" + ex.StackTrace.Trim());
                MessageBox.Show(msg, "Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
        }

        /// <summary>
        /// Initialize the values in the Service combobox
        /// </summary>
        private void InitService()
        {
            try
            {
                using (SQLiteConnection conn = Common.GetSqlConnection())
                {
                    const string sel = "select Name from Service order by Name";
                    var cmd = new SQLiteCommand(sel, conn);
                    var da = new SQLiteDataAdapter(cmd);
                    var ds = new DataSet();
                    da.Fill(ds);
                    service.DataSource = ds.Tables[0];
                    service.DisplayMember = "Name";
                    if (_editingMultipleRecs)
                    {
                        service.SelectedIndex = -1;
                    }
                }
            }
            catch (Exception ex)
            {
                String msg = String.Format("An error occurred.  The error was: \n\r{0}",
                                           ex.Message + "\n\r" + ex.StackTrace.Trim());
                MessageBox.Show(msg, "Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
        }

        /// <summary>
        /// utility to build command parms, used by
        /// both add and update methods
        /// </summary>
        /// <param name="cmdObject">SQLiteCommand cmd object</param>
        private void SetSqlCmdParameters(SQLiteCommand cmdObject)
        {
            cmdObject.Parameters.AddWithValue("@callsign", RecordData.Callsign);
            cmdObject.Parameters.AddWithValue("@center", RecordData.Center);
            cmdObject.Parameters.AddWithValue("@cwshift", RecordData.CWShift);
            cmdObject.Parameters.AddWithValue("@date", RecordData.Date);
            cmdObject.Parameters.AddWithValue("@description", RecordData.Description);
            cmdObject.Parameters.AddWithValue("@filterbandwidth", RecordData.FilterBandwidth);
            cmdObject.Parameters.AddWithValue("@filterorder", RecordData.FilterOrder);
            cmdObject.Parameters.AddWithValue("@filtertype", RecordData.FilterType);
            cmdObject.Parameters.AddWithValue("@frequency", RecordData.Frequency);
            cmdObject.Parameters.AddWithValue("@mode", RecordData.Mode);
            cmdObject.Parameters.AddWithValue("@notes", RecordData.Notes);
            cmdObject.Parameters.AddWithValue("@protocol", RecordData.Protocol);
            cmdObject.Parameters.AddWithValue("@service", RecordData.Service);
            cmdObject.Parameters.AddWithValue("@shiftenabled", RecordData.ShiftEnabled);
            cmdObject.Parameters.AddWithValue("@shift", RecordData.Shift);
            cmdObject.Parameters.AddWithValue("@squelchenabled", RecordData.SquelchEnabled);
            cmdObject.Parameters.AddWithValue("@squelch", RecordData.Squelch);
            cmdObject.Parameters.AddWithValue("@locked", RecordData.Locked);
            cmdObject.Parameters.AddWithValue("@flagged", RecordData.Flagged);
        }

        /// <summary>
        /// updates private data from the dialog fields
        /// </summary>
        private void UpdateRecordData()
        {
            RecordData.Center = long.Parse(center.Text, Common.NsThou);
            RecordData.Frequency = long.Parse(frequency.Text, Common.NsThou);
            RecordData.Callsign = callsign.Text;
            RecordData.CWShift = Int32.Parse(cwShift.Text, Common.NsThou);
            RecordData.Date = DateTime.Today.ToString(Common.DateSaveFormat, CultureInfo.InvariantCulture);
            RecordData.Description = description.Text;
            RecordData.FilterBandwidth = int.Parse(filterBandwidth.Text, Common.NsThou);
            RecordData.FilterOrder = int.Parse(filterOrder.Text, Common.NsThou);
            RecordData.FilterType = filterType.Text;
            RecordData.Mode = mode.Text;
            RecordData.Notes = notes.Text;
            RecordData.Protocol = protocol.Text;
            RecordData.Service = service.Text;
            RecordData.ShiftEnabled = Convert.ToInt32(shiftEnabled.Checked);
            RecordData.Shift = long.Parse(shift.Text, NumberStyles.AllowThousands | NumberStyles.AllowLeadingSign);
            RecordData.SquelchEnabled = Convert.ToInt32(squelchEnabled.Checked);
            RecordData.Squelch = Int32.Parse(squelch.Text);
            RecordData.Flagged = flaggedCheckBox.Checked;
            RecordData.Locked = lockedCheckBox.Checked;
        }

        /// <summary>
        /// performs data validations
        /// </summary>
        /// <param name="sender">Control being validated</param>
        /// <param name="min">minimum allowed value</param>
        /// <param name="max">maximum allowed value</param>
        /// <returns>True if data is good, False if not</returns>
        /// <remarks>ctl.Tag contains the display name of the control,
        /// "Frequency", "CW Shift", etc.  these are assigned at design time.</remarks>
        private Boolean ValidateEntry(object sender, Decimal min, Decimal max)
        {
            if (_allowValidation)
            {
                var ctl = (Control)sender;
                string errorMsg = String.Format("{0} must be between {1} and {2}", ctl.Tag, min, max);
                Decimal val;
                Boolean result = Decimal.TryParse(ctl.Text, Common.NsThouSign, CultureInfo.CurrentCulture, out val);
                if (result)
                {
                    if ((val < min) || (val > max))
                    {
                        errorProvider1.SetError(ctl, errorMsg);
                        errorProvider1.SetIconAlignment(ctl, ErrorIconAlignment.MiddleRight);
                        return false;
                    }
                }
                else
                {
                    errorProvider1.SetError(ctl, errorMsg);
                    errorProvider1.SetIconAlignment(ctl, ErrorIconAlignment.MiddleRight);
                    return false;
                }
            }
            return true;
        }

        /// <summary>
        /// formats a number on-the-fly
        /// </summary>
        private void FormatTextBoxDecimal(object sender, EventArgs e)
        {
            if (_numberRecordsEditing > 1) return;
            
            var tb = (TextBox)sender;
            Decimal val;
            Boolean result = Decimal.TryParse(tb.Text, Common.NsThouSign, CultureInfo.CurrentCulture, out val);
            tb.Text = result ? Convert.ToDecimal(tb.Text).ToString(Common.NumericFormat) : "0";
        }

        /// <summary>
        /// forces control's Text to uppercase
        /// </summary>
        private void FormatTextBoxUpperCase(object sender, EventArgs e)
        {
            var tb = (TextBox)sender;
            tb.Text = tb.Text.ToUpper();
        }

        #endregion

        #region control events

        /// <summary>
        /// cancel button click handler
        /// </summary>
        private void CancelButtonClick(object sender, EventArgs e)
        {
            _allowValidation = false;
            DialogResult = DialogResult.Cancel;
        }

        /// <summary>
        /// handles change in "shift enabled" checkbox
        /// </summary>
        private void ShiftEnabledClick(object sender, EventArgs e)
        {
            shift.Enabled = shiftEnabled.Checked;
        }

        /// <summary>
        /// handles change in "squelch enabled" checkbox
        /// </summary>
        private void SquelchEnabledClick(object sender, EventArgs e)
        {
            squelch.Enabled = squelchEnabled.Checked;
        }

        /// <summary>
        /// handles click of Add button
        /// </summary>
        /// <remarks>note it is possible to create multiple database records having the same 
        /// frequency; this is intentional and is common in HF monitoring.</remarks>
        private void AddButtonClick(object sender, EventArgs e)
        {
            try
            {
                //store current data in internal record
                UpdateRecordData();
                //build insert statement
                using (SQLiteConnection conn = Common.GetSqlConnection())
                {
                    const string ins =
                        "insert into Data ([Call Sign],Center,[CW Shift],Date,Description,[Filter Bandwidth]," +
                        "[Filter Order],[Filter Type],Frequency,Mode,Notes,Protocol,Service," +
                        "[Shift Enabled],Shift,[Squelch Enabled],Squelch,Locked,Flagged)" +
                        " values (@callsign,@center,@cwshift,@date,@description,@filterbandwidth," +
                        "@filterorder,@filtertype,@frequency,@mode,@notes,@protocol,@service,@shiftenabled," +
                        "@shift,@squelchenabled,@squelch,@locked,@flagged)";
                    var cmd = new SQLiteCommand(ins, conn);
                    SetSqlCmdParameters(cmd);
                    //insert record into db
                    cmd.ExecuteNonQuery();
                    //get rowid of the record just inserted.
                    //if one is found, the insert was successful, otherwise the insert failed
                    long rowId = 0;
                    const string getrow = "select last_insert_rowid()";
                    cmd = new SQLiteCommand(getrow, conn);
                    using (SQLiteDataReader reader = cmd.ExecuteReader())
                    {
                        while (reader.Read())
                        {
                            rowId = (long)reader[0];
                        }
                    }
                    if (rowId > 0)
                    {
                        //it worked; update the internal data and close the window
                        RecordData.Id = (Int32)rowId;
                        SaveFreqGroups(RecordData.Id);
                        DialogResult = DialogResult.OK;
                        Close();
                    }
                    else
                    {
                        //insert failed
                        DialogResult = DialogResult.No;
                        Close();
                    }
                }
            }
            catch (Exception ex)
            {
                String msg = String.Format("Record could not be saved due to an error.  The error was: \n\r{0}",
                                           ex.Message + "\n\r" + ex.StackTrace.Trim());
                throw new Exception(msg);
            }
        }

        /// <summary>
        /// handles click of the Update button
        /// </summary>
        private void UpdateButtonClick(object sender, EventArgs e)
        {
            try
            {
                if (_numberRecordsEditing == 1)
                {
                    #region update 1 record
                    //store current data in internal record
                    UpdateRecordData();
                    using (SQLiteConnection conn = Common.GetSqlConnection())
                    {
                        //create update statement
                        const string upd = "update data set [Call Sign]=@callsign,Center=@center,[CW Shift]=@cwshift," +
                                           "Date=@date,Description=@description,[Filter Bandwidth]=@filterbandwidth," +
                                           "[Filter Order]=@filterorder,[Filter Type]=@filtertype,Frequency=@frequency," +
                                           "Mode=@mode,Notes=@notes,Protocol=@protocol,Service=@service," +
                                           "[Shift Enabled]=@shiftenabled,Shift=@shift,[Squelch Enabled]=@squelchenabled," +
                                           "Squelch=@squelch,Locked=@locked,Flagged=@flagged where rowid = @id";
                        var cmd = new SQLiteCommand(upd, conn);
                        SetSqlCmdParameters(cmd);
                        cmd.Parameters.AddWithValue("@id", RecordData.Id);
                        //update the record
                        cmd.ExecuteNonQuery();
                        //delete existing groups
                        const string del = "delete from GroupFreq where FreqId = @freqId";
                        cmd = new SQLiteCommand(del, conn);
                        cmd.Parameters.AddWithValue("@freqId", RecordData.Id);
                        cmd.ExecuteNonQuery();
                    }
                    //add any selected groups
                    SaveFreqGroups(RecordData.Id);

                    #endregion
                }
                else
                {
                    String msg = String.Format("UPDATE  {0}  FREQUENCIES\n\r\n\rAre you sure you want to update these Frequencies?", _numberRecordsEditing);
                    DialogResult result = MessageBox.Show(msg, "Update Frequencies", MessageBoxButtons.OKCancel,
                                             MessageBoxIcon.Warning);
                    if (result != DialogResult.OK) return;

                    #region update multiple records
                    Cursor = Cursors.WaitCursor;
                    using (SQLiteConnection conn = Common.GetSqlConnection())
                    {
                        //create sql statements
                        const string upd = "update data set [Call Sign]=@callsign,Center=@center,[CW Shift]=@cwshift," +
                                           "Date=@date,Description=@description,[Filter Bandwidth]=@filterbandwidth," +
                                           "[Filter Order]=@filterorder,[Filter Type]=@filtertype,Frequency=@frequency," +
                                           "Mode=@mode,Notes=@notes,Protocol=@protocol,Service=@service," +
                                           "[Shift Enabled]=@shiftenabled,Shift=@shift,[Squelch Enabled]=@squelchenabled," +
                                           "Squelch=@squelch,Locked=@locked,Flagged=@flagged where rowid = @id";
                        const string sel = "select * from data where id = @id";

                        for (int i = 0; i < _recs.Length; i++)
                        {
                            var cmd = new SQLiteCommand(sel, conn);
                            cmd.Parameters.AddWithValue("@id", _recs[i]);
                            DataSet ds = new DataSet();
                            using (var da = new SQLiteDataAdapter(cmd))
                            {
                                da.Fill(ds);
                            }
                            DataRow dr = ds.Tables[0].Rows[0];
                            //copy the existing data to our internal record                            
                            RecordData.Frequency = (long)dr["Frequency"];
                            RecordData.Center = (long)dr["Center"];
                            RecordData.Description = dr["Description"].ToString();
                            RecordData.Mode = dr["Mode"].ToString();
                            RecordData.Protocol = dr["Protocol"].ToString();
                            RecordData.Callsign = dr["Call Sign"].ToString();
                            RecordData.Service = dr["Service"].ToString();
                            RecordData.ShiftEnabled = Convert.ToInt32(dr["Shift Enabled"]);
                            RecordData.Shift = (long)dr["Shift"];
                            RecordData.FilterType = dr["Filter Type"].ToString();
                            RecordData.FilterBandwidth = Convert.ToInt32(dr["Filter Bandwidth"]);
                            RecordData.FilterOrder = Convert.ToInt32(dr["Filter Order"]);
                            RecordData.SquelchEnabled = Convert.ToInt32(dr["Squelch Enabled"]);
                            RecordData.Squelch = Convert.ToInt32(dr["Squelch"]);
                            RecordData.CWShift = Convert.ToInt32(dr["CW Shift"]);
                            RecordData.Notes = dr["Notes"].ToString();
                            RecordData.Date = DateTime.Now.ToString(Common.DateSaveFormat, CultureInfo.InvariantCulture);
                            RecordData.Locked = Convert.ToBoolean(dr["Locked"]);
                            RecordData.Flagged = Convert.ToBoolean(dr["Flagged"]);
                            //update the internal record from the form                            
                            if (!String.IsNullOrEmpty(description.Text)) RecordData.Description = description.Text;
                            if (!String.IsNullOrEmpty(mode.Text)) RecordData.Mode = mode.Text;
                            if (!String.IsNullOrEmpty(protocol.Text)) RecordData.Protocol = protocol.Text;
                            if (!String.IsNullOrEmpty(service.Text)) RecordData.Service = service.Text;
                            if (shiftEnabled.CheckState != CheckState.Indeterminate)
                            {
                                RecordData.ShiftEnabled = Convert.ToInt32(shiftEnabled.Checked);
                            }
                            if (!String.IsNullOrEmpty(shift.Text))
                                RecordData.Shift = long.Parse(shift.Text,
                                                              NumberStyles.AllowThousands |
                                                              NumberStyles.AllowLeadingSign);
                            if (!String.IsNullOrEmpty(filterType.Text)) RecordData.FilterType = filterType.Text;
                            if (!String.IsNullOrEmpty(filterBandwidth.Text))
                                RecordData.FilterBandwidth = int.Parse(filterBandwidth.Text, Common.NsThou);
                            if (!String.IsNullOrEmpty(cwShift.Text))
                                RecordData.CWShift = Int32.Parse(cwShift.Text, Common.NsThou);
                            if (!String.IsNullOrEmpty(filterOrder.Text))
                                RecordData.FilterOrder = int.Parse(filterOrder.Text, Common.NsThou);
                            if (lockedCheckBox.CheckState != CheckState.Indeterminate)
                            {
                                RecordData.Locked = lockedCheckBox.Checked;
                            }
                            if (!String.IsNullOrEmpty(squelch.Text)) RecordData.Squelch = Int32.Parse(squelch.Text);
                            if (squelchEnabled.CheckState != CheckState.Indeterminate)
                            {
                                RecordData.SquelchEnabled = Convert.ToInt32(squelchEnabled.Checked);
                            }
                            if (!String.IsNullOrEmpty(notes.Text)) RecordData.Notes = notes.Text;
                            if (flaggedCheckBox.CheckState != CheckState.Indeterminate)
                            {
                                RecordData.Flagged = flaggedCheckBox.Checked;
                            }
                            //update the data table
                            cmd = new SQLiteCommand(upd, conn);
                            SetSqlCmdParameters(cmd);
                            cmd.Parameters.AddWithValue("@id", _recs[i]);
                            cmd.ExecuteNonQuery();
                            //add any selected groups
                            SaveFreqGroups(_recs[i]);
                        }
                    }
                    Cursor = Cursors.Default;
                    #endregion

                }
                FreqUpdated = true;
                //return OK and close the window
                DialogResult = DialogResult.OK;
                Close();
            }
            catch (Exception ex)
            {
                String msg = String.Format("Frequency could not be saved due to an error.  The error was: \n\r{0}",
                                           ex.Message + "\n\r" + ex.StackTrace.Trim());
                throw new Exception(msg);
            }
        }

        /// <summary>
        /// save the group settings for the freq
        /// </summary>
        private void SaveFreqGroups(long freqId)
        {
            //this is only used when adding or editing groups for a frequency
            ListView.CheckedListViewItemCollection checkedItems = groupList.CheckedItems;
            if (checkedItems.Count <= 0) return;
            using (SQLiteConnection conn = Common.GetSqlConnection())
            using (SQLiteTransaction sqlTransaction = conn.BeginTransaction())
            {
                const string ins = "insert into GroupFreq (GroupId,FreqId) values (@groupId,@freqId)";
                foreach (ListViewItem item in checkedItems)
                {
                    var cmd = new SQLiteCommand(ins, conn);
                    cmd.Parameters.AddWithValue("@groupId", Convert.ToInt32(item.Tag));
                    cmd.Parameters.AddWithValue("@freqId", freqId);
                    cmd.ExecuteNonQuery();
                }
                sqlTransaction.Commit();
            }
        }

        /// <summary>
        /// positions the cursor to the description field when the form is shown.
        /// the frequency and some other fields will already contain data by this time.
        /// </summary>
        private void EditShown(object sender, EventArgs e)
        {
            description.Focus();
            _allowValidation = true;
        }

        /// <summary>
        /// cleanup on form closing
        /// </summary>
        private void EditFormClosing(object sender, FormClosingEventArgs e)
        {
            _allowValidation = false;
            Common.SaveWindowSettings(WindowConfigKey, Size, Location);
        }

        /// <summary>
        /// Flips locked on and flagged off, or vice versa.  Only one can be checked.
        /// </summary>
        private void LockedCheckBoxCheckStateChanged(object sender, EventArgs e)
        {
            if (_editingMultipleRecs) return;
            if (!lockedCheckBox.Checked) return;
            flaggedCheckBox.Checked = false;
        }

        /// <summary>
        /// Flips locked on and flagged off, or vice versa.  Only one can be checked.
        /// </summary>
        private void FlaggedCheckBoxCheckStateChanged(object sender, EventArgs e)
        {
            if (_editingMultipleRecs) return;
            if (!flaggedCheckBox.Checked) return;
            lockedCheckBox.Checked = false;
        }

        #endregion

        #region validation events

        private void SquelchValidated(object sender, EventArgs e)
        {
            errorProvider1.SetError(squelch, String.Empty);
        }

        private void SquelchValidating(object sender, CancelEventArgs e)
        {
            if (ValidateEntry(squelch, 0, 100) == false)
            {
                e.Cancel = true;
            }
        }

        private void CWShiftValidating(object sender, CancelEventArgs e)
        {
            if (ValidateEntry(cwShift, 200, 1200) == false)
            {
                e.Cancel = true;
            }
        }

        private void CWShiftValidated(object sender, EventArgs e)
        {
            errorProvider1.SetError(cwShift, String.Empty);
        }

        private void FilterOrderValidating(object sender, CancelEventArgs e)
        {
            if (ValidateEntry(filterOrder, 10, 9999) == false)
            {
                e.Cancel = true;
            }
        }

        private void FilterOrderValidated(object sender, EventArgs e)
        {
            errorProvider1.SetError(filterOrder, String.Empty);
        }

        private void FilterBandwidthValidating(object sender, CancelEventArgs e)
        {
            if (ValidateEntry(filterBandwidth, 10, 250000) == false)
            {
                e.Cancel = true;
            }
        }

        private void FilterBandwidthValidated(object sender, EventArgs e)
        {
            errorProvider1.SetError(filterBandwidth, String.Empty);
        }

        private void ShiftValidated(object sender, EventArgs e)
        {
            errorProvider1.SetError(shift, String.Empty);
        }

        private void ShiftValidating(object sender, CancelEventArgs e)
        {
            if (ValidateEntry(shift, -100000000000000, 100000000000000) == false)
            {
                e.Cancel = true;
            }
        }

        private void CenterValidating(object sender, CancelEventArgs e)
        {
            if (ValidateEntry(frequency, 0, 3000000000) == false)
            {
                e.Cancel = true;
            }
        }

        private void CenterValidated(object sender, EventArgs e)
        {
            errorProvider1.SetError(center, String.Empty);
        }

        private void FrequencyValidating(object sender, CancelEventArgs e)
        {
            if (ValidateEntry(frequency, 0, 3000000000) == false)
            {
                e.Cancel = true;
            }
        }

        private void FrequencyValidated(object sender, EventArgs e)
        {
            errorProvider1.SetError(frequency, String.Empty);
        }

        #endregion

    }
}