﻿//  ******************************************
//  Frequency Manager Collapsible Panel
//  Copyright © Jeff Knapp 2012-2013 
//
//  This window displays the frequency manager controls.
//  ******************************************
//  Date        Changes
//  01/22/13    fixed err in setcurrentmode.
//  01/27/13    improved error message when sqlite couln't be accessed.
//  01/28/13    added config switch to disable freq normalization for scanning.
//  01/31/13    added edit windows for protocols & services.  
//  02/03/13    added multiRecordIndicator to show when there are multiple licensees on a frequency.
//  02/04/13    changed processfreq so that it selects from scanfreqtable when scanning groups.
//              changed StationTableSelect so that is supports selection from two different tables
//              based on type of scan.
//  02/09/13    added escaping of apostrophes for select string in EditButtonClick.
//  02/15/13    fixed issue with live track no selecting rows.
//              improved the finding frequencies in mem table when scanning groups.
//              added thread.sleep(1) after every change of center to give radio time to catch up.
//              fixed issue in sending the correct freq to Browse when manually tuning.
//  02/17/13    wrapped all tooltip settings in try/catch to prevent error dialogs.
//  02/18/13    removed tooltips for currentDescription and groupList because they were too unstable.
//  02/20/13    added setting of multiRecordIndicator to ImportRadioSettings.
//  05/02/13    zefie's log file mod
//  ******************************************

using System;
using System.Collections;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Data.SQLite;
using System.Diagnostics;//only used when debugging or logging
using System.Drawing;
using System.Globalization;
using System.IO;
using System.Linq;
using System.Reflection;
using System.Runtime.InteropServices;
using System.Text.RegularExpressions;
using System.Threading;
using System.Timers;
using System.Windows.Forms;
using SDRSharp.CollapsiblePanel;
using SDRSharp.Common;
using SDRSharp.Radio;
using Timer = System.Timers.Timer;

namespace SDRSharp.FreqMgr
{
    [DesignTimeVisible(true)]
    [Category("SDRSharp")]
    [Description("FreqMgr Panel")]
    public sealed partial class FreqMgrCollapsiblePanel : UserControl
    {
        #region FreqMgr Panel

        #region vars, objects, & constants

        // zefie's note:
        // Okay, so I COULD use SQLite like the rest of the plugin, however
        // I am going for backwards compatiblity. If I modify the SQLite DB
        // to add my config, it will appear "corrupt" to the original plugin,
        // should you decide to downgrade from my copy. Therefore, I will use
        // the SettingsPersister, and xml file like the Trunker does. this
        // will make an extra file, but wont break your database

        private readonly SettingsPersisterLogger _settingsPersister;
        private List<LogSettings> _loggerConfig;
        private Boolean _canScan; //true when we are able to scan
        private Boolean _formClosing;
        private Boolean _haveActivity;
        private Boolean _isScanningFreqRange; //true when scanning a range of freqs rather than a group
        private Boolean _preventFreqChangeRecursion;
        private const int WmKeydown = 0x100;
        private const int WmSyskeydown = 0x104;
        private const String Copyright = "Copyright © Jeff Knapp 2012-2013";
        private const String NormalizeScanFreqsKey = "freqMgrNormalizeScanFreqs";
        private DataSet _ds;
        private DataTable _stationTable = new DataTable();
        private Form _mf;
        private FrequencyActivityEventArgs _fae;
        private Groups _groupForm;
        private Icon _mfIcon;
        private Int32 _scanFreqPtr = -1;
        private long _hiFreq;
        private long _lowFreq;
        private long _stepSize;
        private Protocols _protocolForm;
        private readonly BackgroundWorker _queueMgr = new BackgroundWorker();
        private readonly DataTable _scanFreqTable = new DataTable();//memory table containing the freqs in a group to scan
        private readonly ISharpControl _controlInterface;
        private readonly Queue _logFreqs = new Queue();
        private readonly Timer _queueTimer = new Timer(); //processes queued frequency activities
        private readonly Timer _resetPreventFreqChangeRecursion;
        private readonly Timer _scanBusyTestTimer = new Timer(); //watches an open freq to determine when it closes
        private readonly Timer _scanInterFreqTimer = new Timer(); //time to sit on a freq waiting for a transmission
        private readonly Timer _scanTailTimer = new Timer(); //waits for additional activity after squelch closes
        private readonly ToolTip _appInfo = new ToolTip();
        private ScannerStateChangedEventArgs _ssc;
        private Services _serviceForm;
        private static Browse _browseForm;
        private String _currCallsign = String.Empty;
        private String _currDate = String.Empty;
        private String _currDescription = String.Empty;
        private String _currNotes = String.Empty;
        private String _currProtocol = String.Empty;
        private String _currService = String.Empty;
        private String _scanUnidFreqFound;
        private Thread _browseWindowThread;
        private Thread _groupWindowThread;
        private Thread _protocolWindowThread;
        private Thread _serviceWindowThread;
        public Boolean IsScanning = false;
        public event EventHandler<FrequencyActivityEventArgs> FrequencyActivity;
        public event EventHandler<GroupsUpdatedEventArgs> GroupsUpdated;
        public event EventHandler<ScannerStateChangedEventArgs> ScannerStateChanged;
        private bool loggingConfigured = false;
        private bool loggingEnabled = false;
        private bool loggingSystemMessageDisplayed = false;
        private bool initializing;
        private bool loggerScanning = false;
        private string lastcall;
        private System.Timers.Timer _logTimer = null;
        private string lastlogtext;
        #endregion

        /// <summary>
        /// constructor
        /// </summary>
        public FreqMgrCollapsiblePanel(ISharpControl control)
        {
            try
            {
                InitializeComponent();
                _controlInterface = control;
                //connect event handlers
                _controlInterface.PropertyChanged += ControlPropertyChanged;
                Load += FreqMgrPanelLoad;
                currentFreq.Click += ForceFocus;
                modeLabel.Click += ForceFocus;
                currentMode.Click += ForceFocus;
                descriptionLabel.Click += ForceFocus;
                currentDescription.Click += ForceFocus;
                scanFuncBox.Click += ForceFocus;
                scanGroupLabel.Click += ForceFocus;
                label2.Click += ForceFocus;
                label1.Click += ForceFocus;
                secDelayLabel.Click += ForceFocus;
                msPauseLabel.Click += ForceFocus;
                fmGroupBox.Click += ForceFocus;
                //connect extended tab control
                var nativeTabControl1 = new NativeTabControl();
                nativeTabControl1.AssignHandle(scanTypeTabControl.Handle);

                // get log settings
                initializing = true;
                _settingsPersister = new SettingsPersisterLogger();
                _loggerConfig = _settingsPersister.readConfig();
                ProcessSavedSettings();
                initializing = false;

                #region setup timers

                //helps prevent a stack-destroying loop
                _resetPreventFreqChangeRecursion = new Timer { Enabled = false, AutoReset = false, Interval = 200 };
                _resetPreventFreqChangeRecursion.Elapsed += ResetPreventFreqChangeRecursionElapsed;

                //tests a scanned freq to see if it is still busy
                _scanBusyTestTimer.Interval = 200; //fixed value
                _scanBusyTestTimer.AutoReset = false;
                _scanBusyTestTimer.Elapsed += ScanBusyTestTimerElapsed;

                //delay to wait for additional traffic on a scanned freq
                _scanTailTimer.Interval = 2000; //user-adjustable
                _scanTailTimer.AutoReset = false;
                _scanTailTimer.Elapsed += ScanTailTimerElapsed;

                //how long to wait on each freq for squelch open
                _scanInterFreqTimer.Interval = 500; //user-adjustable
                _scanInterFreqTimer.AutoReset = false;
                _scanInterFreqTimer.Elapsed += ScanInterFreqTimerElapsed;

                //processes the scanned frequency queue
                _queueTimer.Interval = 2000; //fixed value
                _queueTimer.AutoReset = false;
                _queueTimer.Elapsed += QueueTimerElapsed;

                #endregion

                #region scanner objects

                //_scanFreqTable contains the list of frequencies in the selected scan group
                var dc = new DataColumn("Frequency", typeof(long));
                _scanFreqTable.Columns.Add(dc);
                dc = new DataColumn("Center", typeof(long));
                _scanFreqTable.Columns.Add(dc);
                dc = new DataColumn("Mode", typeof(String));
                _scanFreqTable.Columns.Add(dc);
                dc = new DataColumn("FilterType", typeof(String));
                _scanFreqTable.Columns.Add(dc);
                dc = new DataColumn("Shift", typeof(long));
                _scanFreqTable.Columns.Add(dc);
                dc = new DataColumn("ShiftEnabled", typeof(Int32));
                _scanFreqTable.Columns.Add(dc);
                dc = new DataColumn("FilterBandwidth", typeof(Int32));
                _scanFreqTable.Columns.Add(dc);
                dc = new DataColumn("FilterOrder", typeof(Int32));
                _scanFreqTable.Columns.Add(dc);
                dc = new DataColumn("Squelch", typeof(Int32));
                _scanFreqTable.Columns.Add(dc);
                dc = new DataColumn("SquelchEnabled", typeof(Int32));
                _scanFreqTable.Columns.Add(dc);
                dc = new DataColumn("CWShift", typeof(Int32));
                _scanFreqTable.Columns.Add(dc);
                dc = new DataColumn("Id", typeof(Int32));
                _scanFreqTable.Columns.Add(dc);
                dc = new DataColumn("Description", typeof(String));
                _scanFreqTable.Columns.Add(dc);
                dc = new DataColumn("Notes", typeof(String));
                _scanFreqTable.Columns.Add(dc);
                dc = new DataColumn("Service", typeof(String));
                _scanFreqTable.Columns.Add(dc);
                #endregion

                //verify the FreqMgr database exists
                if (!CheckDb())
                {
                    fmGroupBox.Enabled = false;
                    scanFuncBox.Enabled = false;
                    return;
                }
                //load subset of data columns into a searchable table in memory
                LoadStationList();
                //load scan group list
                LoadScanGroupList();
            }
            catch (Exception ex)
            {
                String msg = String.Format("An error occurred.  The error was: \n\r{0}",
                                           ex.Message + "\n\r" + ex.StackTrace.Trim());
                if (ex.Message.Contains("Could not load file or assembly 'System.Data.SQLite"))
                {
                    msg =
                        "Required file System.Data.SQLite.dll could not be opened.  Your installation may be corrupted.\n\r\n\rFrequency Manager + Scanner will be disabled.";
                }
                MessageBox.Show(msg, "Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
                fmGroupBox.Enabled = false;
                scanFuncBox.Enabled = false;
            }
        }

        void ProcessSavedSettings()
        {
            // There is probably a cleaner way, but right now I'm going for functionality.
            try
            {
                // This will fail if we have no existing configuration
                this.singleLogFilePath.Text = _loggerConfig[0].logPath;

                if (this.singleLogFilePath.Text != null)
                {
                    this.loggingConfigured = true;
                    this.singleLogFileEnabled.Checked = _loggerConfig[0].isLoggingEnabled;
                }
            }
            catch (Exception ex)
            {
                // We have no existing config, which is fine, but we need to catch the exception
                // that _trunkerConfig[0] is an invalid index
                DoNothing(ex);
            }
        }

        void WriteConfig()
        {
            // There is probably a cleaner way, but right now I'm going for functionality.

            // This is basically the opposite of ProcessSavedSettings().
            // It shouldn't really matter what order we save these.

            // We are going to try to read from _trunkerConfig[0]. If we are successful, it exists
            // and we can write the settings. If not, we need to catch the exception and create the index.
            if (!initializing)
            {
                // don't write the config when we are initializing.
                // WriteConfig() will be fired when we change the checkbox due to OnChange
                // We want this but not during initialization.
                try
                {
                    _loggerConfig[0].settingsExist();
                }
                catch (Exception ex)
                {
                    DoNothing(ex);
                    _loggerConfig.Insert(0, new LogSettings());
                }
                _loggerConfig[0].logPath = this.singleLogFilePath.Text;
                _loggerConfig[0].isLoggingEnabled = this.singleLogFileEnabled.Checked;
                _settingsPersister.writeConfig(_loggerConfig);
            }

        }

        /// <summary>
        /// get references to main form controls, and do some UI setup,
        /// during our first load
        /// </summary>
        private void FreqMgrPanelLoad(object sender, EventArgs e)
        {
            try
            {
                _mf = ParentForm;
                _mfIcon = _mf.Icon;
                Parent.Disposed += ParentDisposed;

                #region setup some FreqMgr controls

                //button enablements
                scanFlagButton.Enabled = false;
                scanLockOutButton.Enabled = false;
                scanSkipButton.Enabled = false;
                //set scanner values from config
                secDelay.Value = Convert.ToDecimal(Utils.GetLongSetting("freqMgrScanSecDelay", 3));
                msPause.Value = Convert.ToDecimal(Utils.GetLongSetting("freqMgrScanMsPause", 250));
                _scanUnidFreqFound = Utils.GetStringSetting("freqMgrUnidFoundMessage",
                                                            "Unidentified. Found during frequency scan.");
                //configure scan queue objects
                _queueMgr.WorkerSupportsCancellation = true;
                _queueMgr.DoWork += QueueMgrDoWork;
                _queueMgr.RunWorkerCompleted += QueueMgrRunWorkerCompleted;
                //connect scan freq up-down control events
                scanLowerFreqNud.Enter += NudSelectAll;
                scanUpperFreqNud.Enter += NudSelectAll;
                //misc
                _currDescription = Common.UnidText;
                currentDescription.Text = Common.UnidText;
                currentMode.Text = GetCurrentMode();
                GetStepSize();

                //set private freq to match the current radio settings - this needs to be done last
                currentFreq.Text = _controlInterface.Frequency.ToString(Common.NumericFormat);

                #endregion

#if (TESTING)
                //these are just settings I like to use when testing
                try
                {
                    //_controlInterface.DetectorType = DetectorType.NFM;
                    //_preventFreqChangeRecursion = false;
                    //_controlInterface.CenterFrequency = 451025000;
                    //_controlInterface.Frequency = 451025000;
                    //((TrackBar)_mf.Controls["fftContrastTrackBar"]).Value = -16;
                    //_mf.Height = 1020;
                    //_mf.Width = 1010;
                    BrowseButtonCheckChanged(new object(), new EventArgs());
                    //scanTypeTabControl.SelectedIndex = 1;
                    //scanLowerFreqNud.Text = "854,000,000";
                    //scanUpperFreqNud.Text = "860,000,000";
                    //recordButton.Checked = true;
                    //msPause.Text = "50";
                    //secDelay.Text = "0";
                    //groupList.SelectedItem = "test";

                #region configure the controlPanel

                    ControlCollection cc = _mf.Controls["scrollPanel"].Controls["controlPanel"].Controls;
                    cc.RemoveByKey("agcCollapsiblePanel");
                    cc.RemoveByKey("displayCollapsiblePanel");

                    Control[] fmArray = _mf.Controls.Find("FreqMgrCollapsiblePanel", true);
                    var freqMgrCollapsiblePanel = (FreqMgrCollapsiblePanel)fmArray[0];
                    var cp = (CollapsiblePanel.CollapsiblePanel)freqMgrCollapsiblePanel.Parent;
                    cp.PanelState = PanelStateOptions.Expanded;
                    Control[] apArray = _mf.Controls.Find("audioCollapsiblePanel", true);
                    var ap = (CollapsiblePanel.CollapsiblePanel)apArray[0];
                    ap.NextPanel = cp;

                #endregion
                }
                catch
                {
                    //swallow it
                }
#endif
            }
            catch (Exception ex)
            {
                String msg = String.Format("An error occurred.  The error was: \n\r{0}",
                                           ex.Message + "\n\r" + ex.StackTrace.Trim());
                MessageBox.Show(msg, "Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
                fmGroupBox.Enabled = false;
                scanFuncBox.Enabled = false;
            }
        }

        /// <summary>
        /// forces the browse and group window threads to end when this form's
        /// parent is disposed.  otherwise we have an exception in sdr#
        /// </summary>
        private void ParentDisposed(object sender, EventArgs e)
        {
            try
            {
                _formClosing = true;
                //stop all scanning
                _canScan = false;
                _queueTimer.Stop();
                _scanBusyTestTimer.Stop();
                _scanInterFreqTimer.Stop();
                _scanTailTimer.Stop();

                //kill the browse window thread if necessary
                //I know Abort is the wrong usage but I couldn't find an alternative
                try { _browseWindowThread.Abort(); }
                catch (ThreadAbortException)
                { Thread.ResetAbort(); }
                catch (Exception)
                { }

                //kill the group window thread if necessary
                //I know Abort is the wrong usage but I couldn't find an alternative
                try { _groupWindowThread.Abort(); }
                catch (ThreadAbortException)
                { Thread.ResetAbort(); }
                catch (Exception)
                { }

                //kill the protocol window thread if necessary
                //I know Abort is the wrong usage but I couldn't find an alternative
                try { _protocolWindowThread.Abort(); }
                catch (ThreadAbortException)
                { Thread.ResetAbort(); }
                catch (Exception)
                { }

                //kill the service window thread if necessary
                //I know Abort is the wrong usage but I couldn't find an alternative
                try { _serviceWindowThread.Abort(); }
                catch (ThreadAbortException)
                { Thread.ResetAbort(); }
                catch (Exception)
                { }

                Application.Exit();
            }
            catch (Exception)
            {
                //swallow it
            }
        }

        /// <summary>
        /// handle notifications from SDR# that one of its properties has changed
        /// </summary>
        private void ControlPropertyChanged(object sender, PropertyChangedEventArgs e)
        {
            switch (e.PropertyName)
            {
                case "Frequency":
#if (LOG)
{  
  Common.Log("frequency " + _controlInterface.Frequency);
}
#endif
                    ProcessFreqChange();
                    break;
                case "CenterFrequency":
                    //do nothing; this is a placeholder for future use
#if (LOG)
{
Common.Log("center " + _controlInterface.CenterFrequency);
}
#endif
                    break;
                case "DetectorType":
                    currentMode.Text = GetCurrentMode();
                    break;
                case "StepSize":
                    GetStepSize();
                    break;
                case "StopRadio":
                    _queueTimer.Stop();
                    break;
            }
        }

        /// <summary>
        /// ResetPreventFreqChangeRecursion timer elapsed event.
        /// </summary>
        /// <remarks>this timer and event are used when the auto-centering function changes the frequency.
        /// because that change happens inside the frequencyNumericUpDown_ValueChanged event, there
        /// is the risk of a stack-destroying recursion.  using this timer the event builds in a delay
        /// to allow the implicitly fired event to finish without re-firing iteself.</remarks>
        private void ResetPreventFreqChangeRecursionElapsed(object sender, ElapsedEventArgs e)
        {
            _preventFreqChangeRecursion = false;
        }

        /// <summary>
        /// processes a change of the sdrsharp frequency
        /// regardless of how it happened
        /// </summary>
        private void ProcessFreqChange()
        {
#if (LOG)
            Common.Log("enter");
#endif
            if (_formClosing) return;
            if (_preventFreqChangeRecursion)
            {
                //reset the control flag and bail out
                _preventFreqChangeRecursion = false;
                return;
            }
            try
            {
                //initialize our collapsible panel and private vars
                currentFreq.Text = _controlInterface.Frequency.ToString(Common.NumericFormat);
                currentId.Text = "0"; //currentId is a hidden label used to hold the current database record id
                _currCallsign = String.Empty;
                _currProtocol = String.Empty;
                _currService = String.Empty;
                _currNotes = String.Empty;

                //find the frequency in our database that is closest to the one selected 
                //in the spectrum window, and adjust the radio's frequency to match the database.                     

                String low;
                String hi;
                long selectedFreq = _controlInterface.Frequency;
#if (LOG)
{
    Common.Log("received freq " + selectedFreq);
}
#endif
                //if stepsize > 6250 and if we are NOT scanning a range of frequencies,
                //we use approximation to convert "near-misses" 
                //on frequencies into the actual frequencies from the database
                if ((_stepSize > 6250) && (!_isScanningFreqRange))
                {
                    //divide the current bandwidth in two after removing a token 
                    //amount to reduce our possible search frequency range                        
                    Decimal halfBW = (_controlInterface.FilterBandwidth - 100) / 2;
                    low = (selectedFreq - (halfBW)).ToString(CultureInfo.InvariantCulture);
                    hi = (selectedFreq + (halfBW)).ToString(CultureInfo.InvariantCulture);
                }
                else
                {
                    //search based on the actual chosen freq
                    low = selectedFreq.ToString(CultureInfo.InvariantCulture);
                    hi = selectedFreq.ToString(CultureInfo.InvariantCulture);
                }

                //search the in-memory table for this frequency
                String sel;
                DataRow[] result;
                if ((IsScanning) && (!_isScanningFreqRange))
                {
                    //must be scanning a group, look in the station and groupfreq tables
                    sel = "Frequency = {0}";
                    result = StationTableSelect(String.Format(sel, selectedFreq), true);

                }
                else
                {
                    sel = "Frequency >= {0} and Frequency <= {1}";
                    result = StationTableSelect(String.Format(sel, low, hi), false);
                }
                //if there are multiple rows in result, we'll finish up positioned to the last one
                multiRecordIndicator.Visible = false;
                if (result != null)
                {
                    if (result.Length > 0)
                    {
                        //if (result.Length > 1) multiRecordIndicator.Visible = true;
                        multiRecordIndicator.Visible = result.Length > 1;
                        var freq = selectedFreq;
                        foreach (DataRow row in result)
                        {
                            //copy pertinent data to our FreqMgr collapsible panel
                            currentId.Text = row["Id"].ToString();
                            _currDescription = row["Description"].ToString();
                            currentDescription.Text = String.Format("{0} ({1})", row["Description"], row["Service"]);
                            //determine the mode and set the radio
                            if (!_isScanningFreqRange)
                            {
                                SetCurrentMode(row["Mode"].ToString());
                                currentMode.Text = row["Mode"].ToString();
                            }

                            //if the radio frequency doesn't match the lookup frequency
                            //change the radio frequency to match the lookup
                            freq = (long)row.ItemArray[0];
                            if ((_controlInterface.Frequency != freq) && (_stepSize > 6250))
                            {
                                currentFreq.Text = freq.ToString(Common.NumericFormat);
                                _preventFreqChangeRecursion = true;
#if (LOG)
{  
  Common.Log("setting freq "+ freq);
}
#endif
                                _controlInterface.Frequency = freq;
                                //IMPORTANT: call _resetPreventFreqChangeRecursion.Start
                                //immediately after changing controlInterface.Frequency
                                _resetPreventFreqChangeRecursion.Start();
                            }
                        }
                        //update the browse form grid position to this frequency
                        if (_browseForm != null)
                        {
                            if ((IsScanning) && (!_isScanningFreqRange)) //scanning a group
                            {
                                _browseForm.PositionGrid(freq, _currDescription);
                            }
                            else //not scanning or scanning by range
                            {
                                _browseForm.PositionGrid(freq);
                            }
                        }
                    }
                    else
                    {
                        _currDescription = Common.UnidText;
                        currentDescription.Text = Common.UnidText;
                    }
                }

                //handle the case where there are no rows in result; we 
                //want to deselect the current row on the grid because there
                //is no match to the freq in the spectrum window
                if ((_browseForm != null) && (result.Length == 0))
                {
                    _browseForm.DeselectCurrentRow();
                }
            }
            catch (Exception)
            {
                //swallow it
            }
#if (LOG)
            Common.Log("exit");
#endif
        }

        /// <summary>
        /// gets radio settings from a record returned from Edit or Browse
        /// </summary>
        /// <param name="settings">FreqRecord object containing station data</param>
        /// <param name="updateRadio">True/False to send changes to the radio </param>
        internal void ImportRadioSettings(FreqRecord settings, Boolean updateRadio)
        {
            if (_formClosing) return;
            try
            {
                multiRecordIndicator.Visible = false;
                if ((updateRadio) && (settings.Frequency > 0)) //freq will be zero on a delete from the browse window
                {
                    //preventFreqChangeRecursion is used to control activity elsewhere; 
                    //otherwise we get into a stack-destroying recursion
                    _preventFreqChangeRecursion = true;
                    //update the radio
                    try
                    {
                        _controlInterface.FrequencyShift = settings.Shift;
                        _controlInterface.CenterFrequency = settings.Center;
                        Thread.Sleep(1);
#if (LOG)
{
    Common.Log("setting freq " + settings.Frequency);
}
#endif
                        _controlInterface.Frequency = settings.Frequency;
                        _controlInterface.FilterOrder = settings.FilterOrder;
                        _controlInterface.SquelchThreshold = settings.Squelch;
                        _controlInterface.CWShift = settings.CWShift;
                        _controlInterface.FilterType =
                            (WindowType)(Array.IndexOf(Common.FilterTypeArray, settings.FilterType) + 1);
                        _controlInterface.FrequencyShiftEnabled = Convert.ToBoolean(settings.ShiftEnabled);
                        _controlInterface.SquelchEnabled = Convert.ToBoolean(settings.SquelchEnabled);

                        //determine the mode radiobutton
                        SetCurrentMode(settings.Mode);

                        //FilterBandwidth must be set AFTER the mode is changed
                        _controlInterface.FilterBandwidth = settings.FilterBandwidth;

                        //update freqmgr gui
                        currentId.Text = settings.Id.ToString(CultureInfo.InvariantCulture);
                        currentFreq.Text = settings.Frequency.ToString(Common.NumericFormat);
                        currentMode.Text = settings.Mode;
                        _currDescription = settings.Description;
                        _currProtocol = settings.Protocol;
                        _currService = settings.Service;
                        _currCallsign = settings.Callsign;
                        _currNotes = settings.Notes;
                        _currDate = settings.Date;
                        currentDescription.Text = String.Format("{0} ({1})", settings.Description, settings.Service);
                        
                        try
                        {
                            using (SQLiteConnection conn = Common.GetSqlConnection())
                            {
                                String sel = "select count(*) from data where frequency = {0}";
                                var cmd = new SQLiteCommand(String.Format(sel, settings.Frequency), conn);
                                Int32 count = Convert.ToInt32(cmd.ExecuteScalar());
                                if (count > 1)
                                {
                                    multiRecordIndicator.Visible = true;
                                }
                            }
                        }
                        catch (Exception)
                        {
                            //swallow it
                        }



                    }
                    catch (ApplicationException ex)
                    {
                        //this happens when the user tries to scan a sound card.
                        //we cannot test for this source when the scan button is pressed;
                        //SDR# doesn't tell us what the source is.
                        if (ex.Message.Contains("no front end is connected"))
                        {
                            scanButton.Checked = false;
                            _canScan = false;
                            MessageBox.Show("That Source cannot be updated from the Browse window", "Error",
                                            MessageBoxButtons.OK, MessageBoxIcon.Exclamation);
                            return;
                        }
                        throw;
                    }
                }
                //always update the memory data table
                LoadStationList();
                //update scanfreqtable if necessary
                if ((scanTypeTabControl.SelectedIndex == 0) && (_canScan))
                {
                    LoadScanFreqTable();
                }
                //reset the control flag
                _preventFreqChangeRecursion = false;
            }
            catch (ArgumentOutOfRangeException)
            {
                String msg = String.Format("SDR# refused one of your settings as being invalid " +
                                           "but didn't say which one.  " +
                                           "This usually happens when the Frequency and Center values are not in the same range." +
                                           "\n\r\n\rVerify that all values are 'reasonable' radio settings and try again.");
                MessageBox.Show(msg, "Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
            catch (Exception ex)
            {
                MessageBox.Show(String.Format("An error occurred: {0}\n\r{1}", ex.Message, ex.StackTrace), "Error",
                                MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
        }

        /// <summary>
        /// Checks to see if our database exists and creates it if it does not
        /// </summary>
        /// <remarks>Creates the database and its tables, and initializes
        /// the tables with appropriate data</remarks>
        private static Boolean CheckDb()
        {
            try
            {
                //open the database
                String dbFile = String.Format("{0}\\{1}", Common.DbPath, Common.DbFileName);
                if (!Directory.Exists(Common.DbPath))
                {
                    Directory.CreateDirectory(Common.DbPath);
                }
                using (SQLiteConnection conn = Common.GetSqlConnection())
                {
                    Int32 rowCount = conn.GetSchema("Tables").Rows.Count;

                    if (rowCount == 7)
                    {
                        //in future versions we'll check the version number here
                        //and make any changes that are necesary.
                        return true;
                    }
                    if (rowCount == 6)
                    {
                        #region if rev1 then update db version

                        //if we got here it means the database exists but it is the original version
                        const string msg =
                            "THE FREQUENCY MANAGER DATABASE MUST BE UPGRADED.\n\rYou should back up your database before proceeding.\n\rDo you want to upgrade your database right now anyway?";
                        DialogResult dr = MessageBox.Show(msg, "Database Warning", MessageBoxButtons.YesNo,
                                                          MessageBoxIcon.Warning);
                        if (dr == DialogResult.No)
                        {
                            MessageBox.Show(
                                "Exit SDR# and back up your database, then restart SDR# and upgrade your database",
                                "Database Warning",
                                MessageBoxButtons.OK, MessageBoxIcon.Information);
                            return false;
                        }
                        using (SQLiteTransaction sqlTransaction = conn.BeginTransaction())
                        {
                            try
                            {
                                //add columns to data table
                                const string setZero = "update data set {0} = 0";
                                const string newLocked = "alter table data add column Locked NUMBER";
                                var cmd = new SQLiteCommand(newLocked, conn);
                                cmd.ExecuteNonQuery();
                                cmd = new SQLiteCommand(String.Format(setZero, "Locked"), conn);
                                cmd.ExecuteNonQuery();
                                const string newFlagged = "alter table data add column Flagged NUMBER";
                                cmd = new SQLiteCommand(newFlagged, conn);
                                cmd.ExecuteNonQuery();
                                cmd = new SQLiteCommand(String.Format(setZero, "Flagged"), conn);
                                cmd.ExecuteNonQuery();
                                //create version table
                                SetDbVersion(2, conn);
                                //new services
                                var services = new[] { "Broadcast", "Standard Frequency", Common.UnidText };
                                AddServices(services, conn);
                                //new protocols
                                var protocols = new[] { "DAB", Common.UnidText };
                                AddProtocols(protocols, conn);

                                //reformat the Date column to the neutral locale format
                                const string sel = "select Id,Date from Data";
                                cmd = new SQLiteCommand(sel, conn);
                                var da = new SQLiteDataAdapter(cmd);
                                var ds = new DataSet();
                                da.Fill(ds);
                                const string upd = "update data set Date = @newdate where Id = @id";
                                cmd = new SQLiteCommand(upd, conn);
                                foreach (DataRow item in ds.Tables[0].Rows)
                                {
                                    String dateString = item["Date"].ToString();
                                    if (String.IsNullOrEmpty(dateString)) continue;
                                    String newDate;
                                    DateTime oldDate;
                                    Boolean result = DateTime.TryParse(dateString, CultureInfo.CurrentCulture, DateTimeStyles.AssumeLocal, out oldDate);
                                    if (result)
                                    {
                                        newDate = oldDate.ToString(Common.DateSaveFormat, CultureInfo.InvariantCulture);
                                    }
                                    else
                                    {
                                        //we have to reconstruct oldDate because it could not be parsed
                                        String[] parts = dateString.Split('/'); //should be mm/dd/yy
                                        if (parts[2].Length == 2)
                                        {
                                            parts[2] = "20" + parts[2];
                                        }
                                        oldDate = new DateTime(Convert.ToInt32(parts[2]),
                                                               Convert.ToInt32(parts[0]),
                                                               Convert.ToInt32(parts[1]));
                                        newDate = oldDate.ToString(Common.DateSaveFormat, CultureInfo.InvariantCulture);
                                    }
                                    cmd.Parameters.Clear();
                                    cmd.Parameters.AddWithValue("@newdate", newDate);
                                    cmd.Parameters.AddWithValue("@id", item["Id"]);
                                    cmd.ExecuteNonQuery();
                                }
                                //commit data changes and exit
                                sqlTransaction.Commit();
                            }
                            catch (Exception ex)
                            {
                                //rollback transaction and delete the damaged database
                                sqlTransaction.Rollback();
                                sqlTransaction.Dispose();
                                conn.Close();
                                if (File.Exists(dbFile))
                                {
                                    try
                                    {
                                        File.Delete(dbFile);
                                    }
                                    catch (Exception)
                                    {
                                        //swallow it
                                    }
                                }
                                throw new Exception("The upgrade failed and your damaged database has been deleted.", ex);
                            }
                        }
                        MessageBox.Show("The database was successfully upgraded", "Database Upgrade",
                                        MessageBoxButtons.OK, MessageBoxIcon.Information);
                        return true;

                        #endregion
                    }

                    #region process damaged database

                    if ((rowCount > 1) && (rowCount < 6))
                    {
                        //if we got here it means the database exists but it contains 
                        //too few tables; so we delete it and create a new one from scratch
                        const string msg = "The Frequency Manager database exists but it is damaged.\n\r" +
                                           "Do you want to delete it and create a new database?";
                        DialogResult dr = MessageBox.Show(msg, "Database Error", MessageBoxButtons.YesNo,
                                                          MessageBoxIcon.Warning);
                        if (dr == DialogResult.No)
                        {
                            return false;
                        }
                        conn.Close();
                        //without garbage collection closed/disposed objects
                        //will still hold file handles, preventing us from
                        //deleting the database file.
                        GC.Collect();
                        GC.WaitForPendingFinalizers();
                        Thread.Sleep(1);
                        File.Delete(dbFile);
                    }

                    #endregion
                }

                #region create new database

                using (SQLiteConnection conn = Common.GetSqlConnection())
                {
                    using (SQLiteTransaction sqlTransaction = conn.BeginTransaction())
                    {
                        try
                        {
                            #region create tables

                            //create frequency data table
                            String ct = "create table Data (Id INTEGER PRIMARY KEY AUTOINCREMENT, Frequency NUMBER," +
                                        "Center NUMBER,Description TEXT,Mode TEXT," +
                                        "Protocol TEXT,[Call Sign] TEXT, Service TEXT,[Shift Enabled] NUMBER, " +
                                        "Shift NUMBER,[Filter Type] TEXT," +
                                        "[Filter Bandwidth] NUMBER,[Filter Order] NUMBER,[Squelch Enabled] NUMBER," +
                                        "Squelch NUMBER,[CW Shift] NUMBER," +
                                        "Notes TEXT, Date TEXT, Locked NUMBER, Flagged NUMBER )";

                            var cmd = new SQLiteCommand(ct, conn);
                            cmd.ExecuteNonQuery();

                            //create groups table
                            ct = "create table Groups (Id INTEGER PRIMARY KEY AUTOINCREMENT, Name TEXT)";
                            cmd = new SQLiteCommand(ct, conn);
                            cmd.ExecuteNonQuery();

                            //create group-freq table
                            ct =
                                "create table GroupFreq (Id INTEGER PRIMARY KEY AUTOINCREMENT, GroupId NUMBER, FreqId NUMBER)";
                            cmd = new SQLiteCommand(ct, conn);
                            cmd.ExecuteNonQuery();

                            //create service table
                            ct = "create table Service (Name TEXT PRIMARY KEY)";
                            cmd = new SQLiteCommand(ct, conn);
                            cmd.ExecuteNonQuery();

                            #endregion

                            #region services array

                            var services = new[]
                                               {
                                                   "Aeronautical", "Broadcast", "Business", "Corrections", "Data",
                                                   "Emergency Ops",
                                                   "EMS Dispatch", "EMS Tactical", "EMS Talk", "Federal",
                                                   "Fire Dispatch",
                                                   "Fire Tactical", "Fire Talk", "Ham", "Hospital", "Interop",
                                                   "Law Dispatch", "Law Tactical", "Law Talk", "Maritime", "Media",
                                                   "Military",
                                                   "Multi-Dispatch", "Paging", "Public Safety", "Public Works",
                                                   "Schools",
                                                   "Security",
                                                   "Standard Frequency", "Transportation", "Trunked",
                                                   "Utilities", Common.UnidText
                                               };

                            #endregion

                            AddServices(services, conn);

                            //create protocol table
                            ct = "create table Protocol (Name TEXT PRIMARY KEY)";
                            cmd = new SQLiteCommand(ct, conn);
                            cmd.ExecuteNonQuery();

                            #region protocols array

                            var protocols = new[]
                                                {
                                                    "110A", "1382", "141A", "4285", "ACARS", "ALE 400", "AM",
                                                    "AMTOR-ARQ", "AMTOR-FEC", "ARQ-E", "ASCII", "BPSK 125",
                                                    "BPSK 250", "BPSK 31", "BPSK 63", "CCW",
                                                    "CHIP", "CONTESTIA 8/250", "Contestia 8/500",
                                                    "Contestia 16/ 500", "Contestia 3211000", "Contestia 4/125",
                                                    "Contestia 4/250", "Contestia 4/500", "Contestia 8/1000",
                                                    "Contestia 16/1000", "Contestia 8/2000", "Contestia 16/2000",
                                                    "Contestia 3212000", "Contestia 64/2000", "COQUELET", "CW", "DAB",
                                                    "DominoEX-11", "DominoEX-16", "DominoEX-22", "DominoEX-4",
                                                    "DominoEX-5", "DominoEX-8", "DominoF", "DTMF", "ERMES", "FEC 31",
                                                    "FELD HELL X5", "FELD HELL X9", "FSK HELL 105", "FSK HELL 245",
                                                    "FLEX", "FM HELL", "FM NARROW", "FM WIDE", "GMDSS", "HELL 80",
                                                    "HF FAX",
                                                    "HFDL", "IEC 870-5", "JT65", "Lentus", "LSB", "MFSK-16", "MFSK-32",
                                                    "MFSK-4", "MFSK-64", "MFSK-8", "MOBITEX", "MT 63-1000", "MT 63-2000"
                                                    ,
                                                    "MT 63-500", "NAVTEX", "NWR", "Olivia 8/250", "Olivia 8/500",
                                                    "Olivia 16/500",
                                                    "Olivia 32/1000", "Olivia 4/125", "Olivia 4/250", "Olivia 4/500",
                                                    "Olivia 8/1000", "Olivia 16/1000", "Olivia 8/2000", "Olivia 16/2000"
                                                    ,
                                                    "Olivia 32/2000", "Olivia 64/2000", "P25", "PACKET", "PACTOR", "PAX"
                                                    ,
                                                    "POCSAG", "PSK 10", "PSK 220F", "PSK 63F", "PSK HELL",
                                                    "PSK63F DIGISSTV"
                                                    ,
                                                    "PSKAM 10", "PSKAM 31", "PSKAM 50", "QPSK 125", "QPSK 250",
                                                    "QPSK 31",
                                                    "QPSK 63", "QRSS", "RTTY 100", "RTTY 110", "RTTY 150", "RTTY 200",
                                                    "RTTY 45", "RTTY 50", "RTTY 75", "RTTYM 8/250", "RTTYM 8/500",
                                                    "RTTYM 16/ 500", "RTYM 32/1000", "RTTYM 4/125", "RTTYM 4/250",
                                                    "RTTYM 4/500", "RTTYM 8/1000", "RTTYM 16/1000", "RTTYM 8/2000",
                                                    "RTTYM 16/2000", "RTTYM 32/2000", "RTTYM 64/2000", "SELCAL",
                                                    "SITOR A", "SITOR B", "SLOW HELL", "SSTV", "SYNOP", "THOR 16",
                                                    "THOR 4", "THOR 8", "THROB X-1", "THROB X-2", "THROB X-4", "THROB-1"
                                                    ,
                                                    "THROB-2", "THROB-4", "USB", "Voice",
                                                    "WEFAX-IOC288", "WEFAX-IOC576", Common.UnidText
                                                };

                            #endregion

                            AddProtocols(protocols, conn);

                            //create version table
                            SetDbVersion(2, conn);

                            // Commit the changes into the database
                            sqlTransaction.Commit();
                        }
                        catch (Exception ex)
                        {
                            //rollback transaction and delete the damaged database
                            sqlTransaction.Rollback();
                            sqlTransaction.Connection.Close();
                            sqlTransaction.Dispose();
                            conn.Close();
                            if (File.Exists(dbFile))
                            {
                                //without garbage collection closed/disposed objects
                                //will still hold file handles, preventing us from
                                //deleting the database file.
                                GC.Collect();
                                GC.WaitForPendingFinalizers();
                                Thread.Sleep(1);
                                File.Delete(dbFile);
                            }
                            throw new Exception("The upgrade failed and your damaged database has been deleted.", ex);
                        }
                    }
                }

                #endregion

                MessageBox.Show("The Frequency Manager database was created successfully", "Success",
                                MessageBoxButtons.OK, MessageBoxIcon.Information);
            }
            catch (Exception ex)
            {
                String msg = String.Format("An error occurred.  The error was: \n\r{0}",
                                           ex.Message + "\n\r" + ex.StackTrace.Trim());
                MessageBox.Show(msg, "Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
            return true;
        }

        /// <summary>
        /// adds protocols to the database.
        /// we're already in a transaction at this point.
        /// </summary>
        private static void AddProtocols(IEnumerable<string> protocols, SQLiteConnection conn)
        {
            try
            {
                const string ins = "insert into Protocol (name) values (\"{0}\")";
                foreach (SQLiteCommand cmd in protocols.Select(item => new SQLiteCommand(String.Format(ins, item), conn)))
                {
                    cmd.ExecuteNonQuery();
                }
            }
            catch (Exception ex)
            {
                if (!ex.Message.Contains("constraint violation"))
                {
                    throw;
                }
            }
        }

        /// <summary>
        /// adds service types to the database.
        /// we're already in a transaction at this point.
        /// </summary>
        private static void AddServices(IEnumerable<string> services, SQLiteConnection conn)
        {
            try
            {
                const string ins = "insert into Service (name) values (\"{0}\")";
                foreach (SQLiteCommand cmd in services.Select(item => new SQLiteCommand(String.Format(ins, item), conn)))
                {
                    cmd.ExecuteNonQuery();
                }
            }
            catch (Exception ex)
            {
                if (!ex.Message.Contains("constraint violation"))
                {
                    throw;
                }
            }
        }

        /// <summary>
        /// creates the Version table and sets the rev column.
        /// we're already in a transaction at this point.
        /// </summary>
        private static void SetDbVersion(Int16 ver, SQLiteConnection conn)
        {
            try
            {
                //in the future we'll update the table if it already exists
                const string ct = "create table Version (Rev NUMBER)";
                var cmd = new SQLiteCommand(ct, conn);
                cmd.ExecuteNonQuery();
                cmd = new SQLiteCommand(String.Format("insert into Version (Rev) values ({0})", ver), conn);
                cmd.ExecuteNonQuery();
            }
            catch (Exception)
            {
                MessageBox.Show(
                    "Failed to set database version number.  Restore your database from a backup and try again.\n\rIf the problem persists contact the author.");
            }
        }

        /// <summary>
        /// loads partial data rows into a table in memory.  this
        /// table is used for fast local searches for information 
        /// about the current frequency.
        /// </summary>
        internal void LoadStationList()
        {
            try
            {
#if (LOG)
                Common.Log("enter");
#endif
                using (SQLiteConnection conn = Common.GetSqlConnection())
                {
                    lock (_stationTable)
                    {
                        //init database connection
                        using (var dta = new SQLiteDataAdapter("select frequency,description,service,id," +
                                                               "mode,notes,locked from data", conn))
                        {
                            if (_stationTable != null)
                            {
                                _stationTable.Dispose();
                                _stationTable = null;
                            }
                            _stationTable = new DataTable();
                            dta.Fill(_stationTable);
                        }
                    }
                }
#if (LOG)
                Common.Log("exit");
#endif
            }
            catch (Exception ex)
            {
                String msg = String.Format("An error occurred.  The error was: \n\r{0}\n\r{1}", ex.Message, ex.StackTrace.Trim());
                MessageBox.Show(msg, "Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
        }

        /// <summary>
        /// Determine the text from the selected mode button
        /// </summary>
        internal String GetCurrentMode()
        {
            return Enum.GetName(typeof(DetectorType), _controlInterface.DetectorType);
        }

        /// <summary>
        /// sets SDR# mode per database record
        /// </summary>
        /// <param name="modeText">String containing mode text</param>
        private void SetCurrentMode(String modeText)
        {
            if (_formClosing) return;
            try
            {
                _controlInterface.DetectorType = (DetectorType)Enum.Parse(typeof(DetectorType), modeText.Replace("-", String.Empty));
            }
            catch (Exception)
            {
                //swallow it
            }
        }

        /// <summary>
        /// displays the appinfo tooltip.  intercepts
        /// keypressess and looks for Ctrl+Alt+I.
        /// </summary>
        protected override bool ProcessCmdKey(ref Message msg, Keys keyData)
        {
            if ((msg.Msg == WmKeydown) || (msg.Msg == WmSyskeydown))
                switch (keyData)
                {
                    case Keys.Control | Keys.Alt | Keys.I:
                        string[] parms = new string[4];
                        parms[0] = String.Format("Application Version: {0}", GetType().Assembly.GetName().Version);
                        var fileInfo = new FileInfo(Assembly.GetExecutingAssembly().Location);
                        parms[1] = String.Format("Last Modified: {0}", fileInfo.LastWriteTime);
                        parms[2] = String.Format("Database: {0}\\{1}", Common.DbPath, Common.DbFileName);
                        try
                        {
                            using (SQLiteConnection conn = Common.GetSqlConnection())
                            {
                                const string sel = "select Rev from Version";
                                var cmd = new SQLiteCommand(sel, conn);
                                Int32 rev = Convert.ToInt32(cmd.ExecuteScalar());
                                parms[3] = String.Format("Database Version: {0}", rev);
                            }
                        }
                        catch (Exception)
                        {
                            parms[3] = "Database Version: Unable to read the database!";
                        }
                        String info =
                            String.Format("Frequency Manager + Scanner © 2012-2013 Jeff Knapp\n{0}\n{1}\n{2}\n{3}", parms);
                        try { _appInfo.Show(info, this, new Point(fmGroupBox.Width, 20), 5000); }
                        catch (Exception)
                        {
                            //swallow it
                        }
                        break;
                }
            return false;
        }

        /// <summary>
        /// extracts the current stepsize and configures the scan range controls.
        /// </summary>
        private void GetStepSize()
        {
            _stepSize = _controlInterface.StepSize;
            scanLowerFreqNud.Increment = _stepSize;
            scanUpperFreqNud.Increment = _stepSize;
        }

        /// <summary>
        /// when any control in the freqmgr panel is clicked, set windows
        /// focus to that control or to the first interactive control
        /// in the panel.  this enables the panel tooltip to
        /// be shown regardless of where the mouse was clicked in the panel.
        /// </summary>
        private void ForceFocus(object sender, EventArgs e)
        {
            Focus();
        }

        /// <summary>
        /// utility to select all the text in a numeric up/down control
        /// </summary>
        private static void NudSelectAll(object sender, EventArgs e)
        {
            var ctl = (NumericUpDown)sender;
            ctl.Select(0, ctl.ToString().Length);
        }

        /// <summary>
        /// gets text from the clipboard and sends it to the radio
        /// </summary>
        private void PasteFreqClick(object sender, EventArgs e)
        {
            //get data from the clipboard
            IDataObject iData = Clipboard.GetDataObject();
            //ignore non-text
            if (!iData.GetDataPresent(DataFormats.Text)) return;
            //cleanup the text
            String temp = (String)iData.GetData(DataFormats.Text);
            temp = temp.Trim();
            temp = temp.TrimStart('0');
            //try to convert the text into a number
            Decimal frq;
            Boolean result = Decimal.TryParse(temp, NumberStyles.Any, CultureInfo.CurrentCulture, out frq);
            if (!result) return;
            //normalize the number to mhz
            if (frq < 1000)
            {
                frq = frq * 1000000;
            }
            if (frq < 10000)
            {
                frq = frq * 1000000;
            }
            if (frq < 100000)
            {
                frq = frq * 10000;
            }
            if (frq < 1000000)
            {
                frq = frq * 1000;
            }
            _controlInterface.Frequency = (long)frq;
        }

        /// <summary>
        /// shows a tooltip over the paste button
        /// </summary>
        private void PasteFreqMouseHover(object sender, EventArgs e)
        {
            try { _appInfo.SetToolTip(pasteFreq, "Paste a Frequency into the Radio"); }
            catch (Exception)
            {
                //swallow it
            }
        }

        /// <summary>
        /// displays the manage menu
        /// </summary>
        private void ManageButtonClick(object sender, EventArgs e)
        {
            //show the manage menu
            Rectangle rect = manageButton.DisplayRectangle;
            manageMenu.Show((Control)sender, new Point(rect.Left + rect.Width, rect.Top));
        }

        #region FrequencyActivityEventArgs

        /// <summary>
        /// event argument for FrequencyActivity event
        /// </summary>
        public class FrequencyActivityEventArgs : EventArgs
        {
            public DateTime ActivityStart { get; set; }
            public DateTime ActivityEnd { get; set; }
            public String Frequency { get; set; }
            public Int32 Delay { get; set; }
        }

        #endregion

        #endregion

        #region Browse

        /// <summary>
        /// display the browse form non-modally
        /// </summary>
        /// <remarks>the browse button is actually a checkbox styled as a button</remarks>
        private void BrowseButtonCheckChanged(object sender, EventArgs e)
        {
            try
            {
                if (_browseForm == null)
                {
                    //create a browse form and position it
                    //to the current frequency
                    _browseForm = new Browse(_controlInterface.Frequency, this) { Icon = _mfIcon };
                    //connect events
                    _browseForm.FormClosing += BrowseFormFormClosing;
                    _browseForm.StationSelected += BrowseFormStationSelected;
                    _browseForm.StationEdited += BrowseFormStationEdited;
                    //show form
                    _browseWindowThread = new Thread(() => Application.Run(_browseForm));
                    _browseWindowThread.SetApartmentState(ApartmentState.STA);
                    _browseWindowThread.IsBackground = true;
                    _browseWindowThread.Start();
                }
                else
                {
                    //if the form exists, close and destroy it
                    _browseForm.Close();
                    _browseForm = null;
                    //I know Abort is the wrong usage but I couldn't find an alternative
                    try { _browseWindowThread.Abort(); }
                    catch (ThreadAbortException)
                    { Thread.ResetAbort(); }
                    catch (Exception)
                    { }
                }
            }
            catch (Exception ex)
            {
                MessageBox.Show(String.Format("An error occurred: {0}\n\r{1}", ex.Message, ex.StackTrace), "Error",
                                MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
        }

        /// <summary>
        /// handles StationEdited event from the browse window
        /// </summary>
        private void BrowseFormStationEdited(object sender, StationEditedEventArgs e)
        {
            if (e.Record == null)
            {
                //was a bulk edit, reload the entire table
                LoadStationList();
            }
            else
            {
                //data was added or changed; update FreqMgr but not the radio
                ImportRadioSettings(e.Record, false);
            }
        }

        /// <summary>
        /// processes the stationselected event from the browse form
        /// </summary>
        private void BrowseFormStationSelected(object sender, StationSelectedEventArgs e)
        {
            //data was added or changed; update FreqMgr and the radio
            ImportRadioSettings(e.Record, true);
        }

        /// <summary>
        /// processes the browse form closing event
        /// </summary>
        private void BrowseFormFormClosing(object sender, FormClosingEventArgs e)
        {
            browseButton.Checked = false;
            _browseForm = null;
            //I know Abort is the wrong usage but I couldn't find an alternative
            try { _browseWindowThread.Abort(); }
            catch (ThreadAbortException)
            { Thread.ResetAbort(); }
            catch (Exception)
            { }
        }

        #endregion

        #region Edit

        /// <summary>
        /// process the edit button click event
        /// </summary>        
        private void EditButtonClick(object sender, EventArgs e)
        {
            try
            {
                Boolean updating = false;
                //assuming we are adding a record, build the frequency data record for passage to the edit form
                var rec = new FreqRecord
                              {
                                  Id = 0,
                                  Frequency = _controlInterface.Frequency,
                                  Center = _controlInterface.CenterFrequency,
                                  Description = _currDescription,
                                  Mode = currentMode.Text,
                                  Protocol = _currProtocol.Length > 0 ? _currProtocol : Common.UnidText,
                                  Callsign = _currCallsign,
                                  Service = _currService.Length > 0 ? _currService : Common.UnidText,
                                  ShiftEnabled = Convert.ToInt32(_controlInterface.FrequencyShiftEnabled),
                                  Shift = _controlInterface.FrequencyShift,
                                  FilterBandwidth = _controlInterface.FilterBandwidth,
                                  FilterOrder = _controlInterface.FilterOrder,
                                  SquelchEnabled = Convert.ToInt32(_controlInterface.SquelchEnabled),
                                  Squelch = _controlInterface.SquelchThreshold,
                                  CWShift = _controlInterface.CWShift,
                                  Notes = _currNotes,
                                  Date = _currDate,
                                  Flagged = false,
                                  Locked = false
                              };
                var ft = (WindowType)Enum.Parse(typeof(WindowType), _controlInterface.FilterType.ToString());
                Int32 filterTypeSelectedIndex = (int)ft - 1;
                rec.FilterType = Common.FilterTypeArray[filterTypeSelectedIndex];
                //now see if this is already in the database
                String sel = String.Format("select * from Data where frequency = {0} and description = '{1}'", rec.Frequency, _currDescription.Replace("'", "''"));
                using (SQLiteConnection conn = Common.GetSqlConnection())
                using (var cmd = new SQLiteCommand(sel, conn))
                using (var da = new SQLiteDataAdapter(cmd))
                {
                    _ds = new DataSet();
                    da.Fill(_ds);
                }
                //if it is in the db use the db data instead of the initialized form data
                if (_ds.Tables[0].Rows.Count > 0)
                {
                    updating = true;
                    DataRow dr = _ds.Tables[0].Rows[0];
                    rec.Id = Convert.ToInt32(dr["Id"].ToString());
                    rec.Frequency = (long)dr["Frequency"];
                    rec.Center = (long)dr["Center"];
                    rec.Description = dr["Description"].ToString();
                    rec.Mode = dr["Mode"].ToString();
                    rec.Protocol = dr["Protocol"].ToString();
                    rec.Callsign = dr["Call Sign"].ToString();
                    rec.Service = dr["Service"].ToString();
                    rec.Notes = dr["Notes"].ToString();
                    rec.Date = String.IsNullOrEmpty(dr["Date"].ToString()) ? _currDate : dr["Date"].ToString();
                    rec.Locked = Convert.ToBoolean(dr["Locked"]);
                    rec.Flagged = Convert.ToBoolean(dr["Flagged"]);
                }
                String currMode = GetCurrentMode();

                //display the edit form
                using (var form = new Edit(rec, Common.FilterTypeArray, filterTypeSelectedIndex, currMode))
                {
                    //display the edit form modally and wait for return
                    if (form.ShowDialog() == DialogResult.OK)
                    {
                        //data was added or changed; update FreqMgr and the radio
                        rec = form.RecordData;
                        ImportRadioSettings(form.RecordData, true);
                        //update the browse form if it is open
                        if (_browseForm != null)
                        {
                            if (updating)
                            {
                                _browseForm.GridRowUpdate(rec, true);
                            }
                            else
                            {
                                _browseForm.GridRowAdd(rec, true);
                            }
                        }
                    }
                    //update scanfreqtable if necessary
                    if ((scanTypeTabControl.SelectedIndex == 0) && (_canScan))
                    {
                        LoadScanFreqTable();
                    }
                }
                //reload the internal station lookup table
                LoadStationList();
            }
            catch (Exception ex)
            {
                String msg = String.Format("An error occurred.  The error was: \n\r{0}",
                                           ex.Message + "\n\r" + ex.StackTrace.Trim());
                MessageBox.Show(msg, "Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
        }

        #endregion

        #region Groups

        /// <summary>
        /// display the group form non-modally
        /// </summary>
        private void EditGroups(object sender, EventArgs e)
        {
            if (_groupForm == null)
            {
                //create form
                _groupForm = new Groups { Icon = _mfIcon };
                //connect events
                _groupForm.FormClosing += GroupFormFormClosing;
                _groupForm.GroupsUpdated += GroupFormGroupsUpdated;
                //show form
                _groupWindowThread = new Thread(() => Application.Run(_groupForm));
                _groupWindowThread.SetApartmentState(ApartmentState.STA);
                _groupWindowThread.IsBackground = true;
                _groupWindowThread.Start();
            }
            else
            {
                //if the form exists, close and destroy it
                _groupForm.Close();
                _groupForm = null;
                //I know Abort is the wrong usage but I couldn't find an alternative
                try { _groupWindowThread.Abort(); }
                catch (ThreadAbortException)
                { Thread.ResetAbort(); }
                catch (Exception)
                { }
            }
        }

        /// <summary>
        /// sinks GroupsUpdated event from the Groups window
        /// </summary>
        private void GroupFormGroupsUpdated(object sender, FreqMgr.GroupsUpdatedEventArgs e)
        {
            try
            {
                //update the scanner's group list
                String currGroup = Common.NoneText;
                var x = (DataRowView)groupList.SelectedItem;
                if (x != null)
                {
                    currGroup = x.Row["Name"].ToString();
                }
                LoadScanGroupList();
                //restore selection
                groupList.SelectedIndex = groupList.FindStringExact(currGroup, 0);
                if (groupList.SelectedIndex == -1)
                {
                    groupList.SelectedIndex = groupList.FindStringExact(Common.NoneText, 0);
                    groupList.Text = Common.NoneText;
                }
                groupList.Text = currGroup;
                //update the browse window if it is open
                if (GroupsUpdated == null) return;
                GroupsUpdated(this, new GroupsUpdatedEventArgs());
            }
            catch (Exception)
            {
                //swallow it
            }
        }

        /// <summary>
        /// cleanup when group form closes
        /// </summary>
        private void GroupFormFormClosing(object sender, FormClosingEventArgs e)
        {
            _groupForm = null;
            //I know Abort is the wrong usage but I couldn't find an alternative
            try { _groupWindowThread.Abort(); }
            catch (ThreadAbortException)
            { Thread.ResetAbort(); }
            catch (Exception)
            { }
        }

        #endregion

        #region Protocols

        /// <summary>
        /// display the protocl form non-modally
        /// </summary>
        private void EditProtocols(object sender, EventArgs e)
        {
            if (_protocolForm == null)
            {
                //create form
                _protocolForm = new Protocols { Icon = _mfIcon };
                //connect events
                _protocolForm.FormClosing += ProtocolFormFormClosing;
                //show form
                _protocolWindowThread = new Thread(() => Application.Run(_protocolForm));
                _protocolWindowThread.SetApartmentState(ApartmentState.STA);
                _protocolWindowThread.IsBackground = true;
                _protocolWindowThread.Start();
            }
            else
            {
                //if the form exists, close and destroy it
                _protocolForm.Close();
                _protocolForm = null;
                //I know Abort is the wrong usage but I couldn't find an alternative
                try { _protocolWindowThread.Abort(); }
                catch (ThreadAbortException)
                { Thread.ResetAbort(); }
                catch (Exception)
                { }
            }
        }

        /// <summary>
        /// cleanup when Protocol form closes
        /// </summary>
        private void ProtocolFormFormClosing(object sender, FormClosingEventArgs e)
        {
            _protocolForm = null;
            //I know Abort is the wrong usage but I couldn't find an alternative
            try { _protocolWindowThread.Abort(); }
            catch (ThreadAbortException)
            { Thread.ResetAbort(); }
            catch (Exception)
            { }
        }

        #endregion

        #region Services

        /// <summary>
        /// display the services form non-modally
        /// </summary>
        private void EditServices(object sender, EventArgs e)
        {
            if (_serviceForm == null)
            {
                //create form
                _serviceForm = new Services { Icon = _mfIcon };
                //connect events
                _serviceForm.FormClosing += ServiceFormFormClosing;
                //show form
                _serviceWindowThread = new Thread(() => Application.Run(_serviceForm));
                _serviceWindowThread.SetApartmentState(ApartmentState.STA);
                _serviceWindowThread.IsBackground = true;
                _serviceWindowThread.Start();
            }
            else
            {
                //if the form exists, close and destroy it
                _serviceForm.Close();
                _serviceForm = null;
                //I know Abort is the wrong usage but I couldn't find an alternative
                try { _serviceWindowThread.Abort(); }
                catch (ThreadAbortException)
                { Thread.ResetAbort(); }
                catch (Exception)
                { }
            }
        }

        /// <summary>
        /// cleanup when Service form closes
        /// </summary>
        private void ServiceFormFormClosing(object sender, FormClosingEventArgs e)
        {
            _serviceForm = null;
            //I know Abort is the wrong usage but I couldn't find an alternative
            try { _serviceWindowThread.Abort(); }
            catch (ThreadAbortException)
            { Thread.ResetAbort(); }
            catch (Exception)
            { }
        }

        #endregion


        private void WriteLog(String txtToWrite, string filePath)
        {
            // Are the occasional crashs related to trying to write the file too fast?
            try
            {
                using (StreamWriter outfile = new StreamWriter(filePath))
                {
                    outfile.Write(txtToWrite.ToString());
                }
            }
            catch (Exception e)
            {
                DoNothing(e);
            }

        }

        private void DoNothing(Exception e)
        {
            // Does exactly as the function would imply...
        }

        private void writeSingleLog(string outstring = null)
        {
            if (singleLogFilePath.Text != null && this.loggingEnabled)
            {
                if (outstring == null)
                {
                    outstring = getLogText();
                }
                if (outstring != null)
                {
                    WriteLog(outstring, singleLogFilePath.Text);
                }
            }
        }

        #region Scanner

        /// <summary>
        /// handles check of the scan button; 
        /// note the scanButton is a checkbox styled as a button
        /// </summary>
        private void ScanButtonCheckedChanged(object sender, EventArgs e)
        {
            // Write the log when user stops scanning, since they are entering manual mode
            if (!this.scanButton.Checked)
            {
                loggerScanning = false;
                logTimer();
                writeSingleLog();
            }

            #region setup and validation

            _canScan = scanButton.Checked;
            scanTypeTabControl.Enabled = !_canScan;
            msPause.Enabled = !_canScan;
            secDelay.Enabled = !_canScan;
            _isScanningFreqRange = false;
            IsScanning = _canScan;
            if (!_canScan)
            {
                //stop all scaning activities
                _queueTimer.Stop();
                _scanBusyTestTimer.Stop();
                _scanInterFreqTimer.Stop();
                _scanTailTimer.Stop();
                //flush the queue            
                try
                {
                    if (!_queueMgr.IsBusy)
                    {
                        _queueMgr.RunWorkerAsync(_logFreqs);
                    }
                }
                catch (Exception)
                {
                    //swallow it
                }
                //set buttons
                scanSkipButton.Enabled = _canScan;
                scanLockOutButton.Enabled = _canScan;
                scanFlagButton.Enabled = _canScan;
                scanButton.Text = "Scan";
                //notify clients of state change
                if (ScannerStateChanged != null)
                {
                    _ssc = new ScannerStateChangedEventArgs { IsScanning = _canScan };
                    ScannerStateChanged(this, _ssc);
                }
                return;
            }
            //we can scan at this point.  continue with validations and startup.
            //validate frequency range
            if (scanTypeTabControl.SelectedIndex == 1)
            {
                if ((int)scanUpperFreqNud.Value <= (int)scanLowerFreqNud.Value)
                {
                    scanButton.Checked = false;
                    _canScan = false;
                    MessageBox.Show("Lower Frequency must be less than Upper Frequency", "Cannot Scan",
                                    MessageBoxButtons.OK, MessageBoxIcon.Error);
                    return;
                }
            }
            //save the current step size
            GetStepSize();

            #endregion

            //determine type of scan
            if (scanTypeTabControl.SelectedIndex == 0)
            {
                #region scan by group

                //make sure we have a scan group chosen
                Int32 index = groupList.FindStringExact(Common.NoneText, 0);
                if (groupList.SelectedIndex == index)
                {
                    scanButton.Checked = false;
                    _canScan = false;
                    MessageBox.Show("Select a Scan Group first", "Cannot Scan", MessageBoxButtons.OK,
                                    MessageBoxIcon.Error);
                    return;
                }
                //delete existing data from the internal scan frequency table
                _scanFreqTable.Rows.Clear();
                //load the internal scan frequency table
                LoadScanFreqTable();
                if (_scanFreqTable.Rows.Count == 0)
                {
                    scanButton.Checked = false;
                    _canScan = false;
                    MessageBox.Show("That group has no frequencies, or has no unlocked frequencies", "Cannot Scan",
                                    MessageBoxButtons.OK, MessageBoxIcon.Error);
                    _ds = null;
                    return;
                }
                _scanFreqPtr = -1;

                #endregion
            }
            else
            {
                #region scan by frequency

                _isScanningFreqRange = true;

                if (Utils.GetBooleanSetting(NormalizeScanFreqsKey))
                {
                    //normalize frequencies by rounding down and up
                    scanLowerFreqNud.Value =
                        (Decimal)(Common.RoundDownFreq((Double)scanLowerFreqNud.Value / _stepSize, 0) * _stepSize);
                    scanUpperFreqNud.Value =
                        (Decimal)(Common.RoundUpFreq((Double)scanUpperFreqNud.Value / _stepSize, 0) * _stepSize);
                }
                _lowFreq = (long)scanLowerFreqNud.Value;
                _hiFreq = (long)scanUpperFreqNud.Value;
                //initialize the center freq
                try
                {
                    //decrement for the first freq only to make it on the target when it starts
                    _controlInterface.CenterFrequency = _lowFreq - _stepSize;
                    Thread.Sleep(1);
                }
                catch (ApplicationException ex)
                {
                    //this happens when the user tries to scan a sound card.
                    //we cannot test for this source when the scan button is pressed;
                    //SDR# doesn't tell us what the source is.
                    if (ex.Message.Contains("no front end is connected"))
                    {
                        scanButton.Checked = false;
                        _canScan = false;
                        MessageBox.Show("You cannot scan that Source", "Error", MessageBoxButtons.OK,
                                        MessageBoxIcon.Exclamation);
                        return;
                    }
                    throw;
                }
                //ok to continue!
                _controlInterface.Frequency = _controlInterface.CenterFrequency;
#if (LOG)
                {
                    Common.Log("=================================================");
                    Common.Log("starting freq range scan ");
                    Common.Log("center = " + _controlInterface.CenterFrequency);
                    Common.Log("freq = " + _controlInterface.Frequency);
                }
#endif
                #endregion
            }

            //start scanning!
            //set the timers
            if (secDelay.Value == 0)
            {
                _scanTailTimer.Interval = 10;
                //timer intervals cannot be zero, so set it to a token amount of .01 seconds
            }
            else
            {
                _scanTailTimer.Interval = Convert.ToDouble(secDelay.Value) * 1000;
            }
            _scanInterFreqTimer.Interval = Convert.ToDouble(msPause.Value);
            _queueTimer.Enabled = true;
            scanButton.Text = "Stop";
            //notify clients of state change
            if (ScannerStateChanged != null)
            {
                _ssc = new ScannerStateChangedEventArgs { IsScanning = _canScan };
                ScannerStateChanged(this, _ssc);
            }
            //start the scanner
            if (!_controlInterface.IsPlaying)
            {
                _controlInterface.StartRadio();
                Thread.Sleep(1000); //the radio sometimes crashes unless we do this to give it time to settle
            }
            //set buttons
            scanSkipButton.Enabled = _canScan;
            scanLockOutButton.Enabled = _canScan;
            scanFlagButton.Enabled = _canScan;
            //get the first frequency
            DoScan();
        }

        /// <summary>
        /// loads the scan freq table with the freqs that are in the chosen group
        /// </summary>
        private void LoadScanFreqTable()
        {
            try
            {
#if (LOG)
                Common.Log("enter");
#endif
                const String selGroup = "select d.Id, d.Frequency as Frequency, d.Center as Center, " +
                                        "d.Mode as Mode, d.[Filter Type] as FilterType, d.Shift as Shift, " +
                                        "d.[Shift Enabled] as ShiftEnabled, d.Description as Description, d.Service as Service, " +
                                        "d.[Filter Bandwidth] as FilterBandwidth, d.[Filter Order] as FilterOrder, " +
                                        "d.Squelch as Squelch, d.[Squelch Enabled] as SquelchEnabled, d.Notes as Notes, " +
                                        "d.[CW Shift] as CWShift from data as d join groupfreq as g " +
                                        "on d.id = g.freqid where d.Locked = 0 " +
                                        "and g.groupid = @groupId order by Frequency";
                using (SQLiteConnection conn = Common.GetSqlConnection())
                using (var cmd = new SQLiteCommand(selGroup, conn))
                using (var da = new SQLiteDataAdapter(cmd))
                {
                    cmd.Parameters.AddWithValue("@groupId", groupList.SelectedValue);
                    lock (_scanFreqTable)
                    {
                        _scanFreqTable.Rows.Clear();
                        da.Fill(_scanFreqTable);
                        _scanFreqTable.DefaultView.Sort = "Frequency ASC";
                    }
                }
#if (LOG)
                Common.Log("exit");
#endif
            }
            catch (Exception ex)
            {
                String msg = String.Format("An error occurred.  The error was: \n\r{0}",
                                           ex.Message + "\n\r" + ex.StackTrace.Trim());
                MessageBox.Show(msg, "Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
        }

        /// <summary>
        /// timer performing the delay for pausing on a frequency to check if it it transmitting.
        /// if it is transmitting, the busy test timer is started, otherwise DoScan is called to
        /// jump to the next frequency.
        /// </summary>
        private void ScanInterFreqTimerElapsed(object sender, ElapsedEventArgs e)
        {
#if (LOG)
            Common.Log("enter");
#endif
            try
            {
                //if freq is in use
                if (_controlInterface.IsSquelchOpen)
                {
                    _haveActivity = true;
                    scannerLog();
                    //initialize an event argument for consumers
                    _fae = new FrequencyActivityEventArgs
                               {
                                   ActivityStart = DateTime.Now,
                                   ActivityEnd = DateTime.Now,
                                   Frequency = currentFreq.Text,
                                   Delay = Convert.ToInt32(secDelay.Text)
                               };
                    //if record button is checked and scan freqs, then put freq in the queue
                    if ((scanTypeTabControl.SelectedIndex == 1) && (recordButton.Checked))
                    {
                        EnqueueFreqRecord(false, false);
                    }
                    _scanBusyTestTimer.Start();
                }
                else
                {
                    //freq is not active, so move on
                    DoScan();
                }
            }
            catch (Exception ex)
            {
                MessageBox.Show(
                    String.Format("Error in ScanInterFreqTimerElapsed: {0}\n\r{1}", ex.Message, ex.StackTrace), "Error",
                    MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
#if (LOG)
            Common.Log("exit");
#endif
        }

        /// <summary>
        /// puts a new freqrecord into the scanning queue for later processing
        /// </summary>
        private void EnqueueFreqRecord(Boolean flagged, Boolean locked)
        {
            //add the event to the queue
            var ft = (WindowType)Enum.Parse(typeof(WindowType), _controlInterface.FilterType.ToString());
            Int32 filterTypeSelectedIndex = (int)ft - 1;
            _logFreqs.Enqueue(new FreqRecord
                                  {
                                      Frequency = _controlInterface.Frequency,
                                      Center = _controlInterface.CenterFrequency,
                                      Description = _scanUnidFreqFound,
                                      Mode =
                                          ((DetectorType)
                                           Enum.Parse(typeof(DetectorType), _controlInterface.DetectorType.ToString()))
                                          .
                                          ToString(),
                                      Protocol = Common.UnidText,
                                      Callsign = String.Empty,
                                      Service = Common.UnidText,
                                      ShiftEnabled = Convert.ToInt32(_controlInterface.FrequencyShiftEnabled),
                                      Shift = _controlInterface.FrequencyShift,
                                      FilterBandwidth = _controlInterface.FilterBandwidth,
                                      FilterOrder = _controlInterface.FilterOrder,
                                      FilterType = Common.FilterTypeArray[filterTypeSelectedIndex],
                                      SquelchEnabled = Convert.ToInt32(_controlInterface.SquelchEnabled),
                                      Squelch = _controlInterface.SquelchThreshold,
                                      CWShift = _controlInterface.CWShift,
                                      Notes = String.Empty,
                                      Date = DateTime.Now.ToString(Common.DateSaveFormat, CultureInfo.InvariantCulture),
                                      Flagged = flagged,
                                      Locked = locked
                                  });
        }

        /// <summary>
        /// timer that pauses after a tranmission ends, to see if another transmission starts on the same freq.
        /// if the frequency is still active, the busy test timer is started, otherwise DoScan is called to
        /// jump to the next frequency.
        /// </summary>
        private void ScanTailTimerElapsed(object sender, ElapsedEventArgs e)
        {
#if (LOG)
            Common.Log("enter");
#endif
            if (_controlInterface.IsSquelchOpen)
            {
                _scanBusyTestTimer.Start();
            }
            else
            {
                DoScan();
            }
#if (LOG)
            Common.Log("exit");
#endif
        }

        /// <summary>
        /// timer that pauses for a short delay, after which the frequency is checked to see
        /// if it is still in use.  If so, the timer runs again; otherwise the transmission
        /// has ended and the tail timer is started to check for a response transmission.
        /// </summary>
        private void ScanBusyTestTimerElapsed(object sender, ElapsedEventArgs e)
        {
#if (LOG)
            Common.Log("enter");
#endif
            if (_controlInterface.IsSquelchOpen)
            {
                _scanBusyTestTimer.Start();
            }
            else
            {
                if (secDelay.Value > 0)
                {
                    _scanTailTimer.Start();
                }
                else
                {
                    DoScan();
                }
            }
#if (LOG)
            Common.Log("exit");
#endif
        }

        /// <summary>
        /// saves the value of the msPause numeric updown to config
        /// </summary>
        private void MsPauseValueChanged(object sender, EventArgs e)
        {
            lock (msPause)
            {
                Utils.SaveSetting("freqMgrScanMsPause", msPause.Value);
            }
        }

        /// <summary>
        /// saves the value of the secDelay numeric updown to config
        /// </summary>
        private void SecDelayValueChanged(object sender, EventArgs e)
        {
            lock (secDelay)
            {
                Utils.SaveSetting("freqMgrScanSecDelay", secDelay.Value);
            }
        }

        /// <summary>
        /// skips the current locked-on scanner freq
        /// </summary>
        private void ScanSkipButtonClick(object sender, EventArgs e)
        {
            _scanTailTimer.Stop();
            DoScan();
        }

        /// <summary>
        /// primary scanning method.  gets the next frequency,
        /// configures the radio, and starts the control timers.
        /// </summary>
        private void DoScan()
        {
            if ((!_canScan) || (!_controlInterface.IsPlaying) || (_formClosing)) return;
            try
            {
                //stop timers
                _scanTailTimer.Stop();
                _scanInterFreqTimer.Stop();
                _scanBusyTestTimer.Stop();

                Boolean listenToFreq = false;
                scannerLog();
#if (LOG)
                {
                    Common.Log("enter");
                }
#endif
                //scan by group
                if (scanTypeTabControl.SelectedIndex == 0)
                {
                    Boolean goodRecord = false;
                    while (!goodRecord)
                    {
                        //get next freq and program the radio
                        DataRow f = ScanGetNextStation();
                        try
                        {
                            SetCurrentMode(f["Mode"].ToString());
                            _controlInterface.FilterType =
                                (WindowType)
                                (Array.IndexOf(Common.FilterTypeArray, f["FilterType"].ToString()) + 1);
                            _controlInterface.FrequencyShift = (long)f["Shift"];
                            _controlInterface.FrequencyShiftEnabled =
                                Convert.ToBoolean((Int32)f["ShiftEnabled"]);
                            _controlInterface.FilterOrder = (Int32)f["FilterOrder"];
                            _controlInterface.SquelchThreshold = (Int32)f["Squelch"];
                            _controlInterface.SquelchEnabled = Convert.ToBoolean((Int32)f["SquelchEnabled"]);
                            _controlInterface.CWShift = (Int32)f["CWShift"];
                            _controlInterface.CenterFrequency = (long)f["Center"];
                            Thread.Sleep(1);
                            _controlInterface.Frequency = (long)f["Frequency"];
#if (LOG)
{
    Common.Log("group set freq " + f["Frequency"]);
}
#endif
                            //FilterBandwidth must be set AFTER the mode is changed
                            _controlInterface.FilterBandwidth = (Int32)f["FilterBandwidth"];
                            goodRecord = true;
                        }
                        catch (RowNotInTableException)
                        {
                            //it can happen that a user can flag, edit, or lock a frequency
                            //during a group scan.  That can cause ScanGetNextStation() to return
                            //an invalid record.  When that happens we need to reload the
                            //scan table.
                            LoadScanFreqTable();
                        }
                        catch (ApplicationException ex)
                        {
                            //this happens when the user tries to scan a sound card.
                            //we cannot test for this source when the scan button is pressed;
                            //SDR# doesn't tell us what the source is.
                            if (ex.Message.Contains("no front end is connected"))
                            {
                                scanButton.Checked = false;
                                _canScan = false;
                                MessageBox.Show("You cannot scan that Source", "Error", MessageBoxButtons.OK,
                                                MessageBoxIcon.Exclamation);
                                return;
                            }
                            throw;
                        }
                        catch (InvalidOperationException)
                        {
                            //swallow it.  this one sometimes happens when sdr#
                            //is closed while scanning at a high rate of speed.
                        }
                    }
                }
                else //scan by frequency
                {
                    long newFreq = _controlInterface.Frequency + _stepSize;
                    while (!listenToFreq)
                    {
                        //restart at lower limit if we are past the upper limit
                        if (newFreq > _hiFreq)
                        {
                            newFreq = _lowFreq;
                        }
                        const String sel = "frequency = {0}";
                        DataRow[] result = StationTableSelect(String.Format(sel, newFreq), false);
                        //if freq not found, or if freq found and IS NOT locked out, listen to it
                        if ((result.Length == 0) ||
                            ((result.Length > 0) && (!Convert.ToBoolean(result[0]["Locked"]))))
                        {
                            listenToFreq = true;
                        }
                        //if freq found and IS locked out, skip it and get the next freq
                        if ((result.Length > 0) && (Convert.ToBoolean(result[0]["Locked"])))
                        {
                            newFreq += _stepSize;
                        }
                    }
#if (LOG)
    {
        Common.Log("setting center and freq to " + newCenter);
    }
#endif

                    //set new freq
                    _controlInterface.Frequency = newFreq;
                }
                //start control timers as needed
                if ((_controlInterface.IsSquelchOpen) && (listenToFreq))
                {
                    _scanBusyTestTimer.Start();
                }
                else
                {
                    //update clients with info partially prepared earlier
                    if ((FrequencyActivity != null) && (_haveActivity))
                    {
                        _fae.ActivityEnd = DateTime.Now;
                        FrequencyActivity(this, _fae);
                        _haveActivity = false;
                    }
                    //start the inter-freq timer
                    _scanInterFreqTimer.Start();
                }
            }
            catch (Exception ex)
            {
                MessageBox.Show(String.Format("Error in DoScan: {0}\n\r{1}", ex.Message, ex.StackTrace), "Error",
                                MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
#if (LOG)
{
    Common.Log("exit");
}
#endif
        }

        /// <summary>
        /// returns the next frequency in the chosen group
        /// </summary>
        /// <returns>DataRow containing frequency and center</returns>
        private DataRow ScanGetNextStation()
        {
#if (LOG)
            Common.Log("enter");
#endif
            _scanFreqPtr += 1;
            if (_scanFreqPtr >= _scanFreqTable.Rows.Count) _scanFreqPtr = 0;

            DataRow dr = null;
            try
            {
                lock (_scanFreqTable)
                {
                    dr = _scanFreqTable.Rows[_scanFreqPtr];
                }
            }
            catch (RowNotInTableException)
            {
                //swallow it
            }
#if (LOG)
            Common.Log("exit");
#endif
            return dr;
        }

        /// <summary>
        /// Loads the group list dropdown from the database
        /// </summary>
        private void LoadScanGroupList()
        {
#if (LOG)
            Common.Log("enter");
#endif
            DataView dv = Common.GetGroupList(true);
            if (dv.Table.Rows.Count <= 1) return;
            String currGroup = String.Empty;
            if (groupList.Items.Count > 0)
            {
                var x = (DataRowView)groupList.SelectedItem;
                currGroup = x != null ? x.Row["Name"].ToString() : Common.NoneText;
            }
            groupList.DataSource = dv;
            groupList.DisplayMember = "Name";
            groupList.ValueMember = "Id";
            Int32 found = groupList.FindStringExact(currGroup, 0);
            if (found > -1)
            {
                groupList.SelectedIndex = found;
            }
            else
            {
                groupList.SelectedIndex = groupList.FindStringExact(Common.NoneText, 0);
                scanButton.Checked = false;
            }
#if (LOG)
            Common.Log("exit");
#endif
        }

        /// <summary>
        /// adds a flag to the current freq
        /// </summary>
        private void ScanFlagButtonClick(object sender, EventArgs e)
        {
            EnqueueFreqRecord(true, false);
        }

        /// <summary>
        /// locks the current freq from scanning
        /// </summary>
        private void ScanLockOutButtonClick(object sender, EventArgs e)
        {
            EnqueueFreqRecord(false, true);
            ScanSkipButtonClick(new object(), new EventArgs());
        }

        #region ScannerStateChangedEventArgs

        /// <summary>
        /// event argument for ScannerStateChanged event
        /// </summary>
        public class ScannerStateChangedEventArgs : EventArgs
        {
            public Boolean IsScanning { get; set; }
        }

        #endregion

        #region GroupsUpdatedEventArgs

        /// <summary>
        /// event argument for GroupUpdated event
        /// </summary>
        public class GroupsUpdatedEventArgs : EventArgs
        {
            //this space intentionally left blank
        }

        #endregion

        #region ProtocolsUpdatedEventArgs

        /// <summary>
        /// event argument for ProtocolUpdated event
        /// </summary>
        public class ProtocolsUpdatedEventArgs : EventArgs
        {
            //this space intentionally left blank
        }

        #endregion

        #endregion

        #region Scanner Queue

        /// <summary>
        /// processes items in the queue.
        /// this is mostly used when scanning frequency ranges
        /// to add/update the db and the browse window records.
        /// </summary>
        private void QueueMgrDoWork(object sender, DoWorkEventArgs e)
        {
            if (_formClosing) return;
            try
            {
#if (LOG)
                Common.Log("enter");
#endif
                var activities = (Queue)e.Argument;
                Int32 activityCount = activities.Count;
                const String ins =
                    "insert into Data ([Call Sign],Center,[CW Shift],Date,Description,[Filter Bandwidth]," +
                    "[Filter Order],[Filter Type],Frequency,Mode,Notes,Protocol,Service," +
                    "[Shift Enabled],Shift,[Squelch Enabled],Squelch,Locked,Flagged)" +
                    " values (@callsign,@center,@cwshift,@date,@description,@filterbandwidth," +
                    "@filterorder,@filtertype,@frequency,@mode,@notes,@protocol,@service,@shiftenabled," +
                    "@shift,@squelchenabled,@squelch,@locked,@flagged)";
                const String upd = "update Data set Flagged = @flagged, Locked = @locked where Id = @id";
                while (activities.Count > 0)
                {
                    var fr = (FreqRecord)activities.Dequeue();
                    long lastFreq = fr.Frequency;
                    //search the in-memory table                    
                    String sel = String.Format("Frequency = {0} and Mode = '{1}'", fr.Frequency, fr.Mode);
                    DataRow[] result = StationTableSelect(sel, false);
                    using (SQLiteConnection conn = Common.GetSqlConnection())
                    using (SQLiteTransaction sqlTransaction = conn.BeginTransaction())
                    {
                        if (result.Length == 0)
                        {
                            #region insert new record

                            //don't know this one so add it to the db
                            var cmd = new SQLiteCommand(ins, conn);
                            cmd.Parameters.AddWithValue("@callsign", fr.Callsign);
                            cmd.Parameters.AddWithValue("@center", fr.Center);
                            cmd.Parameters.AddWithValue("@cwshift", fr.CWShift);
                            cmd.Parameters.AddWithValue("@date", fr.Date);
                            cmd.Parameters.AddWithValue("@description", fr.Description);
                            cmd.Parameters.AddWithValue("@filterbandwidth", fr.FilterBandwidth);
                            cmd.Parameters.AddWithValue("@filterorder", fr.FilterOrder);
                            cmd.Parameters.AddWithValue("@filtertype", fr.FilterType);
                            cmd.Parameters.AddWithValue("@frequency", fr.Frequency);
                            cmd.Parameters.AddWithValue("@mode", fr.Mode);
                            cmd.Parameters.AddWithValue("@notes", fr.Notes);
                            cmd.Parameters.AddWithValue("@protocol", fr.Protocol);
                            cmd.Parameters.AddWithValue("@service", fr.Service);
                            cmd.Parameters.AddWithValue("@shiftenabled", fr.ShiftEnabled);
                            cmd.Parameters.AddWithValue("@shift", fr.Shift);
                            cmd.Parameters.AddWithValue("@squelchenabled", fr.SquelchEnabled);
                            cmd.Parameters.AddWithValue("@squelch", fr.Squelch);
                            cmd.Parameters.AddWithValue("@locked", Convert.ToInt32(fr.Locked));
                            cmd.Parameters.AddWithValue("@flagged", Convert.ToInt32(fr.Flagged));
                            //insert record into database
                            cmd.ExecuteNonQuery();
                            //get its recordId
                            long rowId = 0;
                            const string getrow = "select last_insert_rowid()";
                            cmd = new SQLiteCommand(getrow, conn);
                            SQLiteDataReader reader = cmd.ExecuteReader();
                            while (reader.Read())
                            {
                                rowId = (long)reader[0];
                            }
                            reader.Dispose();
                            if (rowId > 0)
                            {
                                //it worked; update the internal data
                                fr.Id = (Int32)rowId;
                                //update station table
                                DataRow dr = _stationTable.NewRow();
                                dr["Frequency"] = fr.Frequency;
                                dr["Description"] = fr.Description;
                                dr["Service"] = fr.Service;
                                dr["Id"] = fr.Id;
                                dr["Mode"] = fr.Mode;
                                dr["Notes"] = fr.Notes;
                                dr["Locked"] = fr.Locked;
                                lock (_stationTable)
                                {
                                    _stationTable.Rows.Add(dr);
                                    _stationTable.AcceptChanges();
                                }
                                //update browse
                                if ((_browseForm != null) && (lastFreq > 0))
                                {
                                    _browseForm.GridRowAdd(fr, false);
                                }
                            }

                            #endregion
                        }
                        else
                        {
                            #region update existing records

                            var cmd = new SQLiteCommand(upd, conn);
                            foreach (DataRow dataRow in result)
                            {
                                if ((!fr.Locked) && (!fr.Flagged)) continue;
                                cmd.Parameters.Clear();
                                cmd.Parameters.AddWithValue("@id", dataRow["id"]);
                                cmd.Parameters.AddWithValue("@locked", Convert.ToInt32(fr.Locked));
                                cmd.Parameters.AddWithValue("@flagged", Convert.ToInt32(fr.Flagged));
                                //update record
                                cmd.ExecuteNonQuery();
                                //update station table
                                dataRow["Locked"] = fr.Locked;
                                //update browsegrid
                                if (_browseForm != null)
                                {
                                    _browseForm.SetFreqFlags(fr.Frequency, Convert.ToInt64(fr.Locked),
                                                            Convert.ToInt64(fr.Flagged));
                                }
                            }
                            _stationTable.AcceptChanges();

                            #endregion
                        }
                        sqlTransaction.Commit();
                    }
                }
                //reload the memory table if needed
                if (activityCount <= 0) return;
                //update scanfreqtable if necessary
                if ((scanTypeTabControl.SelectedIndex == 0) && (_canScan))
                {
                    LoadScanFreqTable();
                }
#if (LOG)
                Common.Log("exit");
#endif
            }
            catch (Exception ex)
            {
                MessageBox.Show(String.Format("Error in QueueMgrDoWork: {0}\n\r{1}", ex.Message, ex.StackTrace), "Error",
                                MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
        }

        /// <summary>
        /// executes a select query against the station tables
        /// and returns an array of DataRows
        /// </summary>
        private DataRow[] StationTableSelect(string sel, Boolean isGroup)
        {
            DataRow[] result = null;
            Boolean goodQuery = false;
            //we loop here to ensure we get a result,
            //and we don't skip over the process due to 
            //Exceptions caused by collisions
            while (!goodQuery)
            {
                try
                {
                    if (isGroup) //filter the _scanFreqTable table
                    {
                        lock (_scanFreqTable)
                        {
                            result = _scanFreqTable.Select(sel);
                            goodQuery = true; //if we were able to query the table it's good even if there's no data
                        }
                    }
                    else //filter the entire frequency data table
                    {
                        lock (_stationTable)
                        {
                            result = _stationTable.Select(sel);
                            goodQuery = true; //if we were able to query the table it's good even if there's no data
                        }
                    }
                }
                catch (Exception)
                {
                }
            }
            return result;
        }

        /// <summary>
        /// runs the queue worker when the timer expires
        /// </summary>
        private void QueueTimerElapsed(object sender, ElapsedEventArgs e)
        {
#if (LOG)
            Common.Log("enter");
#endif
            if (_formClosing) return;
            try
            {
                if (!_queueMgr.IsBusy)
                {
                    _queueMgr.RunWorkerAsync(_logFreqs);
                }
            }
            catch (Exception ex)
            {
                MessageBox.Show(String.Format("Error in QueueTimerElapsed: {0}\n\r{1}", ex.Message, ex.StackTrace),
                                "Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
#if (LOG)
            Common.Log("exit");
#endif
        }

        /// <summary>
        /// restarts the queue timer when the queue is empty
        /// </summary>
        private void QueueMgrRunWorkerCompleted(object sender, RunWorkerCompletedEventArgs e)
        {
#if (LOG)
            Common.Log("enter");
#endif
            if (_formClosing) return;
            try
            {
                if (_canScan)
                {
                    _queueTimer.Start();
                }
            }
            catch (Exception ex)
            {
                MessageBox.Show(
                    String.Format("Error in QueueMgrRunWorkerCompleted: {0}\n\r{1}", ex.Message, ex.StackTrace), "Error",
                    MessageBoxButtons.OK, MessageBoxIcon.Error);
            }

#if (LOG)
            Common.Log("exit");
#endif
        }

        /// <summary>
        /// starts and stops the recording of freqs to the database
        /// </summary>
        private void RecordButtonCheckedChanged(object sender, EventArgs e)
        {
            if (!recordButton.Checked)
            {
                _queueTimer.Stop();
                if (!_queueMgr.IsBusy)
                {
                    _queueMgr.RunWorkerAsync(_logFreqs); //flush the queue
                }
            }
            else
            {
                if (_canScan && scanTypeTabControl.SelectedIndex == 1)
                {
                    _queueTimer.Start();
                }
            }
        }

        #endregion

        private void singleLogFileEnabled_CheckedChanged(object sender, EventArgs e)
        {
            if (loggingConfigured && singleLogFilePath.Text != null)
            {

                if (singleLogFileEnabled.Checked)
                {
                    // Enable the logger  
                    this.logFileBrowseButton.Enabled = false;
                    this.loggingEnabled = true;
                    // start our timer if we aren't scanning. the timer will self destruct if the user starts to scan anyway.
                    if (!this.scanButton.Checked)
                        logTimer();

                }
                else
                {
                    // Stop logging
                    this.logFileBrowseButton.Enabled = true;
                    this.loggingEnabled = false;
                }
                WriteConfig();
            }
            else
            {
                if (loggingSystemMessageDisplayed == false)
                {
                    MessageBox.Show("Please set the log file location.", "Where do I write the log?", MessageBoxButtons.OK, MessageBoxIcon.Error);
                    loggingSystemMessageDisplayed = true; // this prevents a double message
                    singleLogFileEnabled.Checked = false;
                }
                else
                {
                    // If the user clicks it again, we will display the error again.
                    loggingSystemMessageDisplayed = false;
                }
            }
        }

        private void logFileBrowseButton_Click(object sender, EventArgs e)
        {
            SaveFileDialog fileDialog = new SaveFileDialog();
            fileDialog.FileName = "single.log";
            fileDialog.DefaultExt = "log";
            fileDialog.Filter = "Log File|*.log|Comma Seperated Value File|*.csv|Standard Text File|*.txt";
            if (fileDialog.ShowDialog() == DialogResult.OK)
            {
                singleLogFilePath.Text = fileDialog.FileName;
                loggingConfigured = true;
                WriteConfig();
            }
        }

        private void currentDescription_TextChanged(object sender, EventArgs e)
        {
            if (singleLogFileEnabled.Checked && !this.scanButton.Checked)
            {

            }
        }

        private void logTimer()
        {

            if (this._logTimer == null)
            {
                this._logTimer = new System.Timers.Timer(250);
                this._logTimer.Elapsed += new System.Timers.ElapsedEventHandler(_logTimer_Elapsed);
                this._logTimer.Start();
            }
        }

        void _logTimer_Elapsed(object sender, System.Timers.ElapsedEventArgs e)
        {
            lock (this._logTimer)
            {
                if (singleLogFileEnabled.Checked && !this.scanButton.Checked)
                {
                    loggerScanning = false;
                    if (getLogText() != lastlogtext)
                    {
                        lastlogtext = getLogText();
                        writeSingleLog(lastlogtext);
                    }
                }
                else
                {
                    // destroy timer if we are scanning
                    _logTimer.Stop();
                    _logTimer.Dispose();
                    _logTimer = null;
                }
            }
        }

        private void scannerLog()
        {
            if (this.scanButton.Checked)
            {
                if (!_controlInterface.IsSquelchOpen)
                {
                    // we dont need to write the 'Scanning...' string each time we change frequencies...
                    // ssd users will appreicate this hack
                    if (!loggerScanning)
                    {
                        loggerScanning = true;
                        string scanstring = "Scanning...";
                        if (lastcall != null)
                            scanstring += " (Last Call: "+lastcall+")";
                        writeSingleLog(scanstring);
                    }
                }
                else
                {
                    loggerScanning = false;
                    lastcall = getLogText();
                    writeSingleLog();
                }
            }
        }

        private string getLogText()
        {
            double curFreq = _controlInterface.Frequency;
            double roundedFreq = Math.Round((curFreq / 1000000), 4);
            return this._currDescription + " (" + roundedFreq.ToString() + "mHz)";
        }
    }

    #region NativeTabControl class
    /// <summary>
    /// custom tab page border-adjuster class, used to hide/minimize the ugly white border around tab pages
    /// </summary>
    /// <remarks>From http://stackoverflow.com/questions/2567172/c-sharp-tabcontrol-border-controls </remarks>
    class NativeTabControl : NativeWindow
    {
        protected override void WndProc(ref Message m)
        {
            if ((m.Msg == TcmAdjustrect))
            {
                Rect rc = (Rect)m.GetLParam(typeof(Rect));
                //Adjust these values to suit, dependant upon Appearance
                rc.Left -= 2;
                rc.Right += 1;
                rc.Top -= 1;
                rc.Bottom += 2;
                Marshal.StructureToPtr(rc, m.LParam, true);
            }
            base.WndProc(ref m);
        }

        private const Int32 TcmFirst = 0x1300;
        private const Int32 TcmAdjustrect = (TcmFirst + 40);
        private struct Rect
        {
            public Int32 Left;
            public Int32 Top;
            public Int32 Right;
            public Int32 Bottom;
        }
    }
    #endregion
}

