﻿//  ******************************************
//  Frequency Manager Services Window
//  Copyright © Jeff Knapp 2013 
//
//  This window displays the Services grid and allows
//  a user to manage service definitions.
//  ******************************************
//  Date        Changes
//  01/31/13    initial release
//  02/17/13    disabled minimize window button
//  ******************************************

using System;
using System.ComponentModel;
using System.Data;
using System.Data.SQLite;
using System.Drawing;
using System.Windows.Forms;

namespace SDRSharp.FreqMgr
{
    public sealed partial class Services : Form
    {
        #region local vars

        private const String Copyright = "Copyright © Jeff Knapp 2013";
        private const String ConfigKey = "freqMgrServicesWindow";
        private readonly SQLiteDataAdapter _dataAdapter;
        private readonly DataTable _dataTable;

        #endregion

        /// <summary>
        /// constructor
        /// </summary>
        public Services()
        {
            InitializeComponent();
            //format window
            Size sz;
            Point lo;
            Common.GetWindowSettings(ConfigKey, out sz, out lo);
            //get window size
            if (!sz.IsEmpty)
            {
                Size = sz;
            }
            //get window position
            if (!lo.IsEmpty)
            {
                Location = lo;
            }

            const string query = "select * from Service order by Name";
            _dataAdapter = new SQLiteDataAdapter(query, Common.GetSqlConnection());
            SQLiteCommandBuilder cBuilder = new SQLiteCommandBuilder(_dataAdapter);
            _dataTable = new DataTable();
            _dataAdapter.Fill(_dataTable);
            BindingSource bSource = new BindingSource {DataSource = _dataTable};
            dgv.DataSource = bSource;

            dgv.Columns["Name"].SortMode = DataGridViewColumnSortMode.Automatic;
            dgv.Sort(dgv.Columns["Name"], ListSortDirection.Ascending);
            dgv.Columns["Name"].HeaderCell.SortGlyphDirection = SortOrder.Ascending;

            SizeChanged += ServicesSizeChanged;
            LocationChanged += ServicesLocationChanged;
            dgv.KeyDown += DGVKeyDown;
        }

        #region utility methods
        private void SaveData()
        {
            _dataAdapter.Update(_dataTable);
        }
        #endregion

        #region control events

        /// <summary>
        /// traps delete key and asks for confirmation
        /// </summary>
        private void DGVKeyDown(object sender, KeyEventArgs e)
        {
            if (e.KeyCode != Keys.Delete) return;
            String msg = String.Format("{0}\n\rAre you sure you want to delete this Service?",
                                       dgv.SelectedRows[0].Cells["Name"].EditedFormattedValue);
            DialogResult result = MessageBox.Show(msg, "Delete Record", MessageBoxButtons.OKCancel,
                                                  MessageBoxIcon.Question);
            if (result == DialogResult.Cancel)
            {
                e.Handled = true;
            }
        }

        /// <summary>
        /// saves the form size and location to the config file
        /// </summary>
        private void ServicesLocationChanged(object sender, EventArgs e)
        {
            Common.SaveWindowSettings(ConfigKey, Size, Location);
        }

        /// <summary>
        /// saves the form size and location to the config file
        /// </summary>
        private void ServicesSizeChanged(object sender, EventArgs e)
        {
            Common.SaveWindowSettings(ConfigKey, Size, Location);
        }

        /// <summary>
        /// handles the close button click
        /// </summary>
        private void CloseButtonClick(object sender, EventArgs e)
        {
            Close();
        }

        /// <summary>
        /// resize objects as needed
        /// </summary>
        private void ServicesResize(object sender, EventArgs e)
        {
            dgv.Height = ClientRectangle.Height - panel1.Height;
        }

        /// <summary>
        /// save window settings when form is closing
        /// </summary>
        private void ServicesFormClosing(object sender, FormClosingEventArgs e)
        {
            SaveData();
            Common.SaveWindowSettings(ConfigKey, Size, Location);
        }
        
        /// <summary>
        /// Update the database after validation
        /// </summary>
        private void DGVValidated(object sender, EventArgs e)
        {
            SaveData();
        }

        #endregion

    }
}